/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2007, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************

    Module Name:
	action.c
 
    Abstract:
	Handle action frame.
 
    Revision History:
    Who         When          What
    --------    ----------    ----------------------------------------------
	Jan Lee		2006	  	created for rt2860 for 802.11n
 */

#include "rt_config.h"

/*  
    ==========================================================================
    Description: 
		action state machine init. Currently there are 8 category defined in 802.11spec.  We use the category value definition to 
		init corresponding action function in action state machine.
		
    Parameters: 
        S - pointer to the association state machine
    Note:
		The state machine looks like the following as name implies its function
    ==========================================================================
 */
VOID ActionStateMachineInit(
    IN	PRTMP_ADAPTER	pAd, 
    IN  STATE_MACHINE *S, 
    OUT STATE_MACHINE_FUNC Trans[]) 
{
	StateMachineInit(S, (STATE_MACHINE_FUNC*)Trans, MAX_ACT_STATE, MAX_ACT_MSG, (STATE_MACHINE_FUNC)Drop, ACT_IDLE, ACT_MACHINE_BASE);

	StateMachineSetAction(S, ACT_IDLE, MT2_PEER_SPECTRUM_CATE, (STATE_MACHINE_FUNC)PeerSpectrumAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_PEER_QOS_CATE, (STATE_MACHINE_FUNC)PeerQOSAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_PEER_DLS_CATE, (STATE_MACHINE_FUNC)PeerDLSAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_PEER_BA_CATE, (STATE_MACHINE_FUNC)PeerBAAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_PEER_RM_CATE, (STATE_MACHINE_FUNC)PeerRMAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_PEER_BSSTRANS_CATE, (STATE_MACHINE_FUNC)PeerBSSTranAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_PEER_HT_CATE, (STATE_MACHINE_FUNC)PeerHTAction);
	
	StateMachineSetAction(S, ACT_IDLE, MT2_MLME_ADD_BA_CATE, (STATE_MACHINE_FUNC)MlmeADDBAAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_MLME_ORI_DELBA_CATE, (STATE_MACHINE_FUNC)MlmeDELBAAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_MLME_REC_DELBA_CATE, (STATE_MACHINE_FUNC)MlmeDELBAAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_MLME_QOS_CATE, (STATE_MACHINE_FUNC)MlmeQOSAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_MLME_DLS_CATE, (STATE_MACHINE_FUNC)MlmeDLSAction);
	StateMachineSetAction(S, ACT_IDLE, MT2_ACT_INVALID, (STATE_MACHINE_FUNC)MlmeInvalidAction);
}

VOID MlmeADDBAAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
	MLME_ADDBA_REQ_STRUCT *pInfo;
	UCHAR			Addr[6];
	PUCHAR		   pOutBuffer = NULL;
	NDIS_STATUS 	NStatus;
	ULONG		Idx;
	FRAME_ADDBA_REQ  Frame;
	ULONG		FrameLen;

	pInfo = (MLME_ADDBA_REQ_STRUCT *)Elem->Msg;	
	NdisZeroMemory(&Frame, sizeof(FRAME_ADDBA_REQ));
	
	if(MlmeAddBAReqSanity(pAd, Elem->Msg, Elem->MsgLen, Addr)) 
	{
		NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer);  //Get an unused nonpaged memory
		if(NStatus != NDIS_STATUS_SUCCESS) 
		{
			DBGPRINT(RT_DEBUG_TRACE,("BA - MlmeADDBAAction() allocate memory failed \n"));
			return;
		}
		// 1. Prepare ADDBA REQ frame. Parameters in ADDBA REQ frame is advisory.
		Idx = pAd->MacTab.Content[pInfo->Wcid].BAOriWcidArray[pInfo->TID];

		ActHeaderInit(pAd, &Frame.Hdr, pInfo->pAddr, pAd->CommonCfg.Bssid);
		Frame.Category = CATEGORY_BA;
		Frame.Action = ADDBA_REQ;
		Frame.BaParm.Rsv = 0;
		Frame.BaParm.BAPolicy = IMMED_BA;
		Frame.BaParm.TID = pInfo->TID;
		Frame.BaParm.BufSize = pInfo->BaBufSize;
		Frame.Token = pInfo->Token;
		Frame.TimeOutValue = pInfo->TimeOutValue;
		Frame.BaStartSeq.field.FragNum = 0;
		Frame.BaStartSeq.field.StartSeq = pAd->MacTab.Content[pInfo->Wcid].BAOriSequence[pInfo->TID];
		DBGPRINT(RT_DEBUG_TRACE,("MlmeADDBAAction - ---->pInfo->Wcid = %d, TID = %d,  BAOriSequence = %x, \n",pInfo->Wcid, pInfo->TID, pAd->MacTab.Content[pInfo->Wcid].BAOriSequence[pInfo->TID]));

		MakeOutgoingFrame(pOutBuffer,				&FrameLen,
					  sizeof(FRAME_ADDBA_REQ),	  &Frame,
					  END_OF_ARGS);
		MiniportMMRequest(pAd, pOutBuffer, FrameLen);
	}
}

/*
    ==========================================================================
    Description:
        send DELBA and delete BaEntry if any
    Parametrs:
        Elem - MLME message MLME_DELBA_REQ_STRUCT
        
	IRQL = DISPATCH_LEVEL

    ==========================================================================
 */
VOID MlmeDELBAAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
	MLME_DELBA_REQ_STRUCT *pInfo;
	PUCHAR		   pOutBuffer = NULL;
	PUCHAR		   pOutBuffer2 = NULL;
	PUCHAR		   pOutBuffer3 = NULL;
	NDIS_STATUS     NStatus;
	ULONG		Idx;
	FRAME_DELBA_REQ  Frame;
	ULONG		FrameLen;
	FRAME_BAR	FrameBar;
	UCHAR				ZeroAddr[6];
	
	pInfo = (MLME_DELBA_REQ_STRUCT *)Elem->Msg;	
	// must send back DELBA 
	NdisZeroMemory(&Frame, sizeof(FRAME_DELBA_REQ));
	NdisZeroMemory(ZeroAddr, 6);
	
	if(MlmeDelBAReqSanity(pAd, Elem->Msg, Elem->MsgLen)) 
	{
		NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer2);  //Get an unused nonpaged memory
		if(NStatus != NDIS_STATUS_SUCCESS) 
		{
			DBGPRINT(RT_DEBUG_TRACE,("BA - MlmeDELBAAction() allocate memory failed \n"));
			return;
		}

		// Send BAR (Send BAR to refresh peer reordering buffer sequence index.)
		Idx = pAd->MacTab.Content[pInfo->Wcid].BAOriWcidArray[pInfo->TID];
		BarHeaderInit(pAd, &FrameBar, pAd->MacTab.Content[pInfo->Wcid].Addr);
		FrameBar.StartingSeq.field.FragNum = 0;  
		FrameBar.StartingSeq.field.StartSeq = pAd->MacTab.Content[pInfo->Wcid].BAOriSequence[pInfo->TID];  
		FrameBar.BarControl.TID = pAd->BATable.BAOriEntry[Idx].TID;  
		FrameBar.BarControl.ACKPolicy = IMMED_BA;   // mandatory use Immediate BA Policy
		FrameBar.BarControl.Compressed = 1;  // mandatory use Compressed 
		FrameBar.BarControl.MTID = 0;  
		MakeOutgoingFrame(pOutBuffer2,				&FrameLen,
					  sizeof(FRAME_BAR),	  &FrameBar,
					  END_OF_ARGS);
		//Send BAR to refresh peer reordering buffer.
		MiniportMMRequest(pAd, pOutBuffer2, FrameLen);
		DBGPRINT(RT_DEBUG_TRACE,("BA - MlmeDELBAAction() . Send BAR to refresh peer reordering buffer \n"));

		// Send DELBA action frame
		FrameLen = 0;
		if (!RTMPEqualMemory(pAd->MacTab.Content[pInfo->Wcid].Addr, ZeroAddr, 6))
		{
			NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer);  //Get an unused nonpaged memory
			if(NStatus != NDIS_STATUS_SUCCESS) 
			{
				DBGPRINT(RT_DEBUG_TRACE,("BA - MlmeDELBAAction() allocate memory failed\n"));
				MlmeFreeMemory(pAd, (PVOID)pOutBuffer2);			
				return;
			}
			NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer3);  //Get an unused nonpaged memory
			if(NStatus != NDIS_STATUS_SUCCESS) 
			{
				DBGPRINT(RT_DEBUG_TRACE,("BA - MlmeDELBAAction() allocate memory failed \n"));
				MlmeFreeMemory(pAd, (PVOID)pOutBuffer2);
				MlmeFreeMemory(pAd, (PVOID)pOutBuffer);
				return;
			}
			ActHeaderInit(pAd, &Frame.Hdr,  pAd->MacTab.Content[pInfo->Wcid].Addr, pAd->CommonCfg.Bssid);
			Frame.Category = CATEGORY_BA;
			Frame.Action = DELBA;
			Frame.DelbaParm.Initiator = pInfo->Initiator;
			Frame.DelbaParm.TID = pInfo->TID;
			MakeOutgoingFrame(pOutBuffer,               &FrameLen,
		              	sizeof(FRAME_DELBA_REQ),    &Frame,
		              	END_OF_ARGS);
			MiniportMMRequest(pAd, pOutBuffer, FrameLen);

			MakeOutgoingFrame(pOutBuffer3,               &FrameLen,
		              	sizeof(FRAME_DELBA_REQ),    &Frame,
		              	END_OF_ARGS);
			MiniportMMRequest(pAd, pOutBuffer3, FrameLen);
		}
	}
}


VOID MlmeQOSAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
}

VOID MlmeDLSAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
}

VOID MlmeInvalidAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
//	PUCHAR		   pOutBuffer = NULL;
	//Return the receiving frame except the MSB of category filed set to 1.  7.3.1.11
}

VOID PeerAddBAReqAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
	//	7.4.4.1
	ULONG	Idx;
	UCHAR	Status = 1;
	UCHAR	pAddr[6];
	FRAME_ADDBA_RSP	ADDframe;
	PUCHAR		   pOutBuffer = NULL;
	NDIS_STATUS 	NStatus;
	PFRAME_ADDBA_REQ  pAddreqFrame = NULL;
	UCHAR				BufSize = 0;
	ULONG		FrameLen;
	USHORT		StartSeq;
	PFRAME_802_11 		pFrame;

	//ADDBA Request from unknown peer, ignore this.
	if (Elem->Wcid >= MAX_LEN_OF_MAC_TABLE)
		return;

	if ((pAd->CommonCfg.BACapability.field.Policy == BA_NOTUSE) || (pAd->CommonCfg.IOTestParm.bNextDisableRxBA == TRUE))
	{
		Status = REASON_DECLINED;
		pFrame = (PFRAME_802_11)Elem->Msg;
		COPY_MAC_ADDR(pAddr, pFrame->Hdr.Addr2);
		DBGPRINT(RT_DEBUG_TRACE,("ACTION - PeerBAAction() Status = %d,   \n", Status));
	}
	else if (PeerAddBAReqActionSanity(pAd, Elem->Msg, Elem->MsgLen, pAddr)) 
	{
		pAddreqFrame = (PFRAME_ADDBA_REQ)(&Elem->Msg[0]);
		// 1. 
		Idx = pAd->MacTab.Content[Elem->Wcid].BARecWcidArray[pAddreqFrame->BaParm.TID];
		if ((pAd->MacTab.Content[Elem->Wcid].RXBAbitmap &(1<<pAddreqFrame->BaParm.TID)) == 0)
		{
			//Newly add
			BufSize = (pAddreqFrame->BaParm.BufSize > pAd->CommonCfg.BACapability.field.RxBAWinLimit)? (UCHAR)pAd->CommonCfg.BACapability.field.RxBAWinLimit:pAddreqFrame->BaParm.BufSize;
			// Intel Station send ADDBA with bufsize=0, we need to return bufsize=64.
			if (BufSize < 2)
				BufSize = 64;
			StartSeq = (pAddreqFrame->BaStartSeq.field.StartSeq == 0)?0xfff:(pAddreqFrame->BaStartSeq.field.StartSeq - 1);
			BATableInsertEntry(pAd, Elem->Wcid,  pAddreqFrame->TimeOutValue, StartSeq, (UCHAR)pAddreqFrame->BaParm.TID, BufSize, 0, TRUE);
		}
		else
		{
			// BA  agreement already setup. So return with existing NMSDU , Status Code = SUCCESS
			// Check !!!  change buffer size? rt2860
			BufSize = pAd->BATable.BARecEntry[Idx].BAWinSize;
		}
		Status = 0;
		DBGPRINT(RT_DEBUG_TRACE,("ACTION - PeerBAAction() BufSize = %d \n", BufSize));
	}
	
	pAddreqFrame = (PFRAME_ADDBA_REQ)(&Elem->Msg[0]);
	
	// 2. Always send back ADDBA Response 
	NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer);  //Get an unused nonpaged memory
	if (NStatus != NDIS_STATUS_SUCCESS) 
	{
		DBGPRINT(RT_DEBUG_ERROR,("ACTION - PeerBAAction() allocate memory failed \n"));
		return;
	}
	
	// 2-1. Prepare ADDBA Response frame.
	ActHeaderInit(pAd, &ADDframe.Hdr, pAddr, pAd->CommonCfg.Bssid);
	ADDframe.Category = CATEGORY_BA;
	ADDframe.Action = ADDBA_RESP;
	ADDframe.Token = pAddreqFrame->Token;
	// What is the Status code??  need to check.
	ADDframe.StatusCode = Status;
	ADDframe.BaParm.Rsv = 0;
	ADDframe.BaParm.BAPolicy = IMMED_BA;
	ADDframe.BaParm.TID = pAddreqFrame->BaParm.TID;
	ADDframe.BaParm.BufSize = BufSize;	
	ADDframe.TimeOutValue = pAddreqFrame->TimeOutValue;
	
	MakeOutgoingFrame(pOutBuffer,				&FrameLen,
			  sizeof(FRAME_ADDBA_RSP),	&ADDframe,
			  END_OF_ARGS);
	MiniportMMRequest(pAd, pOutBuffer, FrameLen);
	// 
	if ((pAd->CommonCfg.BACapability.field.AutoBA))
	{
		BATableInsertEntry(pAd, pAd->MacTab.Content[Elem->Wcid].Aid,  0, 0x1, 0, pAd->MacTab.Content[Elem->Wcid].BaSizeInUse,  Originator_SetAutoAdd, FALSE);
		pAd->MacTab.Content[Elem->Wcid].TXAutoBAbitmap |= (1);
	}
}

VOID PeerAddBARspAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
	UCHAR		Idx;
	PFRAME_ADDBA_RSP	pFrame = NULL;
	PBA_ORI_ENTRY		pBss;
	PUCHAR		   pOutBuffer1 = NULL;
	PUCHAR		   pOutBuffer2 = NULL;
	NDIS_STATUS 	NStatus;
	ULONG		FrameLen;
	FRAME_BAR	FrameBar;
	unsigned long   IrqFlags;

	//ADDBA Response from unknown peer, ignore this.
	if (Elem->Wcid >= MAX_LEN_OF_MAC_TABLE)
		return;
	
	if (PeerAddBARspActionSanity(pAd, Elem->Msg, Elem->MsgLen))
	{
		pFrame = (PFRAME_ADDBA_RSP)(&Elem->Msg[0]);
		
		Idx = pAd->MacTab.Content[Elem->Wcid].BAOriWcidArray[pFrame->BaParm.TID];
		// return on dismatched TID on ADDBA response 
		if ((pAd->BATable.BAOriEntry[Idx].TID != pFrame->BaParm.TID))
			return;
		
		DBGPRINT(RT_DEBUG_TRACE,("PeerAddBARsp---->StatusCode = %d, Wcid = %d, BAid = %d, TID=%d\n", pFrame->StatusCode, Elem->Wcid, Idx, pFrame->BaParm.TID));
		switch (pFrame->StatusCode)
		{
			case 0:
				// I am the originatorwant of this BA session.
				if ((pAd->BATable.BAOriEntry[Idx].ORI_BA_Status != Originator_Done))
				{
					NdisAcquireSpinLock(&pAd->BATabLock, IrqFlags);
					// Change parameter.
					pAd->BATable.BAOriEntry[Idx].BAWinSize = (UCHAR)pFrame->BaParm.BufSize;
					// update parameter from receiver. Ultimate parameter is based on ADDBA responsed frame.
					pBss = &pAd->BATable.BAOriEntry[Idx];
					pBss->ORI_BA_Status = Originator_Done; 
					pBss->TID = (UCHAR)pFrame->BaParm.TID;
					pAd->BATable.numAsOriginator++;
					pBss->BAWinSize = (UCHAR)pFrame->BaParm.BufSize;  //NMSDU will not surpass 8 bits
					if (pBss->BAWinSize == 0)
						pBss->BAWinSize = 1;  //NMSDU will not surpass 8 bits
						
					pBss->TimeOutValue= pFrame->TimeOutValue;
					pAd->MacTab.Content[Elem->Wcid].TXBAbitmap |= (1<<(pFrame->BaParm.TID));
					pAd->MacTab.Content[Elem->Wcid].TXAutoBAbitmap &= (~(1<<(pFrame->BaParm.TID)));
					NdisReleaseSpinLock(&pAd->BATabLock, IrqFlags);
					// SEND BAR ;
					NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer1);  //Get an unused nonpaged memory
					if(NStatus != NDIS_STATUS_SUCCESS) 
					{
						DBGPRINT(RT_DEBUG_TRACE,("BA - MlmeDELBAAction() allocate memory failed \n"));
						return;
					}
					NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer2);  //Get an unused nonpaged memory
					if(NStatus != NDIS_STATUS_SUCCESS) 
					{
						MlmeFreeMemory(pAd, (PVOID) pOutBuffer1);
						DBGPRINT(RT_DEBUG_TRACE,("BA - MlmeDELBAAction() allocate memory failed \n"));
						return;
					}

					BarHeaderInit(pAd, &FrameBar, pAd->MacTab.Content[Elem->Wcid].Addr);
					FrameBar.StartingSeq.field.FragNum = 0; // make sure sequence not clear in DEL function.
					FrameBar.StartingSeq.field.StartSeq = pAd->MacTab.Content[Elem->Wcid].BAOriSequence[pBss->TID]; // make sure sequence not clear in DEL funciton.
					FrameBar.BarControl.TID = pBss->TID; // make sure sequence not clear in DEL funciton.
					FrameBar.BarControl.ACKPolicy = 0; 
					FrameBar.BarControl.Compressed = 1; // make sure sequence not clear in DEL funciton.
					FrameBar.BarControl.MTID = 0; // make sure sequence not clear in DEL funciton.
					DBGPRINT(RT_DEBUG_TRACE,("PeerAddBARsp ----->Wcid = %d, TID = %d,  BAOriSequence = %x, \n",Elem->Wcid, pBss->TID, pAd->MacTab.Content[Elem->Wcid].BAOriSequence[pBss->TID]));
					MakeOutgoingFrame(pOutBuffer1,				&FrameLen,
								  sizeof(FRAME_BAR),	  &FrameBar,
								  END_OF_ARGS);
					MiniportMMRequest(pAd, pOutBuffer1, FrameLen);
					MakeOutgoingFrame(pOutBuffer2,				&FrameLen,
								  sizeof(FRAME_BAR),	  &FrameBar,
								  END_OF_ARGS);
					MiniportMMRequest(pAd, pOutBuffer2, FrameLen);
				}

			break;
			case REASON_DECLINED:
				if (pAd->BATable.BAOriEntry[Idx].ORI_BA_Status == Originator_WaitRes)
				{
					BATableTearORIEntry(pAd, (UCHAR)pFrame->BaParm.TID, Elem->Wcid, TRUE, FALSE);
				}
			break;
			default:
				if (pAd->BATable.BAOriEntry[Idx].ORI_BA_Status == Originator_WaitRes)
				{
					// After retry 10 times, delete.
					if (++pAd->BATable.BAOriEntry[Idx].Token > 20)
					{
						BATableTearORIEntry(pAd, (UCHAR)pFrame->BaParm.TID, Elem->Wcid, FALSE, FALSE);
					}
					else
					{
						// Decrease Buf zise and retry again.
						MLME_ADDBA_REQ_STRUCT	AddbaReq;
						if (pAd->BATable.BAOriEntry[Idx].BAWinSize > 3)
							pAd->BATable.BAOriEntry[Idx].BAWinSize =(pAd->BATable.BAOriEntry[Idx].BAWinSize)*7/8;
						
						NdisZeroMemory(&AddbaReq, sizeof(AddbaReq));
						COPY_MAC_ADDR(AddbaReq.pAddr, pAd->MacTab.Content[Elem->Wcid].Addr);
						AddbaReq.Wcid = (UCHAR)(Elem->Wcid);
						AddbaReq.TID = (UCHAR)pFrame->BaParm.TID;
						AddbaReq.Token = pAd->BATable.BAOriEntry[Idx].Token;
						AddbaReq.BaBufSize = pAd->BATable.BAOriEntry[Idx].BAWinSize;
						AddbaReq.TimeOutValue = pAd->BATable.BAOriEntry[Idx].TimeOutValue;
						MlmeEnqueue(pAd, ACTION_STATE_MACHINE, MT2_MLME_ADD_BA_CATE, sizeof(MLME_ADDBA_REQ_STRUCT), (PVOID)&AddbaReq);
						RTUSBMlmeUp(pAd);	
					}
						
				}
			break;

		}
		DBGPRINT(RT_DEBUG_TRACE,("PeerAddBARspAction - ----> numAsOriginator = %ld, \n",pAd->BATable.numAsOriginator));
	}
}

VOID PeerDelBAAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
	UCHAR		Idx;
	PFRAME_DELBA_REQ  pDelFrame = NULL;
	
	//DELBA Request from unknown peer, ignore this.
	if (PeerDelBAActionSanity(pAd, Elem->Wcid, Elem->Msg, Elem->MsgLen))
	{
		pDelFrame = (PFRAME_DELBA_REQ)(&Elem->Msg[0]);
		if (pDelFrame->DelbaParm.Initiator == ORIGINATOR)
		{
			DBGPRINT(RT_DEBUG_TRACE,("BA - PeerDelBAAction----> ORIGINATOR\n"));
			Idx = pAd->MacTab.Content[Elem->Wcid].BARecWcidArray[pDelFrame->DelbaParm.TID];
			if (pDelFrame->DelbaParm.TID == pAd->BATable.BARecEntry[Idx].TID)
				BATableDeleteRECEntry(pAd, &(pAd->BATable.BARecEntry[Idx]));
		}
		else 
		{
			DBGPRINT(RT_DEBUG_TRACE,("BA - PeerDelBAAction----> RECIPIENT\n"));
			Idx = pAd->MacTab.Content[Elem->Wcid].BAOriWcidArray[pDelFrame->DelbaParm.TID];
			if (pDelFrame->DelbaParm.TID == pAd->BATable.BAOriEntry[Idx].TID)
			BATableDeleteORIEntry(pAd, &(pAd->BATable.BAOriEntry[Idx]));
			if (pAd->CommonCfg.BACapability.field.AutoBA == TRUE)
			{
				if (INFRA_ON(pAd) && (pAd->StaActive.SupportedHtPhy.bHtEnable == TRUE))
				{
					// If add myself as one Recipient of peer (pAd->CommonCfg.Bssid), the sequence 0xffff indicates the BA
					// session not started yet. 
					BATableInsertEntry(pAd, BSSID_WCID,  0, 0x1, (UCHAR)pDelFrame->DelbaParm.TID, pAd->MacTab.Content[BSSID_WCID].BaSizeInUse,  Originator_SetAutoAdd, FALSE);
					pAd->MacTab.Content[BSSID_WCID].TXAutoBAbitmap |= (1<<pDelFrame->DelbaParm.TID);
				}
				else if (pAd->MacTab.Content[Elem->Wcid].Sst == SST_ASSOC)
				{
					BATableInsertEntry(pAd, Elem->Wcid,  0, 0x1, (UCHAR)pDelFrame->DelbaParm.TID, pAd->MacTab.Content[Elem->Wcid].BaSizeInUse,  Originator_SetAutoAdd, FALSE);
					pAd->MacTab.Content[Elem->Wcid].TXAutoBAbitmap |= (1<<pDelFrame->DelbaParm.TID);
				}
			}
		} 
	}
}

VOID PeerSpectrumAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{

	UCHAR	Action = Elem->Msg[LENGTH_802_11+1];
	EXT_CHA_SWITCH_IE	ExtChaSwitchIE;
	
	switch(Action)
	{
		case SPEC_CHANNEL_SWITCH:
		{
//			FRAME_SPETRUM_CS	FrameCS;
			break;
		}
		case SPEC_EXTCHANNEL_SWITCH:	// 7.3.2.51
			RTMPMoveMemory(&ExtChaSwitchIE, &Elem->Msg[LENGTH_802_11+2], 4);
			DBGPRINT(RT_DEBUG_TRACE,("Spectrum. New Channel = %d, New Class = %d \n", ExtChaSwitchIE.NewChannel, ExtChaSwitchIE.NewClass));
			break;
	}
// unknown category field value received.
// return to source with most significant bit set to 1 if received packet's MSB = 0
}

VOID PeerQOSAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
}

VOID PeerDLSAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
}

VOID PeerBAAction(
    IN PRTMP_ADAPTER pAd, 
    IN MLME_QUEUE_ELEM *Elem) 
{
	UCHAR	Action = Elem->Msg[LENGTH_802_11+1];
	
	switch(Action)
	{
		case ADDBA_REQ:
			PeerAddBAReqAction(pAd,Elem);
			DBGPRINT(RT_DEBUG_TRACE,("BA - PeerBAAction() Action = ADDBA_REQ \n"));
			break;
		case ADDBA_RESP:
			PeerAddBARspAction(pAd,Elem);
			DBGPRINT(RT_DEBUG_TRACE,("BA - PeerBAAction() Action = ADDBA_RESP \n"));
			break;
		case DELBA:
			PeerDelBAAction(pAd,Elem);
			DBGPRINT(RT_DEBUG_TRACE,("BA - PeerBAAction() Action = PeerDelBAAction( \n"));
			break;
	}
}

VOID PeerRMAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
}

VOID PeerBSSTranAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
}

VOID PeerHTAction(
	IN PRTMP_ADAPTER pAd, 
	IN MLME_QUEUE_ELEM *Elem) 
{
	UCHAR	Action = Elem->Msg[LENGTH_802_11+1];
	UCHAR	Information;
	
	if (Elem->Wcid >= MAX_LEN_OF_MAC_TABLE)
		return;
	
	switch(Action)
	{
		case Notify_BW_ACTION:
			DBGPRINT(RT_DEBUG_TRACE,("ACTION - HT Notify Channel bandwidth action----> \n"));
			if (Elem->Msg[LENGTH_802_11+2] == 0)	// 7.4.8.2. if value is 1, keep the same as supported channel bandwidth. 
				pAd->MacTab.Content[Elem->Wcid].HTPhyMode.field.BW = 0;
			
			break;
		case SMPS_ACTION:
			// 7.3.1.25
 			DBGPRINT(RT_DEBUG_TRACE,("ACTION - SMPS action----> \n"));
			if (((Elem->Msg[LENGTH_802_11+2]&0x1) == 0))
			{
				pAd->MacTab.Content[Elem->Wcid].MmpsMode = MMPS_ENABLE;
			}
			else if (((Elem->Msg[LENGTH_802_11+2]&0x2) == 0))
			{
				pAd->MacTab.Content[Elem->Wcid].MmpsMode = MMPS_STATIC;
			}
			else
			{
				pAd->MacTab.Content[Elem->Wcid].MmpsMode = MMPS_DYNAMIC;
			}
			break;
 
		case SETPCO_ACTION:
			break;
		case MIMO_CHA_MEASURE_ACTION:
			break;
		case HT_INFO_EXCHANGE:
			Information = Elem->Msg[LENGTH_802_11+2];
 			DBGPRINT(RT_DEBUG_TRACE,("ACTION - HT Information Exchange= %x----> \n", Information));
			break;
	}
}

VOID RECBATimerTimeout(
    IN PVOID SystemSpecific1, 
    IN PVOID FunctionContext, 
    IN PVOID SystemSpecific2, 
    IN PVOID SystemSpecific3) 
{
	RTMP_ADAPTER	*pAd = (RTMP_ADAPTER *)FunctionContext;
	PBA_REC_ENTRY	pBAEntry;
	ULONG	 RecTabSize;
	PRTMP_REORDERBUF pBuffer;
	UCHAR		count, num;
	unsigned long   IrqFlags;
	
	RecTabSize = pAd->BATable.numAsRecipient;
	for (num = 0; num < MAX_LEN_OF_BA_REC_TABLE;num++)
	{
		if (RecTabSize == 0)
			break;
		pBAEntry = &pAd->BATable.BARecEntry[num];
		//if ((pAd->MacTab.Content[BSSID_WCID].RXBAbitmap&(1<<UserPriority)) != 0) 
		if (pBAEntry->REC_BA_Status == Recipient_Accept) 
		{
			RecTabSize--;
			if ((pBAEntry->LastIndSeq == pBAEntry->LastIndSeqAtTimer) && (pBAEntry->NumOfRxPkt > 0))
			{
				DBGPRINT(RT_DEBUG_TRACE, ("6-%d. Timeout Refresh. LastIndSeq = 0x%x.	NumOfRxPkt = 0x%x\n",num, pBAEntry->LastIndSeq, pBAEntry->NumOfRxPkt));
				NdisAcquireSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
				count = pBAEntry->BAWinSize;
				do
				{
					pBuffer = &pAd->LocalRxReorderBuf[pBAEntry->RxBufIdxUsed].MAP_RXBuf[pBAEntry->Curindidx];
					//pBuffer = &pBAEntry->MAP_RXBuf[pBAEntry->Curindidx];
					if (pBuffer->IsFull == TRUE)
					{
						if (pBuffer->IsAMSDU)
						{
							REPORT_AMSDU_FRAMES_TO_LLC(pAd, (PUCHAR)((PUCHAR)(pBuffer->pBuffer)+pBuffer->DataOffset), pBuffer->Datasize);
						}
						else
							REPORT_ETHERNET_FRAME_TO_LLC(pAd, pBuffer->Header802_3, (PUCHAR)((PUCHAR)(pBuffer->pBuffer)+pBuffer->DataOffset), pBuffer->Datasize);
						pBuffer->IsFull = FALSE;
						pBAEntry->NumOfRxPkt--;
						DBGPRINT(RT_DEBUG_INFO, ("66-%d.  INDI. %dth buffer. pBuffer->Sequence.value = 0x%x\n" ,num,pBAEntry->Curindidx, pBuffer->Sequence.value));
						if (pBAEntry->NumOfRxPkt == 0)
						{
							// We already indicate sequence, so we still to head the index brefore break.
							INC_RING_INDEX(pBAEntry->Curindidx, MAX_RX_REORDERBUF);
							pBAEntry->LastIndSeq = pBuffer->Sequence.value&MAXSEQ;	
							break;
						}
					}
					pBAEntry->LastIndSeq++;	
					pBAEntry->LastIndSeq &= MAXSEQ;
					INC_RING_INDEX(pBAEntry->Curindidx, MAX_RX_REORDERBUF);
					count--;
				}while (count > 0);
				NdisReleaseSpinLock(&pBAEntry->RxReRingLock, IrqFlags);

				DBGPRINT(RT_DEBUG_INFO, ("666. LastIndSeq = 0x%x. NumOfRxPkt = %d\n" ,pBAEntry->LastIndSeq, pBAEntry->NumOfRxPkt));
			}
			pBAEntry->LastIndSeqAtTimer = pBAEntry->LastIndSeq;
				
		}			
	}
}

/*
	==========================================================================
	Description:
		Retry sending ADDBA Reqest.
		
	IRQL = DISPATCH_LEVEL
	
	Parametrs:
	p8023Header: if this is already 802.3 format, p8023Header is NULL
	
	Return	: TRUE if put into rx reordering buffer, shouldn't indicaterxhere.
				FALSE , then continue indicaterx at this moment.
	==========================================================================
 */
VOID ORIBATimerTimeout(
	IN	PRTMP_ADAPTER	pAd) 
{
	MAC_TABLE_ENTRY	*pEntry;
	ULONG	i,  total;
//	PUCHAR			pOutBuffer = NULL;

	total = pAd->MacTab.Size * NUM_OF_TID;
	
	for (i = 1; ((i <MAX_LEN_OF_BA_ORI_TABLE) && (total > 0)) ; i++)
	{
		if  ((pAd->BATable.BAOriEntry[i].ORI_BA_Status == Originator_Done) && 
			TKIP_ON(pAd) && (pAd->RalinkCounters.OneSecTransmittedByteCount > 1000))
		{
			/*NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer);  //Get an unused nonpaged memory
			if(NStatus != NDIS_STATUS_SUCCESS) 
			{
				DBGPRINT(RT_DEBUG_TRACE,("BA - MlmeDELBAAction() allocate memory failed \n"));
				return;
			}
			NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer);  //Get an unused nonpaged memory
			if(NStatus != NDIS_STATUS_SUCCESS) 
			{
				DBGPRINT(RT_DEBUG_ERROR,("BA - MlmeADDBAAction() allocate memory failed \n"));
				return;
			}
			DBGPRINT(RT_DEBUG_TRACE,("BA - ORIBATimerTimeout.OneSecTransmittedByteCount  = 0x%x \n", pAd->RalinkCounters.OneSecTransmittedByteCount));
			pEntry = &pAd->MacTab.Content[pAd->BATable.BAOriEntry[i].Wcid];
			if (pEntry->NoDataIdleCount == 0)
			{
				Sequence = pEntry->BAOriSequence[pAd->BATable.BAOriEntry[i].TID];
				if (Sequence < pAd->BATable.BAOriEntry[i].BAWinSize)
					Sequence = 0;
				else
					Sequence -= pAd->BATable.BAOriEntry[i].BAWinSize;
				Sequence ++;
				BarHeaderInit(pAd, &FrameBar, pAd->MacTab.Content[pAd->BATable.BAOriEntry[i].Wcid].Addr);

				FrameBar.StartingSeq.field.FragNum = 0; // make sure sequence not clear in DEL function.
				FrameBar.StartingSeq.field.StartSeq = Sequence; // make sure sequence not clear in DEL funciton.
				FrameBar.BarControl.TID = pAd->BATable.BAOriEntry[i].TID; // make sure sequence not clear in DEL funciton.
				MakeOutgoingFrame(pOutBuffer,				&FrameLen,
							  	sizeof(FRAME_BAR),	  &FrameBar,
								END_OF_ARGS);
				MiniportMMRequest(pAd, pOutBuffer, FrameLen);*/
		}
		else if ((pAd->BATable.BAOriEntry[i].ORI_BA_Status == Originator_WaitRes) && (pAd->BATable.BAOriEntry[i].Token++ < 25))
		{
			MLME_ADDBA_REQ_STRUCT	 AddbaReq;	
			pEntry = &pAd->MacTab.Content[pAd->BATable.BAOriEntry[i].Wcid];
			NdisZeroMemory(&AddbaReq, sizeof(AddbaReq));
			COPY_MAC_ADDR(AddbaReq.pAddr, pEntry->Addr);
			AddbaReq.Wcid = (UCHAR)(pEntry->Aid);
			AddbaReq.TID = pAd->BATable.BAOriEntry[i].TID;
			AddbaReq.BaBufSize = pEntry->BaSizeInUse;
			AddbaReq.TimeOutValue = 0;
			AddbaReq.Token = (++pAd->BATable.BAOriEntry[i].Token);
			MlmeEnqueue(pAd, ACTION_STATE_MACHINE, MT2_MLME_ADD_BA_CATE, sizeof(MLME_ADDBA_REQ_STRUCT), (PVOID)&AddbaReq);
			RTUSBMlmeUp(pAd);	
			DBGPRINT(RT_DEBUG_TRACE,("BA - ADDBATimer. Send to wcid = %d, TID= %d\n", AddbaReq.Wcid, AddbaReq.TID));
		}
		else if ((pAd->BATable.BAOriEntry[i].ORI_BA_Status == Originator_WaitAdd) &&
			    (NO_SECURITY_ON(pAd)||AES_ON(pAd)))
		{
			// Check 
			MLME_ADDBA_REQ_STRUCT	 AddbaReq;	
			pAd->BATable.BAOriEntry[i].ORI_BA_Status = Originator_WaitRes;
			pEntry = &pAd->MacTab.Content[pAd->BATable.BAOriEntry[i].Wcid];
			NdisZeroMemory(&AddbaReq, sizeof(AddbaReq));
			COPY_MAC_ADDR(AddbaReq.pAddr, pEntry->Addr);
			AddbaReq.Wcid = (UCHAR)(pEntry->Aid);
			AddbaReq.TID = pAd->BATable.BAOriEntry[i].TID;
			AddbaReq.BaBufSize = pEntry->BaSizeInUse;
			AddbaReq.TimeOutValue = 0;
			AddbaReq.Token = 1;
			MlmeEnqueue(pAd, ACTION_STATE_MACHINE, MT2_MLME_ADD_BA_CATE, sizeof(MLME_ADDBA_REQ_STRUCT), (PVOID)&AddbaReq);
			RTUSBMlmeUp(pAd);	
			DBGPRINT(RT_DEBUG_TRACE,("BA - ADDBATimer. Send to wcid = %d, TID= %d\n", AddbaReq.Wcid, AddbaReq.TID));
		}
		else if ((pAd->BATable.BAOriEntry[i].ORI_BA_Status == Originator_WaitAdd))
		{
			pEntry = &pAd->MacTab.Content[pAd->BATable.BAOriEntry[i].Wcid];
			if ((pAd->RalinkCounters.OneSecTxRetryOkCount > 10) && (pEntry->bIAmBadAtheros == FALSE) &&
			CLIENT_STATUS_TEST_FLAG(&pAd->MacTab.Content[BSSID_WCID], fCLIENT_STATUS_RALINK_CHIPSET))
			{
				MLME_ADDBA_REQ_STRUCT	 AddbaReq;	
				pAd->BATable.BAOriEntry[i].ORI_BA_Status = Originator_WaitRes;
				NdisZeroMemory(&AddbaReq, sizeof(AddbaReq));
				COPY_MAC_ADDR(AddbaReq.pAddr, pEntry->Addr);
				AddbaReq.Wcid = (UCHAR)(pEntry->Aid);
				AddbaReq.TID = pAd->BATable.BAOriEntry[i].TID;
				AddbaReq.BaBufSize = pEntry->BaSizeInUse;
				AddbaReq.TimeOutValue = 0;
				AddbaReq.Token = (++pAd->BATable.BAOriEntry[i].Token);
				MlmeEnqueue(pAd, ACTION_STATE_MACHINE, MT2_MLME_ADD_BA_CATE, sizeof(MLME_ADDBA_REQ_STRUCT), (PVOID)&AddbaReq);
				RTUSBMlmeUp(pAd);	
				DBGPRINT(RT_DEBUG_TRACE,("BA - 2ADDBATimer. Send to wcid = %d, TID= %d\n", AddbaReq.Wcid, AddbaReq.TID));
			}
			else if ((pEntry->bIAmBadAtheros == TRUE) || (pAd->CommonCfg.IOTestParm.bNextDisableRxBA == TRUE))
			{
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_AMSDU_INUSED);
				pAd->BATable.BAOriEntry[i].ORI_BA_Status = Originator_NONE;
				DBGPRINT(RT_DEBUG_TRACE,("BA - 2ADDBATimer. fCLIENT_STATUS_AMSDU_INUSED\n"));
			}
			else
				DBGPRINT(RT_DEBUG_TRACE,("BA - 2ADDBATimer. pAd->RalinkCounters.OneSecTxRetryOkCount  = %ld \n", pAd->RalinkCounters.OneSecTxRetryOkCount ));
		}
		else if ((pAd->BATable.BAOriEntry[i].ORI_BA_Status == Originator_WaitRes))
		{
			if (pAd->CommonCfg.BACapability.field.AutoBA == 1)
			{
				pAd->BATable.BAOriEntry[i].ORI_BA_Status = Originator_SetAutoAdd;
				pEntry = &pAd->MacTab.Content[pAd->BATable.BAOriEntry[i].Wcid];
				pEntry->TXAutoBAbitmap |= (1<<(pAd->BATable.BAOriEntry[i].TID));
			}
			else
			pAd->BATable.BAOriEntry[i].ORI_BA_Status = Originator_NONE;
			
			pAd->BATable.BAOriEntry[i].Token = 1;
		}
		total --;
	}
	pAd->RalinkCounters.OneSecTransmittedByteCount = 0;
}

VOID ActHeaderInit(
	IN	PRTMP_ADAPTER	pAd, 
	IN OUT PHEADER_802_11 pHdr80211, 
	IN PUCHAR pDA, 
	IN PUCHAR pBssid) 
{
	NdisZeroMemory(pHdr80211, sizeof(HEADER_802_11));
	pHdr80211->FC.Type = BTYPE_MGMT;
	pHdr80211->FC.SubType = SUBTYPE_ACTION;
	COPY_MAC_ADDR(pHdr80211->Addr1, pDA);
	COPY_MAC_ADDR(pHdr80211->Addr2, pAd->CurrentAddress);
	COPY_MAC_ADDR(pHdr80211->Addr3, pBssid);
}

VOID BarHeaderInit(
	IN	PRTMP_ADAPTER	pAd, 
	IN OUT PFRAME_BAR pCntlBar, 
	IN PUCHAR pDA) 
{
	
	NdisZeroMemory(pCntlBar, sizeof(FRAME_BAR));
	pCntlBar->FC.Type = BTYPE_CNTL;
	pCntlBar->FC.SubType = SUBTYPE_BLOCK_ACK_REQ;
	pCntlBar->BarControl.MTID = 0;
	pCntlBar->BarControl.Compressed = 1;
	pCntlBar->BarControl.ACKPolicy = 0;

	pCntlBar->Duration = 16 + RTMPCalcDuration(pAd, RATE_1, sizeof(FRAME_BA));

	COPY_MAC_ADDR(pCntlBar->Addr1, pDA);
	COPY_MAC_ADDR(pCntlBar->Addr2, pAd->CurrentAddress);
}

/*
    ==========================================================================
    Description:
		Check Every QoSData frame's sequence in BA session to decide whether 1.put to reordering buffer or 
		2. indicate receive because it's in sequence.
        
	IRQL = DISPATCH_LEVEL
	
	Parametrs:
	p8023Header: if this is already 802.3 format, p8023Header is NULL
	
	Return	: TRUE if put into rx reordering buffer, shouldn't indicaterxhere.
				FALSE , then continue indicaterx at this moment.
    ==========================================================================
 */
BOOLEAN QosBADataParse(
    IN PRTMP_ADAPTER	pAd, 
	IN BOOLEAN bAMSDU,
    IN PUCHAR p8023Header,
    IN UCHAR	Wcid,
    IN UCHAR	TID,
    IN USHORT Sequence,
    IN UCHAR DataOffset, 
    IN PUCHAR pData, 
    IN USHORT Datasize)
{
	ULONG	Idx = 3;
	BOOLEAN 	brc = FALSE;
	PBA_REC_ENTRY	pBAEntry;
	PRTMP_REORDERBUF	pReBuffer;
	PHEADER_802_11	pHeader;
	INT	i;
	unsigned long   IrqFlags;
	
	pHeader = (PHEADER_802_11)(pData + 4 + RXWI_SIZE); 	// USB has 4 bytes as "Length"

	if (Wcid < MAX_LEN_OF_MAC_TABLE)
	{
		// if this receiving packet is from SA that is in our OriEntry. Since WCID <9 has direct mapping. no need search.
		Idx = pAd->MacTab.Content[Wcid].BARecWcidArray[TID];
		pBAEntry = &pAd->BATable.BARecEntry[Idx];
	}
	else
	{
		return FALSE;
	}
	
	if (pBAEntry->REC_BA_Status == Recipient_Accept)
	{
		// find corresponding BaEntry and IMME BACK policy in use, so put into reordering buffer
		//  *****************************************************************************
		//  start insert , NO Wait For BAR
		//  *****************************************************************************
		// 0-0. update last sequence 
		// 1. check if in order.
		if (SEQ_STEPONE (pHeader->Sequence, pBAEntry->LastIndSeq) )
		{
			NdisAcquireSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
			pBAEntry->LastIndSeq = pHeader->Sequence;
			INC_RING_INDEX(pBAEntry->Curindidx, MAX_RX_REORDERBUF);			
			NdisReleaseSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
			brc = TRUE;
			if (bAMSDU)
			{
				REPORT_AMSDU_FRAMES_TO_LLC(pAd, (PUCHAR)((pData)+DataOffset), Datasize);
			}
			else
				REPORT_ETHERNET_FRAME_TO_LLC(pAd, p8023Header, (PUCHAR)(pData+DataOffset), Datasize);

			if (pBAEntry->NumOfRxPkt > 0)
			{
				NdisAcquireSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
				for ( i=1; i < pBAEntry->BAWinSize ; i++)
				{
					pReBuffer = &pAd->LocalRxReorderBuf[pBAEntry->RxBufIdxUsed].MAP_RXBuf[pBAEntry->Curindidx];
					//pBuffer = &pBAEntry->MAP_RXBuf[pBAEntry->Curindidx];
					if (pReBuffer->IsFull == TRUE)
					{
						if (pReBuffer->IsAMSDU)
						{
							REPORT_AMSDU_FRAMES_TO_LLC(pAd, (PUCHAR)((PUCHAR)(pReBuffer->pBuffer)+pReBuffer->DataOffset), pReBuffer->Datasize);
						}
						else
							REPORT_ETHERNET_FRAME_TO_LLC(pAd, pReBuffer->Header802_3, (PUCHAR)((pReBuffer->pBuffer)+pReBuffer->DataOffset), pReBuffer->Datasize);
						pBAEntry->LastIndSeq = pReBuffer->Sequence.value&MAXSEQ;		
						pReBuffer->IsFull = FALSE;
						pBAEntry->NumOfRxPkt-- ;
						DBGPRINT(RT_DEBUG_INFO, ("22-%ld. INDI. %dth buffer. pBuffer->Sequence = 0x%x, NumOfRxPkt = %d\n" , Idx,pBAEntry->Curindidx, pReBuffer->Sequence.value, pBAEntry->NumOfRxPkt));
						INC_RING_INDEX(pBAEntry->Curindidx, MAX_RX_REORDERBUF);
					}
					else
					{
						break;
					}
				}
			    	NdisReleaseSpinLock(&pBAEntry->RxReRingLock,  IrqFlags);
			}
			return TRUE;

		}
		//  rule 2.  Duplicate check.
		else if (pHeader->Sequence == pBAEntry->LastIndSeq)
		{
			DBGPRINT(RT_DEBUG_INFO, ("33-%ld. DUPLICATE receive pHeader->Sequence = 0x%x ,LastIndSeq = 0x%x. drop \n" , Idx, pHeader->Sequence, pBAEntry->LastIndSeq));
			pAd->WlanCounters.FrameDuplicateCount.LowPart ++;
			//drop this
			brc = TRUE;
		}
		// ====================================================
		// check rule 2-2.   Duplicate check.
		else if(SEQ_SMALLER(pHeader->Sequence, pBAEntry->LastIndSeq))
		{
			DBGPRINT(RT_DEBUG_INFO, ("3-%ld. DUPLICATE receive pHeader->Sequence = 0x%x ,LastIndSeq = 0x%x. drop \n" , Idx, pHeader->Sequence, pBAEntry->LastIndSeq));
			pAd->WlanCounters.FrameDuplicateCount.LowPart ++;
			//drop this
			brc = TRUE;
		}
		// 4 Receive seq within Window Size
		else if (SEQ_WITHIN_WIN(pHeader->Sequence, pBAEntry->LastIndSeq, pBAEntry->BAWinSize))
		{
			// Put to CurIndex + (RecSeq - LastSeq) -1
			i = (pHeader->Sequence > pBAEntry->LastIndSeq)?(pHeader->Sequence - pBAEntry->LastIndSeq -1):(pHeader->Sequence + MAXSEQ - pBAEntry->LastIndSeq);
			if ( i > pBAEntry->BAWinSize)
			{
				DBGPRINT(RT_DEBUG_TRACE, ("!!pHeader->Sequence = 0x%x,pBAEntry->LastIndSeq = 0x%x, Header->Frag = 0x%x \n", pHeader->Sequence, pBAEntry->LastIndSeq, pHeader->Frag));
				brc = TRUE;
			}
			else
			{
				pReBuffer = &pAd->LocalRxReorderBuf[pBAEntry->RxBufIdxUsed].MAP_RXBuf[((pBAEntry->Curindidx+i)%MAX_RX_REORDERBUF)];
				if (pReBuffer->IsFull == FALSE)
				{
					// Write RxD buffer address & allocated buffer length
					NdisAcquireSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
					pReBuffer->Sequence.value = (USHORT)pHeader->Sequence & MAXSEQ;
					pReBuffer->Datasize = Datasize;
					NdisMoveMemory(pReBuffer->Header802_3, p8023Header, LENGTH_802_3);
					pReBuffer->DataOffset = DataOffset;
					//Copy packet to reorder buffer
					NdisMoveMemory(pReBuffer->pBuffer, pData, (Datasize+DataOffset));
					pBAEntry->NumOfRxPkt++;  
					pReBuffer->IsFull = TRUE;
					pReBuffer->IsAMSDU = bAMSDU;
					NdisReleaseSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
					DBGPRINT(RT_DEBUG_INFO, ("4-%ld. Rx out of order. Insert[%dth] to [%dth]   pBAEntry->LastIndSeq =0x%x  pHeader->Seq = 0x%x, DataOffset = 0x%x, Datasize = 0x%x\n",\
						 Idx,i, pBAEntry->Curindidx ,pBAEntry->LastIndSeq , pHeader->Sequence, pReBuffer->DataOffset, pReBuffer->Datasize));
					// don't indicate 
					brc = TRUE;
				}
				else
				{
					DBGPRINT(RT_DEBUG_INFO,  ("44-%ld. Duplicate  Reorder Buffer pHeader->Sequence = 0x%x,\n", Idx, pHeader->Sequence));
					// don't indicate 
					brc = TRUE;
				}
			}
		}
		// 5.1 Receive seq surpasses Win(lastseq + nMSDU). So refresh all reorder buffer
		else
		{
			DBGPRINT(RT_DEBUG_TRACE, ("5-%ld. Refresh. pHeader->Sequence=0x%x, #RxPkt = %d. Curindidx = 0x%x  LastIndSeq  = 0x%x.\n", Idx,pHeader->Sequence,pBAEntry->NumOfRxPkt, pBAEntry->Curindidx, pBAEntry->LastIndSeq  ));
			// Then refresh all reorder buffer. and update LastSeq = ReceiveSeq
			NdisAcquireSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
			DBGPRINT(RT_DEBUG_INFO, ("55-%ld. INDI Sequences are : \n", Idx));
			for ( i=0; i < pBAEntry->BAWinSize; i++)
			{
				pReBuffer = &pAd->LocalRxReorderBuf[pBAEntry->RxBufIdxUsed].MAP_RXBuf[pBAEntry->Curindidx];
				//pBuffer = &pBAEntry->MAP_RXBuf[pBAEntry->Curindidx];
				if (pReBuffer->IsFull == TRUE)
				{
					if (pReBuffer->IsAMSDU)
					{
						REPORT_AMSDU_FRAMES_TO_LLC(pAd, (PUCHAR)((PUCHAR)(pReBuffer->pBuffer)+pReBuffer->DataOffset), pReBuffer->Datasize);
					}
					else
						REPORT_ETHERNET_FRAME_TO_LLC(pAd, pReBuffer->Header802_3, (PUCHAR)(pReBuffer->pBuffer+pReBuffer->DataOffset), pReBuffer->Datasize);
					pBAEntry->LastIndSeq = pReBuffer->Sequence.value&MAXSEQ;	
					pReBuffer->IsFull = FALSE;
					if (pBAEntry->NumOfRxPkt == 0)
					{	DBGPRINT_ERR( ("55 Error Counting Reorder Buffer\n"));
					}
					else
						pBAEntry->NumOfRxPkt--;
					DBGPRINT(RT_DEBUG_INFO, ("0x%x, ", pReBuffer->Sequence.value));
				}
				INC_RING_INDEX(pBAEntry->Curindidx, MAX_RX_REORDERBUF);
			}
			DBGPRINT(RT_DEBUG_INFO, ("\n"));
			NdisReleaseSpinLock(&pBAEntry->RxReRingLock, IrqFlags);

			if (pBAEntry->LastIndSeq == pHeader->Sequence  )
			{
				DBGPRINT_ERR( ("555-%ld refresh . pHeader->Sequence = %x\n", Idx, pHeader->Sequence));
				//not receive here
				brc = TRUE;
			}
			else
			{
				pBAEntry->LastIndSeq = pHeader->Sequence;
				DBGPRINT(RT_DEBUG_TRACE, ("55-%ld  INDI.  pHeader->Sequence  0x%x\n" , Idx,pBAEntry->LastIndSeq));
				// receive here
				brc = FALSE;
			}
				
			pBAEntry->NumOfRxPkt  = 0;			
		}
		
		return brc ;
	}
	else
	{
		return FALSE;
	}
}

/*
Description : Handle subtype = BlockAckRequest control frame (BAR).
BAR frame is used to sync frame sequence in BA session. 
*/
BOOLEAN CntlEnqueueForRecv(
	IN	PRTMP_ADAPTER	pAd, 
	IN ULONG Wcid, 
	IN ULONG MsgLen, 
	IN PFRAME_BA_REQ pMsg) 
{
	PFRAME_BA_REQ	pFrame = pMsg;
	PRTMP_REORDERBUF	pBuffer;
	BA_REC_ENTRY *pBAEntry;
	ULONG	Idx;
	UCHAR	TID, i;
	unsigned long   IrqFlags;
	
	TID = (UCHAR)pFrame->BARControl.TID;
	// Do nothing if the driver is starting halt state.
	// This might happen when timer already been fired before cancel timer with mlmehalt
	if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS))
		return FALSE;

	// First check the size, it MUST not exceed the mlme queue size
	if (MsgLen > MGMT_DMA_BUFFER_SIZE)
	{
		DBGPRINT_ERR(("MlmeEnqueueForRecv: frame too large, size = %ld \n", MsgLen));
		return FALSE;
	}
	else if (MsgLen != sizeof(FRAME_BA_REQ))
	{
		DBGPRINT_ERR(("MlmeEnqueueForRecv: BlockAck Request frame length size = %ld incorrect\n", MsgLen));
		return FALSE;
	}
		
	if ((Wcid < MAX_LEN_OF_MAC_TABLE) && (TID < 8))
	{
		// if this receiving packet is from SA that is in our OriEntry. Since WCID <9 has direct mapping. no need search.
		Idx = pAd->MacTab.Content[Wcid].BARecWcidArray[TID];
		pBAEntry = &pAd->BATable.BARecEntry[Idx];
	}
	else
	{
		return FALSE;
	}
	
	// Then refresh all reorder buffer. and update LastSeq = ReceiveSeq
	if (SEQ_SMALLER(pBAEntry->LastIndSeq, pFrame->BAStartingSeq.field.StartSeq))
	{
		DBGPRINT(RT_DEBUG_TRACE, ("Receive BAR : LastIndSeq  = 0x%x, Starting Seq = 0x%x, [wcid = %ld, TID = %d].  Curindidx = %d \n", pBAEntry->LastIndSeq, pFrame->BAStartingSeq.field.StartSeq, Wcid, TID,pBAEntry->Curindidx));
		NdisAcquireSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
		for ( i=0; i < pBAEntry->BAWinSize; i++)
		{
			if (SEQ_STEPONE (pFrame->BAStartingSeq.field.StartSeq, pBAEntry->LastIndSeq) )
			{
				if (pBAEntry->NumOfRxPkt > 0)
				{
					for ( i=1; i < pBAEntry->BAWinSize ; i++)
					{
						pBuffer = &pAd->LocalRxReorderBuf[pBAEntry->RxBufIdxUsed].MAP_RXBuf[pBAEntry->Curindidx];
						//pBuffer = &pBAEntry->MAP_RXBuf[pBAEntry->Curindidx];
						if (pBuffer->IsFull == TRUE)
						{
							if (pBuffer->IsAMSDU)
							{
								REPORT_AMSDU_FRAMES_TO_LLC(pAd, (PUCHAR)((PUCHAR)(pBuffer->pBuffer)+pBuffer->DataOffset), pBuffer->Datasize);
							}
							else
								REPORT_ETHERNET_FRAME_TO_LLC(pAd, pBuffer->Header802_3, (PUCHAR)((PUCHAR)(pBuffer->pBuffer)+pBuffer->DataOffset), pBuffer->Datasize);
							pBAEntry->LastIndSeq = pBuffer->Sequence.value&MAXSEQ;		
							pBuffer->IsFull = FALSE;
							if (pBAEntry->NumOfRxPkt > 0)
								pBAEntry->NumOfRxPkt-- ;
							DBGPRINT(RT_DEBUG_INFO, ("BAR : 22-%ld. INDI. %dth buffer. pBuffer->Sequence = 0x%x, NumOfRxPkt = %d\n" , Idx,pBAEntry->Curindidx, pBuffer->Sequence.value, pBAEntry->NumOfRxPkt));
							INC_RING_INDEX(pBAEntry->Curindidx, MAX_RX_REORDERBUF);
						}
						else
							break;
					}
				}
				// After we flush the reordering queue, break BAR ;
				break;
			}
			pBuffer = &pAd->LocalRxReorderBuf[pBAEntry->RxBufIdxUsed].MAP_RXBuf[pBAEntry->Curindidx];
			if (pBuffer->IsFull == TRUE)
			{	
				if (pBuffer->IsAMSDU)
				{
					REPORT_AMSDU_FRAMES_TO_LLC(pAd, (PUCHAR)((PUCHAR)(pBuffer->pBuffer)+pBuffer->DataOffset), pBuffer->Datasize);
				}
				else
					REPORT_ETHERNET_FRAME_TO_LLC(pAd, pBuffer->Header802_3, (PUCHAR)((PUCHAR)(pBuffer->pBuffer)+pBuffer->DataOffset), pBuffer->Datasize);
				pBuffer->IsFull = FALSE;
				if (pBAEntry->NumOfRxPkt == 0)
				{	
					DBGPRINT_ERR( ("55 Error Counting Reorder Buffer\n"));
				}
				else
					pBAEntry->NumOfRxPkt--;
				DBGPRINT(RT_DEBUG_TRACE, ("0x%x, ", pBuffer->Sequence.value));
			}
			pBAEntry->LastIndSeq++;	
			pBAEntry->LastIndSeq &= MAXSEQ;
			INC_RING_INDEX(pBAEntry->Curindidx, MAX_RX_REORDERBUF);
		}
		DBGPRINT(RT_DEBUG_TRACE, ("\n"));

		NdisReleaseSpinLock(&pBAEntry->RxReRingLock, IrqFlags);
	}
			
	return TRUE;
}
/*
Description : When there is a Ba policy switch.  (Auto <-> Manual)

Definition :
Auto : After link up, when there is a unicast frame, BA session will be automatically setup.
Manual : BA session won't be setup unless use "ADD button" on RAUI 11n setting page. 
*/
VOID BaAutoManSwitch(
	IN	PRTMP_ADAPTER	pAd) 
{
	UCHAR	i, j;
	MAC_TABLE_ENTRY *pEntry = NULL;
	
	if (pAd->MacTab.Size == 0)
		return;
	
	DBGPRINT(RT_DEBUG_TRACE, ("BaAutoManSwitch====>"));
	// Now switch from Manual to Auto
	if (pAd->CommonCfg.BACapability.field.AutoBA)
	{
		DBGPRINT_RAW(RT_DEBUG_TRACE, (" : Auto\n"));
		for (i = 1; i <MAX_LEN_OF_MAC_TABLE; i++)
		{
			pEntry = &pAd->MacTab.Content[i];
			if (pEntry->ValidAsCLI)
			{
				for (j = 0; j <NUM_OF_TID; j++)
				{
					// BaEntry starts from 1, so  0 means no entry is assigned to this {peer,Tid} yet.
					if (pEntry->BAOriWcidArray[j] == 0)
					{
						BATableInsertEntry(pAd, pEntry->Aid,  0, 0x1, j, pEntry->BaSizeInUse,  Originator_SetAutoAdd, FALSE);
						pEntry->TXAutoBAbitmap |= (1<<j);
					}
				}
			}
		}
	}
	// Now switch from Auto to Manual.
	else
	{
		DBGPRINT_RAW(RT_DEBUG_TRACE, (" : Manual\n"));
		for (i = 0; i <MAX_LEN_OF_BA_ORI_TABLE ; i++)
		{
 			if (pAd->BATable.BAOriEntry[i].ORI_BA_Status != Originator_Done)
			{
				pAd->BATable.BAOriEntry[i].ORI_BA_Status = Originator_NONE;
			}
		}
	}

}
/*
Description : Send PSMP Action frame If PSMP mode switches.
*/
VOID SendPSMPAction(
	IN	PRTMP_ADAPTER	pAd,
	IN    UCHAR  Wcid,
	IN    UCHAR  Psmp) 
{
	PUCHAR			pOutBuffer = NULL;
	NDIS_STATUS 	NStatus;
//	ULONG			Idx;
	FRAME_PSMP_ACTION	Frame;
	ULONG			FrameLen;
	
	NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer);  //Get an unused nonpaged memory
	if(NStatus != NDIS_STATUS_SUCCESS) 
	{
		DBGPRINT(RT_DEBUG_ERROR,("BA - SendPSMPAction() allocate memory failed \n"));
		return;
	}
	ActHeaderInit(pAd, &Frame.Hdr, pAd->MacTab.Content[Wcid].Addr, pAd->CommonCfg.Bssid);
	Frame.Category = CATEGORY_HT;
	Frame.Action = SMPS_ACTION;
	switch (Psmp)
	{
		case MMPS_ENABLE:
			Frame.Psmp = 0;
			break;
		case MMPS_DYNAMIC:
			Frame.Psmp = 3;
			break;
		case MMPS_STATIC:
			Frame.Psmp = 1;
			break;
	}
	MakeOutgoingFrame(pOutBuffer,				&FrameLen,
				  sizeof(FRAME_PSMP_ACTION),	  &Frame,
				  END_OF_ARGS);
	MiniportMMRequest(pAd, pOutBuffer, FrameLen);
	DBGPRINT(RT_DEBUG_ERROR,("HT - SendPSMPAction( %d )  \n", Frame.Psmp));
		
}

