/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************
 
    Module Name:
    soft_ap.c
 
    Abstract:
    Access Point specific routines and MAC table maintenance routines
 
    Revision History:
    Who         When          What
    --------    ----------    ----------------------------------------------
    John Chang  08-04-2003    created for 11g soft-AP

 */

#include "rt_config.h"

char const *pEventText[EVENT_MAX_EVENT_TYPE] = {
	"restart access point",
	"successfully associated",
	"has disassociated",
	"has been aged-out and disassociated" ,    
	"active countermeasures",
	"has disassociated with invalid PSK password"};


/*
	==========================================================================
	Description:
		This routine is used to clean up a specified power-saving queue. It's
		used whenever a wireless client is deleted.
	==========================================================================
 */
VOID APCleanupPsQueue(
	IN  PRTMP_ADAPTER   pAd,
	IN  PQUEUE_HEADER   pQueue)
{
	PQUEUE_ENTRY pEntry;
	PNDIS_PACKET pPacket;

	while (pQueue->Head)
	{
		pEntry = RemoveHeadQueue(pQueue);
		pPacket = QUEUE_ENTRY_TO_PACKET(pEntry);
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
	}
}

VOID  MacTableInitialize(
	IN  PRTMP_ADAPTER   pAd)
{
	UCHAR         i;
	unsigned long   IrqFlags;
	
	NdisAcquireSpinLock(&pAd->MacTabLock, IrqFlags);
	pAd->MacTab.Size = 0;
	for (i=1; i<MAX_LEN_OF_MAC_TABLE; i++)
	{
		pAd->MacTab.Content[i].ValidAsCLI = FALSE;
		pAd->MacTab.Content[i].ValidAsWDS = FALSE;
		pAd->MacTab.Content[i].bIAmBadAtheros = FALSE;
		RTMPZeroMemory(pAd->MacTab.Content[i].Addr, MAC_ADDR_LEN);
	}
	NdisZeroMemory(&pAd->MacTab, sizeof(MAC_TABLE));
	InitializeQueueHeader(&pAd->MacTab.McastPsQueue);
	NdisReleaseSpinLock(&pAd->MacTabLock, IrqFlags);
}
/*
	==========================================================================
	Description:
		This routine reset the entire MAC table. All packets pending in
		the power-saving queues are freed here.
	==========================================================================
 */
VOID MacTableReset(
	IN  PRTMP_ADAPTER  pAd)
{
	UCHAR         i;
	PCHAR       pOutBuffer = NULL;
	NDIS_STATUS NStatus;
	ULONG       FrameLen = 0;
	HEADER_802_11 DisassocHdr;
	USHORT      Reason;
	RT_SET_ASIC_WCID	SetAsicWcid;
	unsigned long   IrqFlags;
	
	NdisAcquireSpinLock(&pAd->MacTabLock, IrqFlags);
	DBGPRINT(RT_DEBUG_TRACE, ("McastPsQueue.Number %ld...\n",pAd->MacTab.McastPsQueue.Number ));
	APCleanupPsQueue(pAd, &pAd->MacTab.McastPsQueue);
	InitializeQueueHeader(&pAd->MacTab.McastPsQueue);
	for (i=1; i<MAX_LEN_OF_MAC_TABLE; i++)
	{
		if (pAd->MacTab.Content[i].ValidAsCLI == TRUE)
		{
			APCleanupPsQueue(pAd, &pAd->MacTab.Content[i].PsQueue);
			DBGPRINT(RT_DEBUG_TRACE, (" %dth PsQueue.Number %ld...\n", i, pAd->MacTab.Content[i].PsQueue.Number ));

			BATableTearRECEntry(pAd, 0, i, TRUE);
			BATableTearORIEntry(pAd, 0, i, TRUE, TRUE);
			pAd->MacTab.Content[i].ValidAsCLI = FALSE;
			// Before reset MacTable, send disassociation packet to client.
			if (pAd->MacTab.Content[i].Sst == SST_ASSOC)
			{
				//  send out a DISASSOC request frame
				NStatus = MlmeAllocateMemory(pAd, (PVOID *)&pOutBuffer);
				if (NStatus != NDIS_STATUS_SUCCESS) 
					return;

				Reason = REASON_DISASSOC_INACTIVE;
				DBGPRINT(RT_DEBUG_ERROR, ("ASSOC - Send DISASSOC  Reason = %d frame  TO %x %x %x %x %x %x \n",Reason,pAd->MacTab.Content[i].Addr[0],
					pAd->MacTab.Content[i].Addr[1],pAd->MacTab.Content[i].Addr[2],pAd->MacTab.Content[i].Addr[3],pAd->MacTab.Content[i].Addr[4],pAd->MacTab.Content[i].Addr[5]));
				MgtMacHeaderInit(pAd, &DisassocHdr, SUBTYPE_DISASSOC, 0, pAd->MacTab.Content[i].Addr, pAd->CommonCfg.Bssid);
				MakeOutgoingFrame(pOutBuffer, &FrameLen, sizeof(HEADER_802_11), &DisassocHdr, 2, &Reason, END_OF_ARGS);
				MiniportMMRequest(pAd, pOutBuffer, FrameLen);

				RTMPusecDelay(2000);
			}
			NdisZeroMemory(pAd->MacTab.Content[i].Addr,6);
			SetAsicWcid.WCID = i;
			SetAsicWcid.SetTid = 0xffffffff;
			SetAsicWcid.DeleteTid = 0xffffffff;
			RTUSBEnqueueInternalCmd(pAd, RT_CMD_SET_ASIC_WCID, &SetAsicWcid, sizeof(RT_SET_ASIC_WCID));
			//AsicDelWcidTab(pAd, i);
		}

	}

	NdisZeroMemory(&pAd->MacTab, sizeof(MAC_TABLE));
	NdisReleaseSpinLock(&pAd->MacTabLock, IrqFlags);
	if (pAd->MacTab.Size == 0)
	{
		pAd->CommonCfg.AddHTInfo.AddHtInfo2.OperaionMode = 0;
	}
	return;

}
/*
================================================================
Description : because WDS and CLI share the same WCID table in ASIC. 
WDS entry also insert to pAd->MacTab.content[].  Such is marked as ValidAsWDS = TRUE.
Also fills the pairwise key.
Because front MAX_AID_BA entries have direct mapping to BAEntry, which is only used as CLI. So we insert WDS
from index MAX_AID_BA.
================================================================
*/
MAC_TABLE_ENTRY *MacTableInsertWDSEntry(
	IN  PRTMP_ADAPTER   pAd, 
	IN  PUCHAR pAddr,
	IN  UINT  WdsTabIdx) 
{
	int i;
	MAC_TABLE_ENTRY *pEntry = NULL;
	unsigned long   IrqFlags;
		
	// if FULL, return
	if (pAd->MacTab.Size >= MAX_LEN_OF_MAC_TABLE) 
		return NULL;

	// allocate one MAC entry
	NdisAcquireSpinLock(&pAd->MacTabLock, IrqFlags);
	// Put WDS to index>=MAX_AID_BA. because front MAX_AID_BA MACEntry have direct mapping to BAEntry.
	for (i = (1); i< MAX_LEN_OF_MAC_TABLE; i++)   // skip entry#0 so that "entry index == AID" for fast lookup
	{
		// pick up the first available vacancy
		if ((pAd->MacTab.Content[i].ValidAsCLI == FALSE) && (pAd->MacTab.Content[i].ValidAsWDS == FALSE))
		{
			pEntry = &pAd->MacTab.Content[i];
			NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
			pEntry->ValidAsWDS = TRUE;
			pEntry->ValidAsCLI = FALSE;

			pEntry->pAd = pAd;
			pEntry->AuthMode = pAd->CommonCfg.AuthMode;
			pEntry->WepStatus = pAd->CommonCfg.WepStatus;

			if (pEntry->AuthMode < Ndis802_11AuthModeWPA)
				pEntry->WpaState = AS_NOTUSE;
			else
				pEntry->WpaState = AS_INITIALIZE;
			pEntry->PairwiseKey.KeyLen = 0;
			pEntry->PairwiseKey.CipherAlg = CIPHER_NONE;
			pEntry->PortSecured = WPA_802_1X_PORT_NOT_SECURED;
			AsicRemovePairwiseKeyEntry(pAd, (UCHAR)i);
			COPY_MAC_ADDR(pEntry->Addr, pAddr);
			pEntry->Sst = SST_NOT_AUTH;
			pEntry->AuthState = AS_NOT_AUTH;
			pEntry->Aid = (USHORT)i; 
			pEntry->CapabilityInfo = 0;
			pEntry->PsMode = PWR_ACTIVE;
			pEntry->MaxSupportedRate = RATE_11;
			pEntry->CurrTxRate = RATE_11;
			pEntry->PsQIdleCount = 0;
			pEntry->NoDataIdleCount = 0;
			InitializeQueueHeader(&pEntry->PsQueue);
			pAd->MacTab.Size ++;
			pAd->WdsTab.MacTabMatchWCID[WdsTabIdx] = (UCHAR)i;
			pEntry->MatchWDSTabIdx = WdsTabIdx;
			// be sure WdsTab.Wpa_key is right value.
			AsicAddPairwiseKeyEntry(pAd, 
									pAddr, 
									(UCHAR)i, 
									&pAd->WdsTab.Wpa_key);

			DBGPRINT(RT_DEBUG_TRACE, ("MacTableInsertWDSEntry - allocate entry #%d, Total= %d\n",i, pAd->MacTab.Size));
			break;
		}
	}

	// don't need to add this MAC entry into HASH table

	NdisReleaseSpinLock(&pAd->MacTabLock, IrqFlags);
	return pEntry;
}

/*
	==========================================================================
	Description:
		Add and new entry into MAC table
	==========================================================================
 */
MAC_TABLE_ENTRY *MacTableInsertEntry(
	IN  PRTMP_ADAPTER   pAd, 
	IN  PUCHAR pAddr,
	IN BOOLEAN	CleanAll) 
{
	UCHAR HashIdx;
	int i;
	MAC_TABLE_ENTRY *pEntry = NULL, *pCurrEntry;
	unsigned long   IrqFlags;
	// if FULL, return
	if (pAd->MacTab.Size >= MAX_LEN_OF_MAC_TABLE) 
		return NULL;

	// allocate one MAC entry
	NdisAcquireSpinLock(&pAd->MacTabLock, IrqFlags);
	for (i = 1; i< MAX_LEN_OF_MAC_TABLE; i++)   // skip entry#0 so that "entry index == AID" for fast lookup
	{
		// pick up the first available vacancy
		if ((pAd->MacTab.Content[i].ValidAsCLI == FALSE) && (pAd->MacTab.Content[i].ValidAsWDS == FALSE))
		{
			pEntry = &pAd->MacTab.Content[i];
			if (CleanAll == TRUE)
				NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
			pEntry->ValidAsCLI = TRUE;
			pEntry->ValidAsWDS = FALSE;

			pEntry->pAd = pAd;
			pEntry->CMTimerRunning = FALSE;
			pEntry->RSNIE_Len = 0;
			NdisZeroMemory(pEntry->R_Counter, sizeof(pEntry->R_Counter));
			pEntry->AuthMode = pAd->CommonCfg.AuthMode;
			pEntry->WepStatus = pAd->CommonCfg.WepStatus;

			if (pEntry->AuthMode < Ndis802_11AuthModeWPA)
				pEntry->WpaState = AS_NOTUSE;
			else
				pEntry->WpaState = AS_INITIALIZE;
			pEntry->GTKState = REKEY_NEGOTIATING;
			pEntry->PrivacyFilter = Ndis802_11PrivFilterAcceptAll;
			pEntry->PairwiseKey.KeyLen = 0;
			pEntry->PairwiseKey.CipherAlg = CIPHER_NONE;
			pEntry->PortSecured = WPA_802_1X_PORT_NOT_SECURED;
			AsicRemovePairwiseKeyEntry(pAd, (UCHAR)i);
			COPY_MAC_ADDR(pEntry->Addr, pAddr);
			pEntry->Sst = SST_NOT_AUTH;
			pEntry->AuthState = AS_NOT_AUTH;
			pEntry->Aid = (USHORT)i;  //0;
			pEntry->CapabilityInfo = 0;
			pEntry->PsMode = PWR_ACTIVE;
			pEntry->MaxSupportedRate = RATE_11;
			pEntry->CurrTxRate = RATE_11;
			pEntry->PsQIdleCount = 0;
			pEntry->NoDataIdleCount = 0;
			InitializeQueueHeader(&pEntry->PsQueue);
			pAd->MacTab.Size ++;
			// Add this pAddr into ASIC WCID Table Outside This function.
			// Because Spin Lock will raise to > PASSIVE_LEVEL, so that VenderRequest runs only at PASSIVE_LEVEL.
			RTUSBEnqueueInternalCmd(pAd, RT_OID_SET_CLIENT_MAC_ENTRY, pEntry, sizeof(MAC_TABLE_ENTRY));
			break;
		}
	}

	// add this MAC entry into HASH table
	if (pEntry)
	{
		HashIdx = MAC_ADDR_HASH_INDEX(pAddr);
		if (pAd->MacTab.Hash[HashIdx] == NULL)
		{
			pAd->MacTab.Hash[HashIdx] = pEntry;
		}
		else
		{
			pCurrEntry = pAd->MacTab.Hash[HashIdx];
			while (pCurrEntry->pNext != NULL)
				pCurrEntry = pCurrEntry->pNext;
			pCurrEntry->pNext = pEntry;
		}
	}

	NdisReleaseSpinLock(&pAd->MacTabLock, IrqFlags);
	return pEntry;
}
/*
	==========================================================================
	Description:
		Delete all WDS Entry in pAd->MacTab
	==========================================================================
 */
BOOLEAN MacTableDeleteWDSEntry(
	IN PRTMP_ADAPTER pAd) 
{
//	MAC_TABLE_ENTRY *pEntry, *pPrevEntry;
	int i;
	unsigned long   IrqFlags;
	
	NdisAcquireSpinLock(&pAd->MacTabLock, IrqFlags);
	for (i = 1; i< MAX_LEN_OF_MAC_TABLE; i++)   // skip entry#0 so that "entry index == AID" for fast lookup
	{
		if (pAd->MacTab.Content[i].ValidAsWDS == TRUE)
		{
			NdisZeroMemory(&pAd->MacTab.Content[i], sizeof(MAC_TABLE_ENTRY));
		}
	}
	NdisReleaseSpinLock(&pAd->MacTabLock, IrqFlags);

	return TRUE;
}

/*
	==========================================================================
	Description:
		Delete a specified client from MAC table
	==========================================================================
 */
BOOLEAN MacTableDeleteEntry(
	IN PRTMP_ADAPTER pAd, 
	IN USHORT wcid,
	IN PUCHAR pAddr) 
{
	USHORT HashIdx;
	MAC_TABLE_ENTRY *pEntry, *pPrevEntry;
	UCHAR	delWcid=0xff;
	unsigned long   IrqFlags;
	
	RT_SET_ASIC_WCID	SetAsicWcid;

	if (wcid >= MAX_LEN_OF_MAC_TABLE)
		return FALSE;

	HashIdx = MAC_ADDR_HASH_INDEX(pAddr);
	//pEntry = pAd->MacTab.Hash[HashIdx];
	pEntry = &pAd->MacTab.Content[wcid];

	if (pEntry)
	{
		NdisAcquireSpinLock(&pAd->MacTabLock, IrqFlags);
		if (MAC_ADDR_EQUAL(pEntry->Addr, pAddr))
		{
			delWcid = (UCHAR)wcid;
			// Delete this pAddr from ASIC on-chip WCID Table
			BATableTearRECEntry(pAd, 0, (UCHAR)pEntry->Aid, TRUE);
			BATableTearORIEntry(pAd, 0, (UCHAR)pEntry->Aid, TRUE,TRUE);
			pAd->MacTab.Hash[HashIdx] = pEntry->pNext;
			APCleanupPsQueue(pAd, &pEntry->PsQueue); // return all NDIS packet in PSQ
			NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
			pAd->MacTab.Size --;
			DBGPRINT(RT_DEBUG_TRACE, ("MacTableDeleteEntry1 - Total= %d\n", pAd->MacTab.Size));
		}
		else
		{
			while (pEntry->pNext)
			{
				pPrevEntry = pEntry;
				pEntry = pEntry->pNext;
				if (MAC_ADDR_EQUAL(pEntry->Addr, pAddr))
				{
					delWcid = (UCHAR)pEntry->Aid;
					BATableTearRECEntry(pAd, 0, (UCHAR)pEntry->Aid, TRUE);
					BATableTearORIEntry(pAd, 0, (UCHAR)pEntry->Aid, TRUE, TRUE);
					pPrevEntry->pNext = pEntry->pNext;
					APCleanupPsQueue(pAd, &pEntry->PsQueue); // return all NDIS packet in PSQ
					NdisZeroMemory(pEntry, sizeof(MAC_TABLE_ENTRY));
					pAd->MacTab.Size --;
					DBGPRINT(RT_DEBUG_TRACE, ("MacTableDeleteEntry2 - Total= %d\n", pAd->MacTab.Size));
					break;
				}
			}
		}
		NdisReleaseSpinLock(&pAd->MacTabLock, IrqFlags);

		SetAsicWcid.WCID = delWcid;
		SetAsicWcid.SetTid = 0xffffffff;
		SetAsicWcid.DeleteTid = 0xffffffff;
		RTUSBEnqueueInternalCmd(pAd, RT_CMD_SET_ASIC_WCID, &SetAsicWcid, sizeof(RT_SET_ASIC_WCID));
		DBGPRINT(RT_DEBUG_TRACE,("Aid= 0x%x.SetTid= 0x%lx. \n",  pEntry->Aid, SetAsicWcid.SetTid));

	}

	if (pAd->MacTab.Size == 0)
	{
		pAd->CommonCfg.AddHTInfo.AddHtInfo2.OperaionMode = 0;
	}
	return TRUE;
}

/*
	==========================================================================
	Description:
		Look up the MAC address in the MAC table. Return NULL if not found.
	Return:
		pEntry - pointer to the MAC entry; NULL is not found
	==========================================================================
*/
MAC_TABLE_ENTRY *MacTableLookup(
	IN PRTMP_ADAPTER pAd, 
	PUCHAR pAddr) 
{
	ULONG HashIdx;
	MAC_TABLE_ENTRY *pEntry;

	HashIdx = MAC_ADDR_HASH_INDEX(pAddr);
	pEntry = pAd->MacTab.Hash[HashIdx];

	while (pEntry && (pEntry->ValidAsCLI ||pEntry->ValidAsWDS))
	{
		if (MAC_ADDR_EQUAL(pEntry->Addr, pAddr)) 
		{
			break;
		}
		else
			pEntry = pEntry->pNext;
	}
	return pEntry;
}

/*
	==========================================================================
	Description:
		This routine is called by APMlmePeriodicExec() every second to check if
		1. any associated client in PSM. If yes, then TX MCAST/BCAST should be
		   out in DTIM only
		2. any client being idle for too long and should be aged-out from MAC table
		3. garbage collect PSQ
	==========================================================================
*/
VOID MacTableMaintenance(
	IN PRTMP_ADAPTER pAd)
{
	int i;
	unsigned long   IrqFlags;
	
	pAd->MacTab.fAnyStationInPsm = FALSE;
	for (i = 1; i < MAX_LEN_OF_MAC_TABLE; i++) 
	{
		MAC_TABLE_ENTRY *pEntry = &pAd->MacTab.Content[i];

		if (pEntry->ValidAsCLI == FALSE)
			continue;

		pEntry->NoDataIdleCount ++;  

		// 0. STA failed to complete association should be removed to save MAC table space.
		if ((pEntry->Sst != SST_ASSOC) && (pEntry->NoDataIdleCount >= MAC_TABLE_ASSOC_TIMEOUT))
		{
			DBGPRINT(RT_DEBUG_TRACE, ("%02x:%02x:%02x:%02x:%02x:%02x fail to complete ASSOC in %d sec\n",
					pEntry->Addr[0],pEntry->Addr[1],pEntry->Addr[2],pEntry->Addr[3],
					pEntry->Addr[4],pEntry->Addr[5],MAC_TABLE_ASSOC_TIMEOUT));
			MacTableDeleteEntry(pAd, (USHORT)i, pEntry->Addr);
			continue;
		}

		// 1. check if there's any associated STA in power-save mode. this affects outgoing
		//    MCAST/BCAST frames should be stored in PSQ till DtimCount=0
		if (pEntry->PsMode == PWR_SAVE)
			pAd->MacTab.fAnyStationInPsm = TRUE;

		// 2. delete those MAC entry that has been idle for a long time
		if (pEntry->NoDataIdleCount >= (MAC_TABLE_AGEOUT_TIME/2))
		{
			DBGPRINT(RT_DEBUG_TRACE, ("ageout %02x:%02x:%02x:%02x:%02x:%02x after %d-sec silence\n",
					pEntry->Addr[0],pEntry->Addr[1],pEntry->Addr[2],pEntry->Addr[3],
					pEntry->Addr[4],pEntry->Addr[5],MAC_TABLE_AGEOUT_TIME));
			MacTableDeleteEntry(pAd, pEntry->Aid, pEntry->Addr);
		}
		if (pEntry->NoDataIdleCount >= MAC_TABLE_AGEOUT_TIME)
		{
			DBGPRINT(RT_DEBUG_TRACE, ("ageout %02x:%02x:%02x:%02x:%02x:%02x after %d-sec silence\n",
					pEntry->Addr[0],pEntry->Addr[1],pEntry->Addr[2],pEntry->Addr[3],
					pEntry->Addr[4],pEntry->Addr[5],MAC_TABLE_AGEOUT_TIME));
			MacTableDeleteEntry(pAd, pEntry->Aid, pEntry->Addr);
			continue;
		}

		// 3. garbage collect the PsQueue if the STA has being idle for a while
		if (pEntry->PsQueue.Head)
		{
			pEntry->PsQIdleCount ++;  
			if (pEntry->PsQIdleCount > 2) 
			{
				NdisAcquireSpinLock(&pAd->MacTabLock, IrqFlags);
				APCleanupPsQueue(pAd, &pEntry->PsQueue);
				NdisReleaseSpinLock(&pAd->MacTabLock, IrqFlags);
				pEntry->PsQIdleCount = 0;
			}
		}
		else
			pEntry->PsQIdleCount = 0;
	}

	// 4. garbage collect pAd->MacTab.McastPsQueue if backlogged MCAST/BCAST frames
	//    stale in queue. Since MCAST/BCAST frames always been sent out whenever 
	//    DtimCount==0, the only case to let them stale is surprise removal of the NIC,
	//    so that ASIC-based Tbcn interrupt stops and DtimCount dead.
	if (pAd->MacTab.McastPsQueue.Head)
	{
		pAd->MacTab.PsQIdleCount ++;
		if (pAd->MacTab.PsQIdleCount > 1)
		{
			NdisAcquireSpinLock(&pAd->MacTabLock, IrqFlags);
			APCleanupPsQueue(pAd, &pAd->MacTab.McastPsQueue);
			NdisReleaseSpinLock(&pAd->MacTabLock, IrqFlags);
			pAd->MacTab.PsQIdleCount = 0;
		}
	}
	else
		pAd->MacTab.PsQIdleCount = 0;
}
/*
Description: Insert CLI that connects to one of our WDS-AP[index].
*/
WDS_ENTRY *WdsTableInsertEntry(
	IN  PRTMP_ADAPTER   pAd, 
	IN  PUCHAR          pAddr,
	IN  ULONG           index) 
{
	UCHAR HashIdx;
	int i;
	WDS_ENTRY *pEntry = NULL, *pCurrEntry;
	unsigned long   IrqFlags;
	
	if (pAd->WdsTab.MacTab[index].Size >= MAX_LEN_OF_MAC_TABLE) 
		return NULL;

	// allocate one MAC entry
	NdisAcquireSpinLock(&pAd->WdsTabLock, IrqFlags);
	for (i = 0; i< MAX_LEN_OF_MAC_TABLE; i++)
	{
		// pick up the first available vacancy
		if (pAd->WdsTab.MacTab[index].Content[i].Valid == FALSE)
		{
			pEntry = &pAd->WdsTab.MacTab[index].Content[i];
			NdisZeroMemory(pEntry, sizeof(WDS_ENTRY));
			pEntry->Valid = TRUE;
			COPY_MAC_ADDR(pEntry->Addr, pAddr);
			pEntry->NoDataIdleCount = 0;
			pAd->WdsTab.MacTab[index].Size ++;
			DBGPRINT(RT_DEBUG_TRACE, ("WdsTableInsertEntry - allocate entry #%d, Total= %d\n",i, pAd->WdsTab.MacTab[index].Size));
			break;
		}
	}

	// add this MAC entry into HASH table
	if (pEntry)
	{
		HashIdx = MAC_ADDR_HASH_INDEX(pAddr);
		if (pAd->WdsTab.MacTab[index].Hash[HashIdx] == NULL)
		{
			pAd->WdsTab.MacTab[index].Hash[HashIdx] = pEntry;
		}
		else
		{
			pCurrEntry = pAd->WdsTab.MacTab[index].Hash[HashIdx];
			while (pCurrEntry->pNext != NULL)
				pCurrEntry = pCurrEntry->pNext;
			pCurrEntry->pNext = pEntry;
		}
		DBGPRINT(RT_DEBUG_TRACE, ("WdsTableInsertEntry - allocate entry #%d, Total= %d\n",i, pAd->WdsTab.MacTab[index].Size));
	}

	// Copy to MACTAB.CONTENT to. Set ValidAsWDS = TRUE
	NdisReleaseSpinLock(&pAd->WdsTabLock, IrqFlags);
	return pEntry;
}

BOOLEAN WdsTableDeleteEntry(
	IN PRTMP_ADAPTER pAd, 
	IN PUCHAR        pAddr,
	IN ULONG         index) 
{
	USHORT HashIdx;
	WDS_ENTRY *pEntry, *pPrevEntry;
	unsigned long   IrqFlags;
	
	NdisAcquireSpinLock(&pAd->WdsTabLock, IrqFlags);
	HashIdx = MAC_ADDR_HASH_INDEX(pAddr);
	pEntry = pAd->WdsTab.MacTab[index].Hash[HashIdx];
	if (pEntry)
	{
		if (MAC_ADDR_EQUAL(pEntry->Addr, pAddr))
		{
			pAd->WdsTab.MacTab[index].Hash[HashIdx] = pEntry->pNext;
			NdisZeroMemory(pEntry, sizeof(WDS_ENTRY));
			pAd->WdsTab.MacTab[index].Size --;
			DBGPRINT(RT_DEBUG_TRACE, ("WdsTableDeleteEntry - Total= %d\n", pAd->WdsTab.MacTab[index].Size));
		}
		else
		{
			while (pEntry->pNext)
			{
				pPrevEntry = pEntry;
				pEntry = pEntry->pNext;
				if (MAC_ADDR_EQUAL(pEntry->Addr, pAddr))
				{
					pPrevEntry->pNext = pEntry->pNext;
					NdisZeroMemory(pEntry, sizeof(WDS_ENTRY));
					pAd->WdsTab.MacTab[index].Size --;
					DBGPRINT(RT_DEBUG_TRACE, ("WdsTableDeleteEntry - Total= %d\n", pAd->WdsTab.MacTab[index].Size));
					break;
				}
			}
		}
	}

	NdisReleaseSpinLock(&pAd->WdsTabLock, IrqFlags);
	return TRUE;
}

WDS_ENTRY *WdsTableLookup(
	IN PRTMP_ADAPTER pAd, 
	IN PUCHAR        pAddr,
	IN ULONG         index) 
{
	ULONG HashIdx;
	WDS_ENTRY *pEntry;

	HashIdx = MAC_ADDR_HASH_INDEX(pAddr);
	pEntry = pAd->WdsTab.MacTab[index].Hash[HashIdx];

	while (pEntry && (pEntry->Valid))
	{
		if (MAC_ADDR_EQUAL(pEntry->Addr, pAddr)) 
		{
			pEntry->NoDataIdleCount = 0;
			break;
		}
		else
			pEntry = pEntry->pNext;
	}
	return pEntry;
}

/*
	==========================================================================
	Description:
		This routine is called by APMlmePeriodicExec() every second to check if
		1. any associated client in PSM. If yes, then TX MCAST/BCAST should be
		   out in DTIM only
		2. any client being idle for too long and should be aged-out from MAC table
		3. garbage collect PSQ
	==========================================================================
*/
VOID WdsTableMaintenance(
	IN PRTMP_ADAPTER pAd)
{
	ULONG   i, idx;
	for (idx = 0; idx < pAd->WdsTab.Size; idx++)
	{
		for (i = 0; i < MAX_LEN_OF_MAC_TABLE; i++) 
		{
			WDS_ENTRY *pEntry = &pAd->WdsTab.MacTab[idx].Content[i];

			if (pEntry->Valid == FALSE)
				continue;

			pEntry->NoDataIdleCount += 5;  

			// 2. delete those MAC entry that has been idle for a long time
			if (pEntry->NoDataIdleCount >= MAC_TABLE_AGEOUT_TIME)
			{
				DBGPRINT(RT_DEBUG_TRACE, ("ageout %02x:%02x:%02x:%02x:%02x:%02x from WDS #%ld after %d-sec silence\n",
						pEntry->Addr[0],pEntry->Addr[1],pEntry->Addr[2],pEntry->Addr[3],
						pEntry->Addr[4],pEntry->Addr[5], idx, MAC_TABLE_AGEOUT_TIME));
				WdsTableDeleteEntry(pAd, pEntry->Addr, idx);
			}
		}
	}
}

