
#ifndef __RTUSB_H__
#define __RTUSB_H__

//usb header files
#include <linux/usb.h>

#define STATUS_SUCCESS				0x00
#define STATUS_UNSUCCESSFUL 		0x01
#define NT_SUCCESS(status)			(((status) > 0) ? (1):(0))
#define InterlockedIncrement 	 	atomic_inc
#define NdisInterlockedIncrement 	atomic_inc
#define InterlockedDecrement		atomic_dec
#define NdisInterlockedDecrement 	atomic_dec
#define InterlockedExchange			atomic_set
#define NdisMSendComplete			RTMP_SendComplete
#define NdisMCancelTimer			RTMPCancelTimer
#define NdisFreeMemory(a, b, c) 	kfree((a))
#define NdisMSleep					RTMPusecDelay		/* unit: microsecond */
#define NdisGetCurrentSystemTime	RTMP_GetCurrentSystemTime

#ifndef USB_ST_NOERROR
#define USB_ST_NOERROR     0
#endif

#define USBD_TRANSFER_DIRECTION_OUT		0
#define USBD_TRANSFER_DIRECTION_IN		0
#define USBD_SHORT_TRANSFER_OK			0
#define PURB			purbb_t

#define RTUSB_FREE_URB(pUrb)	usb_free_urb(pUrb)

#undef MlmeAllocateMemory
#undef MlmeFreeMemory

typedef int				NTSTATUS;
typedef struct usb_device	* PUSB_DEV;

/* MACRO for linux usb */
typedef struct urb *purbb_t;
typedef struct usb_ctrlrequest devctrlrequest;
#define PIRP		PVOID
#define PMDL		PVOID
#define NDIS_OID	UINT
#ifndef USB_ST_NOERROR
#define  USB_ST_NOERROR     0
#endif

// vendor-specific control operations
#define CONTROL_TIMEOUT_JIFFIES (0.1 * HZ)
#define UNLINK_TIMEOUT_MS		3

#define USBD_TRANSFER_DIRECTION_OUT		0
#define USBD_TRANSFER_DIRECTION_IN		0
#define USBD_SHORT_TRANSFER_OK			0


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
/* allocate urb */
#define RTUSB_ALLOC_URB(iso)	usb_alloc_urb(iso, GFP_ATOMIC)
/* submit urb	*/	
#define RTUSB_SUBMIT_URB(pUrb)	usb_submit_urb(pUrb, GFP_ATOMIC)
/* allocate dma-consistent buffer	*/
#define	RTUSB_URB_ALLOC_BUFFER(pUsb_Dev, BufSize, pDma_addr)	\
	usb_buffer_alloc(pUsb_Dev, BufSize, GFP_ATOMIC, pDma_addr)
/* free dma buffer	*/
#define	RTUSB_URB_FREE_BUFFER(pUsb_Dev, BufSize, pTransferBuf, Dma_addr)	\
	usb_buffer_free(pUsb_Dev, BufSize, pTransferBuf, Dma_addr)
	
#else
/* allocate urb */
#define RTUSB_ALLOC_URB(iso)	usb_alloc_urb(iso)
/* submit urb	*/
#define RTUSB_SUBMIT_URB(pUrb)	usb_submit_urb(pUrb)
/* allocate dma-consistent buffer	*/
#define RTUSB_URB_ALLOC_BUFFER(pUsb_Dev, BufSize, pDma_addr)	\
	kmalloc(BufSize, GFP_ATOMIC)
/* free dma buffer	*/	
#define	RTUSB_URB_FREE_BUFFER(pUsb_Dev, BufSize, pTransferBuf, Dma_addr)	\
	kfree(pTransferBuf)	

#endif

/* unlink urb	*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,7)
#define RTUSB_UNLINK_URB(pUrb)		usb_kill_urb(pUrb)
#else
#define RTUSB_UNLINK_URB(pUrb)		usb_unlink_urb(pUrb)
#endif

/* free urb		*/
#define RTUSB_FREE_URB(pUrb)	usb_free_urb(pUrb)


// Prototypes of completion funuc.
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#define RTUSBBulkOutDataPacketComplete(purb, pt_regs)    RTUSBBulkOutDataPacketComplete(purb)
#define RTUSBBulkOutNullFrameComplete(pUrb, pt_regs)     RTUSBBulkOutNullFrameComplete(pUrb)
#define RTUSBBulkOutRTSFrameComplete(pUrb, pt_regs)      RTUSBBulkOutRTSFrameComplete(pUrb)
#define RTUSBBulkOutMLMEPacketComplete(pUrb, pt_regs)    RTUSBBulkOutMLMEPacketComplete(pUrb)
#define RTUSBBulkOutPsPollComplete(pUrb, pt_regs)        RTUSBBulkOutPsPollComplete(pUrb)
#define RTUSBBulkRxComplete(pUrb, pt_regs)               RTUSBBulkRxComplete(pUrb)
#endif
VOID RTUSBBulkOutDataPacketComplete(purbb_t purb, struct pt_regs *pt_regs);
VOID RTUSBBulkOutNullFrameComplete(purbb_t pUrb, struct pt_regs *pt_regs);
VOID RTUSBBulkOutRTSFrameComplete(purbb_t pUrb, struct pt_regs *pt_regs);
VOID RTUSBBulkOutMLMEPacketComplete(purbb_t pUrb, struct pt_regs *pt_regs);
VOID RTUSBBulkOutPsPollComplete(purbb_t pUrb, struct pt_regs *pt_regs);
VOID RTUSBBulkRxComplete(purbb_t pUrb, struct pt_regs *pt_regs);


#define RTUSBMlmeUp(pAd)	        \
{								    \
	POS_COOKIE pObj = (POS_COOKIE) pAd->OS_Cookie;	\
	if(pObj->MLMEThr_pid>0)		    \
        up(&(pAd->mlme_semaphore)); \
}

#define RTUSBCMDUp(pAd)	                \
{									    \
	POS_COOKIE pObj = (POS_COOKIE) pAd->OS_Cookie;	\
	if(pObj->RTUSBCmdThr_pid>0)		    \
	    up(&(pAd->RTUSBCmd_semaphore)); \
}


static inline NDIS_STATUS RTMPAllocateMemory(
	OUT PVOID *ptr,
	IN size_t size)
{
	*ptr = kmalloc(size, GFP_ATOMIC);
	if(*ptr)
		return NDIS_STATUS_SUCCESS;
	else
		return NDIS_STATUS_RESOURCES;
}

// 
// For WPA SUPPLICANT: WIRELESS EXT support wireless events: v14 or newer
//
#if WIRELESS_EXT >= 14
#define WPA_SUPPLICANT_SUPPORT  1
#else
#define WPA_SUPPLICANT_SUPPORT  0
#endif

/* oid.h */
// Cipher suite type for mixed mode group cipher, P802.11i-2004
typedef enum _RT_802_11_CIPHER_SUITE_TYPE {
	Cipher_Type_NONE,
	Cipher_Type_WEP40,
	Cipher_Type_TKIP,
	Cipher_Type_RSVD,
	Cipher_Type_CCMP,
	Cipher_Type_WEP104
} RT_802_11_CIPHER_SUITE_TYPE, *PRT_802_11_CIPHER_SUITE_TYPE;

//RT_OID_MULTI_READ_MAC
//RT_OID_MULTI_WRITE_MAC
//RT_OID_USB_VENDOR_EEPROM_READ
//RT_OID_USB_VENDOR_EEPROM_WRITE
typedef	struct	_CMDHandler_TLV	{
	USHORT		Offset;
	USHORT		Length;
	UCHAR		DataFirst;
}	CMDHandler_TLV, *PCMDHandler_TLV;

// New for MeetingHouse Api support
#define RT_OID_USB_VENDOR_RESET						0x0D730101	// cmd
#define RT_OID_USB_VENDOR_UNPLUG					0x0D730102	// cmd
#define RT_OID_USB_VENDOR_SWITCH_FUNCTION			0x0D730103	// cmd
#define RT_OID_MULTI_WRITE_MAC						0x0D730107	// cmd
#define RT_OID_MULTI_READ_MAC						0x0D730108	// cmd
#define RT_OID_USB_VENDOR_EEPROM_WRITE				0x0D73010A	// cmd
#define RT_OID_USB_VENDOR_EEPROM_READ				0x0D73010B	// cmd
#define RT_OID_USB_VENDOR_ENTER_TESTMODE			0x0D73010C	// cmd
#define RT_OID_USB_VENDOR_EXIT_TESTMODE				0x0D73010D	// cmd
#define RT_OID_VENDOR_WRITE_BBP						0x0D730119	// cmd
#define RT_OID_VENDOR_READ_BBP						0x0D730118	// cmd
#define RT_OID_VENDOR_WRITE_RF						0x0D73011A	// cmd
#define RT_OID_VENDOR_FLIP_IQ						0x0D73011D	// cmd
#define RT_OID_USB_RESET_BULK_OUT					0x0D730210	// cmd
#define RT_OID_USB_RESET_BULK_IN					0x0D730211	// cmd
#define RT_OID_SET_PSM_BIT_SAVE						0x0D730212	// cmd
#define RT_OID_SET_RADIO							0x0D730214	// cmd
#define RT_OID_UPDATE_TX_RATE						0x0D730216	// cmd
#define OID_802_11_ADD_KEY_WEP						0x0D730218	// cmd
#define RT_OID_RESET_FROM_ERROR						0x0D73021A	// cmd
#define RT_OID_LINK_DOWN							0x0D73021B	// cmd
#define RT_OID_RESET_FROM_NDIS						0x0D73021C	// cmd
#define RT_OID_PERIODIC_EXECUT						0x0D73021D	// cmd
#define RT_OID_CHECK_GPIO							0x0D730215	// cmd
#define RT_PERFORM_SOFT_DIVERSITY					0x0D730221	// cmd
#define RT_OID_FORCE_WAKE_UP						0x0D730222	// cmd
#define RT_OID_SET_PSM_BIT_ACTIVE					0x0D730223	// cmd
#define RT_CMD_RESET_MLME							0x0D730224	// cmd
#define RT_CMD_SET_BW								0x0D730225	// cmd
#define RT_CMD_SET_ASIC_WCID						0x0D730226	// cmd
#define RT_CMD_SET_ASIC_WCID_CIPHER					0x0D730227	// cmd
#define RT_OID_QKERIODIC_EXECUT						0x0D73023D	// cmd
#define RT_OID_SET_CLIENT_MAC_ENTRY					0x0D73023E	// cmd
#define RT_OID_802_11_QUERY_HARDWARE_REGISTER		0x0D710105	// cmd
#define RT_OID_802_11_SET_PHY_MODE					0x0D79010C	// cmd
#define RT_OID_802_11_SET_STA_CONFIG				0x0D790111	// cmd
#define RT_OID_802_11_SET_PREAMBLE					0x0D790101	// cmd

#define WPA1AKMBIT	    0x01
#define WPA2AKMBIT	    0x02
#define WPA1PSKAKMBIT   0x04
#define WPA2PSKAKMBIT   0x08
#define TKIPBIT         0x01
#define CCMPBIT         0x02

#endif //__RTUSB_H__
