#include "rt_config.h"


static void HTParametersHook(
	IN	PRTMP_ADAPTER pAd, 
	IN	CHAR		  *pValueStr,
	IN	CHAR		  *pInput);


//add by kathy

 /**
  * strstr - Find the first substring in a %NUL terminated string
  * @s1: The string to be searched
  * @s2: The string to search for
  */
char * rtstrstr(const char * s1,const char * s2)
{
	INT l1, l2;

	l2 = strlen(s2);
	if (!l2)
		return (char *) s1;
	
	l1 = strlen(s1);
	
	while (l1 >= l2)
	{
		l1--;
		if (!memcmp(s1,s2,l2))
			return (char *) s1;
		s1++;
	}
	
	return NULL;
}
 
/**
 * rstrtok - Split a string into tokens
 * @s: The string to be searched
 * @ct: The characters to search for
 * * WARNING: strtok is deprecated, use strsep instead. However strsep is not compatible with old architecture.
 */
char * __rstrtok;
char * rstrtok(char * s,const char * ct)
{
	char *sbegin, *send;

	sbegin  = s ? s : __rstrtok;
	if (!sbegin)
	{
		return NULL;
	}

	sbegin += strspn(sbegin,ct);
	if (*sbegin == '\0')
	{
		__rstrtok = NULL;
		return( NULL );
	}

	send = strpbrk( sbegin, ct);
	if (send && *send != '\0')
		*send++ = '\0';

	__rstrtok = send;

	return (sbegin);
}

/*
    ========================================================================

    Routine Description:
        Find key section for Get key parameter.

    Arguments:
        buffer                      Pointer to the buffer to start find the key section
        section                     the key of the secion to be find

    Return Value:
        NULL                        Fail
        Others                      Success
    ========================================================================
*/
PUCHAR  RTMPFindSection(
    IN  PCHAR   buffer)
{
    CHAR temp_buf[32];
    PUCHAR  ptr;

    strcpy(temp_buf, "Default");

    if((ptr = rtstrstr(buffer, temp_buf)) != NULL)
            return (ptr+strlen("\n"));
        else
            return NULL;
}

/*
    ========================================================================

    Routine Description:
        Get key parameter.

    Arguments:
        key                         Pointer to key string
        dest                        Pointer to destination      
        destsize                    The datasize of the destination
        buffer                      Pointer to the buffer to start find the key

    Return Value:
        TRUE                        Success
        FALSE                       Fail

    Note:
        This routine get the value with the matched key (case case-sensitive)
    ========================================================================
*/
INT RTMPGetKeyParameter(
    IN  PCHAR   key,
    OUT PCHAR   dest,   
    IN  INT     destsize,
    IN  PCHAR   buffer)
{
    UCHAR *temp_buf1 = NULL;
    UCHAR *temp_buf2 = NULL;
    CHAR *start_ptr;
    CHAR *end_ptr;
    CHAR *ptr;
    CHAR *offset = 0;
    INT  len;

	os_alloc_mem(NULL, &temp_buf1, MAX_PARAM_BUFFER_SIZE);

	if(temp_buf1 == NULL)
        return (FALSE);
	
	os_alloc_mem(NULL, &temp_buf2, MAX_PARAM_BUFFER_SIZE);
	if(temp_buf2 == NULL)
	{
		os_free_mem(NULL, temp_buf1);
        return (FALSE);
	}
	
    //find section
    if((offset = RTMPFindSection(buffer)) == NULL)
    {
    	os_free_mem(NULL, temp_buf1);
    	os_free_mem(NULL, temp_buf2);
        return (FALSE);
    }

    strcpy(temp_buf1, "\n");
    strcat(temp_buf1, key);
    strcat(temp_buf1, "=");

    //search key
    if((start_ptr=rtstrstr(offset, temp_buf1))==NULL)
    {
		os_free_mem(NULL, temp_buf1);
    	os_free_mem(NULL, temp_buf2);
        return (FALSE);
    }

    start_ptr+=strlen("\n");
    if((end_ptr=rtstrstr(start_ptr, "\n"))==NULL)
       end_ptr=start_ptr+strlen(start_ptr);

    if (end_ptr<start_ptr)
    {
		os_free_mem(NULL, temp_buf1);
    	os_free_mem(NULL, temp_buf2);
        return (FALSE);
    }

    memcpy(temp_buf2, start_ptr, end_ptr-start_ptr);
    temp_buf2[end_ptr-start_ptr]='\0';
    len = strlen(temp_buf2);
    strcpy(temp_buf1, temp_buf2);
    if((start_ptr=rtstrstr(temp_buf1, "=")) == NULL)
    {
		os_free_mem(NULL, temp_buf1);
    	os_free_mem(NULL, temp_buf2);
        return (FALSE);
    }

    strcpy(temp_buf2, start_ptr+1);
    ptr = temp_buf2;
    //trim space or tab
    while(*ptr != 0x00)
    {
        if( (*ptr == ' ') || (*ptr == '\t') )
            ptr++;
        else
           break;
    }

    len = strlen(ptr);    
    memset(dest, 0x00, destsize);
    strncpy(dest, ptr, len >= destsize ?  destsize: len);

	os_free_mem(NULL, temp_buf1);
    os_free_mem(NULL, temp_buf2);
    return TRUE;
}

static void rtmp_read_key_parms_from_file(IN  PRTMP_ADAPTER pAd, char *tmpbuf, char *buffer)
{
	char		tok_str[16];					
	INT			idx;
	ULONG		KeyType;
	ULONG		KeyLen;
	ULONG		KeyIdx;
	UCHAR		CipherAlg = CIPHER_WEP64;

	//DefaultKeyID
	if(RTMPGetKeyParameter("DefaultKeyID", tmpbuf, 25, buffer))
	{            
		KeyIdx = simple_strtol(tmpbuf, 0, 10);	
		if((KeyIdx >= 1 ) && (KeyIdx <= 4))
			pAd->CommonCfg.DefaultKeyId = (UCHAR) (KeyIdx - 1 );
		else
			pAd->CommonCfg.DefaultKeyId = 0;

		DBGPRINT(RT_DEBUG_TRACE, ("DefaultKeyID(0~3)=%d\n", pAd->CommonCfg.DefaultKeyId));
	}

	for (idx=0; idx<4; idx++) 
	{
		sprintf(tok_str, "Key%dType", idx+1);
		//Key%dType
		if(RTMPGetKeyParameter(tok_str, tmpbuf, 32, buffer))
		{
			KeyType = simple_strtol(tmpbuf, 0, 10);

			sprintf(tok_str, "Key%dStr", idx+1);
			//Key%dStr
			if(RTMPGetKeyParameter(tok_str, tmpbuf, 128, buffer))
			{
				KeyLen = strlen(tmpbuf);
				if(KeyType == 0)
				{//Hex type
					if( (KeyLen == 10) || (KeyLen == 26))
					{
						pAd->SharedKey[BSS0][idx].KeyLen = KeyLen / 2;
						AtoH(tmpbuf, pAd->SharedKey[BSS0][idx].Key, KeyLen / 2);
						if (KeyLen == 10)
							CipherAlg = CIPHER_WEP64;
						else
							CipherAlg = CIPHER_WEP128;
						pAd->SharedKey[BSS0][idx].CipherAlg = CipherAlg;

						DBGPRINT(RT_DEBUG_TRACE, ("Key%dStr=%s and type=%s\n", idx+1, tmpbuf, (KeyType==0) ? "Hex":"Ascii"));
					}
					else
					{ //Invalid key length
						DBGPRINT(RT_DEBUG_ERROR, ("Key%dStr is Invalid key length!\n", idx+1));
					}								
				}
				else
				{//Ascii								
					if( (KeyLen == 5) || (KeyLen == 13))
					{
						pAd->SharedKey[BSS0][idx].KeyLen = KeyLen;
						NdisMoveMemory(pAd->SharedKey[BSS0][idx].Key, tmpbuf, KeyLen);
						if (KeyLen == 5)
							CipherAlg = CIPHER_WEP64;
						else
							CipherAlg = CIPHER_WEP128;
						pAd->SharedKey[BSS0][idx].CipherAlg = CipherAlg;
					
						DBGPRINT(RT_DEBUG_TRACE, ("Key%dStr=%s and type=%s\n", idx+1, tmpbuf, (KeyType==0) ? "Hex":"Ascii"));		
					}
					else
					{ //Invalid key length
						DBGPRINT(RT_DEBUG_ERROR, ("Key%dStr is Invalid key length!\n", idx+1));
					}
				}
			}
		}
	}
}

/*
    ========================================================================

    Routine Description:
        In kernel mode read parameters from file

    Arguments:
        src                     the location of the file.
        dest					put the parameters to the destination.
        Length                  size to read.

    Return Value:
        None

    Note:

    ========================================================================
*/
static void rtmp_read_wmm_parms_from_file(IN  PRTMP_ADAPTER pAd, char *tmpbuf, char *buffer)
{
	PUCHAR					macptr;						
	INT						i=0;
	BOOLEAN					bWmmEnable = FALSE;
	
	//WmmCapable
	if(RTMPGetKeyParameter("WmmCapable", tmpbuf, 32, buffer))
	{
		if((simple_strtol(tmpbuf, 0, 10) != 0) && (pAd->NumberOfPipes >= 5))  //Enable
		{
			pAd->CommonCfg.bWmmCapable = TRUE;
			bWmmEnable = TRUE;
		}
		else //Disable
		{
			pAd->CommonCfg.bWmmCapable = FALSE;
		}
		
		DBGPRINT(RT_DEBUG_TRACE, ("WmmCapable=%d\n", pAd->CommonCfg.bWmmCapable));
	}
	//AckPolicy for AC_BK, AC_BE, AC_VI, AC_VO
	if(RTMPGetKeyParameter("AckPolicy", tmpbuf, 32, buffer))
	{			
		for (i = 0, macptr = rstrtok(tmpbuf,";"); macptr; macptr = rstrtok(NULL,";"), i++)
		{
			pAd->CommonCfg.AckPolicy[i] = (UCHAR)simple_strtol(macptr, 0, 10);

			DBGPRINT(RT_DEBUG_TRACE, ("AckPolicy[%d]=%d\n", i, pAd->CommonCfg.AckPolicy[i]));
		}
	}

#if 0 //TODO: not ready!
	if (bWmmEnable)
	{
		//APSDCapable
		if(RTMPGetKeyParameter("APSDCapable", tmpbuf, 10, buffer))
		{
			if(simple_strtol(tmpbuf, 0, 10) != 0)  //Enable
				pAd->CommonCfg.bAPSDCapable = TRUE;
			else
				pAd->CommonCfg.bAPSDCapable = FALSE;

			DBGPRINT(RT_DEBUG_TRACE, ("APSDCapable=%d\n", pAd->CommonCfg.bAPSDCapable));
		}

		//APSDAC for AC_BE, AC_BK, AC_VI, AC_VO
		if(RTMPGetKeyParameter("APSDAC", tmpbuf, 32, buffer))
		{
			BOOLEAN apsd_ac[4];
						
			for (i = 0, macptr = rstrtok(tmpbuf,";"); macptr; macptr = rstrtok(NULL,";"), i++)
			{
				apsd_ac[i] = (BOOLEAN)simple_strtol(macptr, 0, 10);

				DBGPRINT(RT_DEBUG_TRACE, ("APSDAC%d  %d\n", i,  apsd_ac[i]));
			}
					
			pAd->CommonCfg.bAPSDAC_BE = apsd_ac[0];
			pAd->CommonCfg.bAPSDAC_BK = apsd_ac[1];
			pAd->CommonCfg.bAPSDAC_VI = apsd_ac[2];
			pAd->CommonCfg.bAPSDAC_VO = apsd_ac[3];
		}
	}
#endif	

}

NDIS_STATUS	RTMPReadParametersHook(
	IN	PRTMP_ADAPTER pAd)
{
	PUCHAR					src;
	struct file				*srcf;
	INT 					retval, orgfsuid, orgfsgid;
   	mm_segment_t			orgfs;
	CHAR					*buffer;
	CHAR					*tmpbuf;
	UCHAR					keyMaterial[40];
	ULONG					RtsThresh;
	ULONG					FragThresh;
	UCHAR					ucInfo;
	ULONG					ulInfo;
	UCHAR					Channel, PhyMode;

	os_alloc_mem(NULL, (PUCHAR *)&buffer, MAX_INI_BUFFER_SIZE);
	if(buffer == NULL)
        return NDIS_STATUS_FAILURE;

	os_alloc_mem(NULL, (PUCHAR *)&tmpbuf, MAX_PARAM_BUFFER_SIZE);
	if(tmpbuf == NULL)
	{
		os_free_mem(NULL, (PUCHAR)buffer);
        return NDIS_STATUS_FAILURE;
	}

   	src = PROFILE_PATH;

	// Save uid and gid used for filesystem access.
	// Set user and group to 0 (root)	
	orgfsuid = current->fsuid;
	orgfsgid = current->fsgid;
	current->fsuid=current->fsgid = 0;
    orgfs = get_fs();
    set_fs(KERNEL_DS);

	if (src && *src)
	{
		srcf = filp_open(src, O_RDONLY, 0);
		if (IS_ERR(srcf)) 
		{
			DBGPRINT(RT_DEBUG_TRACE, ("--> Error %ld opening %s\n", -PTR_ERR(srcf),src));
		}
		else 
		{
			// The object must have a read method
			if (srcf->f_op && srcf->f_op->read)
			{
				memset(buffer, 0x00, MAX_INI_BUFFER_SIZE);
				retval=srcf->f_op->read(srcf, buffer, MAX_INI_BUFFER_SIZE, &srcf->f_pos);
				if (retval < 0)
				{
					DBGPRINT(RT_DEBUG_TRACE, ("--> Read %s error %d\n", src, -retval));
				}
				else
				{
					DBGPRINT(RT_DEBUG_TRACE, ("====> Read from the profile .....\n"));

					// set file parameter to portcfg
					//CountryRegion
					if(RTMPGetKeyParameter("CountryRegion", tmpbuf, 25, buffer))
					{
						ucInfo = (UCHAR) simple_strtol(tmpbuf, 0, 10);
						if (ucInfo <= REGION_MAXIMUM_BG_BAND)
							pAd->CommonCfg.CountryRegion = ucInfo; 
						else
						{	
							ucInfo = 0; // ch 0~11
							pAd->CommonCfg.CountryRegion = ucInfo;	
						}
						DBGPRINT(RT_DEBUG_TRACE, ("CountryRegion=%d\n", pAd->CommonCfg.CountryRegion));
					}
					//CountryRegionABand
					if(RTMPGetKeyParameter("CountryRegionABand", tmpbuf, 25, buffer))
					{
						ucInfo = (UCHAR) simple_strtol(tmpbuf, 0, 10);
						if (ucInfo <= REGION_MAXIMUM_A_BAND)
							pAd->CommonCfg.CountryRegionForABand = ucInfo;
						else
						{
							ucInfo = 7;
							pAd->CommonCfg.CountryRegionForABand = ucInfo;
						}					
						DBGPRINT(RT_DEBUG_TRACE, ("CountryRegionABand=%d\n", pAd->CommonCfg.CountryRegionForABand));
					}
					//SSID
					if(RTMPGetKeyParameter("SSID", tmpbuf, 256, buffer))
					{
						if (strlen(tmpbuf) <= 32)
						{
		 					pAd->CommonCfg.SsidLen = (UCHAR) strlen(tmpbuf);
							memcpy(pAd->CommonCfg.Ssid, tmpbuf, pAd->CommonCfg.SsidLen);
							// prepare ssid for auto reconection
							pAd->MlmeAux.AutoReconnectSsidLen = pAd->CommonCfg.SsidLen;
							memcpy(pAd->MlmeAux.AutoReconnectSsid, tmpbuf, pAd->MlmeAux.AutoReconnectSsidLen);
							pAd->MlmeAux.SsidLen = pAd->CommonCfg.SsidLen;
							memcpy(pAd->MlmeAux.Ssid, tmpbuf, pAd->MlmeAux.SsidLen);

							DBGPRINT(RT_DEBUG_TRACE, ("SSID=%s\n", tmpbuf));
						}								
					}
					//NetworkType
					if (RTMPGetKeyParameter("NetworkType", tmpbuf, 25, buffer))
					{
						pAd->bConfigChanged = TRUE;
						if (strcmp(tmpbuf, "Adhoc") == 0)
							pAd->StaCfg.BssType = BSS_ADHOC;
						else //Default Infrastructure mode
							pAd->StaCfg.BssType = BSS_INFRA;
						// Reset Ralink supplicant to not use, it will be set to start when UI set PMK key
						pAd->StaCfg.WpaState = SS_NOTUSE;
						DBGPRINT(RT_DEBUG_TRACE, ("NetworkType=%d\n", pAd->StaCfg.BssType));
					}
					//Channel
					if(RTMPGetKeyParameter("Channel", tmpbuf, 10, buffer))
					{
						Channel = (UCHAR)simple_strtol(tmpbuf, 0, 10);
						pAd->CommonCfg.AdhocChannel = Channel;
						if (ChannelSanity(pAd, Channel) == TRUE)
						{
							pAd->CommonCfg.Channel = Channel;
							DBGPRINT(RT_DEBUG_TRACE, ("Channel=%d\n", pAd->CommonCfg.Channel));
						}
					}
					//WirelessMode
					if(RTMPGetKeyParameter("WirelessMode", tmpbuf, 10, buffer))
					{
						ucInfo = (UCHAR)simple_strtol(tmpbuf, 0, 10);
						// switch to phy mode
						if (ucInfo <= PHY_11ABG_MIXED) 
							PhyMode = ucInfo;
						else if (ucInfo == 4)
							PhyMode = PHY_11ABGN_MIXED;
						else
							PhyMode = PHY_11BGN_MIXED;
	
						pAd->CommonCfg.PhyMode = PhyMode;
					
						DBGPRINT(RT_DEBUG_TRACE, ("PhyMode=%d\n", pAd->CommonCfg.PhyMode));
					}
					//TxRate
					if(RTMPGetKeyParameter("TxRate", tmpbuf, 10, buffer))
					{
						ulInfo = simple_strtol(tmpbuf, 0, 10);
#if 1					
						//TODO: Always set to "Auto" in the beginning (temporality)
						RTMPSetDesiredRates(pAd, -1);
#else
						switch (pAd->CommonCfg.PhyMode)
						{
							case PHY_11B:
								RTMPSetDesiredRates(pAd, ((ulInfo > 0) && (ulInfo <= 4)) ? (LONG)(RateIdToMbps[ulInfo-1] * 1000000) : (-1));
								break;

							case PHY_11BG_MIXED:
							case PHY_11ABG_MIXED:
							case PHY_11ABGN_MIXED:
							case PHY_11BGN_MIXED:
								RTMPSetDesiredRates(pAd, ((ulInfo > 0) && (ulInfo <= 12)) ? (LONG)(RateIdToMbps[ulInfo-1] * 1000000) : (-1));
								break;

							case PHY_11A:
							case PHY_11G:
							case PHY_11GN_MIXED:
							case PHY_11AN_MIXED:
							case PHY_11AGN_MIXED:
								RTMPSetDesiredRates(pAd, ((ulInfo > 4) && (ulInfo <= 12)) ? (LONG)(RateIdToMbps[ulInfo-1] * 1000000) : (-1));
								break;
			
							default:	// Auto
								RTMPSetDesiredRates(pAd, -1);
								break;
						}		
#endif
						DBGPRINT(RT_DEBUG_TRACE, ("TxRate=%ld\n", ulInfo));
					}				
					//BGProtection
					if(RTMPGetKeyParameter("BGProtection", tmpbuf, 10, buffer))
					{
#ifndef WIFI_TEST
						pAd->CommonCfg.UseBGProtection = 2;// disable b/g protection for throughput test
#else
						switch (simple_strtol(tmpbuf, 0, 10))
						{
							case 1: //Always On
								pAd->CommonCfg.UseBGProtection = 1;
								break;
							case 2: //Always OFF
								pAd->CommonCfg.UseBGProtection = 2;
								break;
							case 0: //AUTO
							default:
								pAd->CommonCfg.UseBGProtection = 0;
								break;
						}
#endif
						DBGPRINT(RT_DEBUG_TRACE, ("BGProtection=%ld\n", pAd->CommonCfg.UseBGProtection));
					}
					//TxPreamble
					if(RTMPGetKeyParameter("TxPreamble", tmpbuf, 10, buffer))
					{
						switch (simple_strtol(tmpbuf, 0, 10))
						{
							case Rt802_11PreambleShort:
								pAd->CommonCfg.TxPreamble = Rt802_11PreambleShort;
								break;
							case Rt802_11PreambleLong:
							case Rt802_11PreambleAuto:
							default:
								// if user wants AUTO, initialize to LONG here, then change according to AP's
								// capability upon association.
								pAd->CommonCfg.TxPreamble = Rt802_11PreambleLong;
								MlmeSetTxPreamble(pAd, Rt802_11PreambleLong);
								break;
						}
						DBGPRINT(RT_DEBUG_TRACE, ("TxPreamble=%d\n", pAd->CommonCfg.TxPreamble));
					}
					//RTSThreshold
					if(RTMPGetKeyParameter("RTSThreshold", tmpbuf, 10, buffer))
					{
						RtsThresh = simple_strtol(tmpbuf, 0, 10);
						if( (RtsThresh < 0) || (RtsThresh > MAX_RTS_THRESHOLD) )
						{
							//illegal RtsThreshold so we set it to default
							RtsThresh = MAX_RTS_THRESHOLD;
						}
						
						pAd->CommonCfg.RtsThreshold  = (USHORT)RtsThresh;
						DBGPRINT(RT_DEBUG_TRACE, ("RTSThreshold=%d\n", pAd->CommonCfg.RtsThreshold));
					}
					//FragThreshold
					if(RTMPGetKeyParameter("FragThreshold", tmpbuf, 10, buffer))
					{		
						FragThresh = simple_strtol(tmpbuf, 0, 10);
						pAd->CommonCfg.bUseZeroToDisableFragment = FALSE;

						if ((FragThresh > MAX_FRAG_THRESHOLD) || (FragThresh < MIN_FRAG_THRESHOLD))
						{ //illegal FragThresh so we set it to default
							FragThresh = MAX_FRAG_THRESHOLD;
							pAd->CommonCfg.bUseZeroToDisableFragment = TRUE;
						}

						pAd->CommonCfg.FragmentThreshold = (USHORT)FragThresh;
						DBGPRINT(RT_DEBUG_TRACE, ("FragThreshold=%d\n", pAd->CommonCfg.FragmentThreshold));
					}
					//TxBurst
					if(RTMPGetKeyParameter("TxBurst", tmpbuf, 10, buffer))
					{
#ifdef WIFI_TEST
						pAd->CommonCfg.bEnableTxBurst = FALSE;
#else
						if(simple_strtol(tmpbuf, 0, 10) != 0)  //Enable
							pAd->CommonCfg.bEnableTxBurst = TRUE;
						else //Disable
							pAd->CommonCfg.bEnableTxBurst = FALSE;
#endif
						DBGPRINT(RT_DEBUG_TRACE, ("TxBurst=%d\n", pAd->CommonCfg.bEnableTxBurst));
					}
					
#ifdef AGGREGATION_SUPPORT
					//PktAggregate (Ralink proprietary feature)
					if(RTMPGetKeyParameter("PktAggregate", tmpbuf, 10, buffer))
					{
						if(simple_strtol(tmpbuf, 0, 10) != 0)  //Enable
							pAd->CommonCfg.bAggregationCapable = TRUE;
						else //Disable
							pAd->CommonCfg.bAggregationCapable = FALSE;

						DBGPRINT(RT_DEBUG_TRACE, ("PktAggregate=%d\n", pAd->CommonCfg.bAggregationCapable));
					}
#endif	/* AGGREGATION_SUPPORT */
					
					// WmmCapable
					rtmp_read_wmm_parms_from_file(pAd, tmpbuf, buffer);
					
					//IEEE80211H
					if(RTMPGetKeyParameter("IEEE80211H", tmpbuf, 10, buffer))
					{
						if(simple_strtol(tmpbuf, 0, 10) != 0)  //Enable
							pAd->CommonCfg.bIEEE80211H = TRUE;
						else //Disable
							pAd->CommonCfg.bIEEE80211H = FALSE;

						DBGPRINT(RT_DEBUG_TRACE, ("IEEE80211H=%d\n", pAd->CommonCfg.bIEEE80211H));
					}
					//PSMode
					if (RTMPGetKeyParameter("PSMode", tmpbuf, 10, buffer))
					{
						if (pAd->StaCfg.BssType == BSS_INFRA)
						{
							if ((strcmp(tmpbuf, "MAX_PSP") == 0) || (strcmp(tmpbuf, "max_psp") == 0))
							{
								// do NOT turn on PSM bit here, wait until MlmeCheckForPsmChange()
								// to exclude certain situations.
								//	   MlmeSetPsm(pAd, PWR_SAVE);
								OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM);
								if (pAd->StaCfg.bWindowsACCAMEnable == FALSE)
									pAd->StaCfg.WindowsPowerMode = Ndis802_11PowerModeFast_PSP;
								pAd->StaCfg.WindowsBatteryPowerMode = Ndis802_11PowerModeFast_PSP;
								pAd->StaCfg.DefaultListenCount = 3;
							}							
							else if ((strcmp(tmpbuf, "Fast_PSP") == 0) || (strcmp(tmpbuf, "fast_psp") == 0) 
								|| (strcmp(tmpbuf, "FAST_PSP") == 0))
							{
								// do NOT turn on PSM bit here, wait until MlmeCheckForPsmChange()
								// to exclude certain situations.
								//	   MlmeSetPsmBit(pAd, PWR_SAVE);
								OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM);
								if (pAd->StaCfg.bWindowsACCAMEnable == FALSE)
									pAd->StaCfg.WindowsPowerMode = Ndis802_11PowerModeFast_PSP;
								pAd->StaCfg.WindowsBatteryPowerMode = Ndis802_11PowerModeFast_PSP;
								pAd->StaCfg.DefaultListenCount = 3;
							}
							else
							{ //Default Ndis802_11PowerModeCAM
								// clear PSM bit immediately
								MlmeSetPsm(pAd, PWR_ACTIVE);
								OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM);
								if (pAd->StaCfg.bWindowsACCAMEnable == FALSE)
									pAd->StaCfg.WindowsPowerMode = Ndis802_11PowerModeCAM;
								pAd->StaCfg.WindowsBatteryPowerMode = Ndis802_11PowerModeCAM;
							}
							DBGPRINT(RT_DEBUG_TRACE, ("PSMode=%ld\n", pAd->StaCfg.WindowsPowerMode));
						}
					}
			        // FastRoaming
					if (RTMPGetKeyParameter("FastRoaming", tmpbuf, 32, buffer))
					{
                   	    ulInfo = simple_strtol(tmpbuf, 0, 10);

					    if (ulInfo == 1)
					        pAd->StaCfg.bFastRoaming = TRUE;
                        else
                            pAd->StaCfg.bFastRoaming = FALSE;

                        DBGPRINT(RT_DEBUG_TRACE, ("FastRoaming=%d\n", pAd->StaCfg.bFastRoaming));
					}
					// RoamThreshold
					if (RTMPGetKeyParameter("RoamThreshold", tmpbuf, 32, buffer))
					{
                        ulInfo = simple_strtol(tmpbuf, 0, 10);
                            
                        if (ulInfo > 90 || ulInfo < 60)
                            pAd->StaCfg.dBmToRoam = -70;
                        else    
                            pAd->StaCfg.dBmToRoam = (-1)*ulInfo;

                        DBGPRINT(RT_DEBUG_TRACE, ("RoamThreshold=%d  dBm\n", pAd->StaCfg.dBmToRoam));
					}
					//AuthMode
					if (RTMPGetKeyParameter("AuthMode", tmpbuf, 32, buffer))
					{ 
						if ((strcmp(tmpbuf, "WEPAUTO") == 0) || (strcmp(tmpbuf, "wepauto") == 0))
							pAd->CommonCfg.AuthMode = Ndis802_11AuthModeAutoSwitch;
						else if ((strcmp(tmpbuf, "SHARED") == 0) || (strcmp(tmpbuf, "shared") == 0))
							pAd->CommonCfg.AuthMode = Ndis802_11AuthModeShared;
						else if ((strcmp(tmpbuf, "WPAPSK") == 0) || (strcmp(tmpbuf, "wpapsk") == 0))
							pAd->CommonCfg.AuthMode = Ndis802_11AuthModeWPAPSK;
						else if ((strcmp(tmpbuf, "WPANONE") == 0) || (strcmp(tmpbuf, "wpanone") == 0))
							pAd->CommonCfg.AuthMode = Ndis802_11AuthModeWPANone;
						else if ((strcmp(tmpbuf, "WPA2PSK") == 0) || (strcmp(tmpbuf, "wpa2psk") == 0))
							pAd->CommonCfg.AuthMode = Ndis802_11AuthModeWPA2PSK;
                        else if ((strcmp(tmpbuf, "WPA") == 0) || (strcmp(tmpbuf, "wpa") == 0))
                            pAd->CommonCfg.AuthMode = Ndis802_11AuthModeWPA;  
                        else if ((strcmp(tmpbuf, "WPA2") == 0) || (strcmp(tmpbuf, "wpa2") == 0))
                            pAd->CommonCfg.AuthMode = Ndis802_11AuthModeWPA2;
						else
							pAd->CommonCfg.AuthMode = Ndis802_11AuthModeOpen;

						pAd->StaCfg.PortSecured = WPA_802_1X_PORT_NOT_SECURED;
						DBGPRINT(RT_DEBUG_TRACE, ("AuthMode=%d\n", pAd->CommonCfg.AuthMode));
					}
					//EncrypType
					if (RTMPGetKeyParameter("EncrypType", tmpbuf, 32, buffer))
					{
						if ((strcmp(tmpbuf, "WEP") == 0) || (strcmp(tmpbuf, "wep") == 0))
							pAd->CommonCfg.WepStatus	= Ndis802_11WEPEnabled;
						else if ((strcmp(tmpbuf, "TKIP") == 0) || (strcmp(tmpbuf, "tkip") == 0))
							pAd->CommonCfg.WepStatus	 = Ndis802_11Encryption2Enabled;
						else if ((strcmp(tmpbuf, "AES") == 0) || (strcmp(tmpbuf, "aes") == 0))
							pAd->CommonCfg.WepStatus	= Ndis802_11Encryption3Enabled;
						else
							pAd->CommonCfg.WepStatus	= Ndis802_11WEPDisabled;

						pAd->CommonCfg.PairCipher	= pAd->CommonCfg.WepStatus;
						pAd->CommonCfg.GroupCipher	= pAd->CommonCfg.WepStatus;
						
						RTMPMakeRSNIE(pAd,pAd->CommonCfg.GroupCipher);
						DBGPRINT(RT_DEBUG_TRACE, ("EncrypType=%d\n", pAd->CommonCfg.WepStatus));
					}			 
					//WPAPSK KEY
					if (RTMPGetKeyParameter("WPAPSK", tmpbuf, 255, buffer))
					{			        
						INT		err = 0;
                        tmpbuf[strlen(tmpbuf)]='\0'; // the end of this string
                        
						if ((pAd->CommonCfg.AuthMode == Ndis802_11AuthModeWPAPSK) ||\
							(pAd->CommonCfg.AuthMode == Ndis802_11AuthModeWPA2PSK) ||\
							(pAd->CommonCfg.AuthMode == Ndis802_11AuthModeWPANone) )
						{	
							if ((strlen(tmpbuf) >= 8) && (strlen(tmpbuf) < 64))	// passphrase
							{
								PasswordHash((char *)tmpbuf, pAd->CommonCfg.Ssid, pAd->CommonCfg.SsidLen, keyMaterial);
								NdisMoveMemory(pAd->StaCfg.PMK, keyMaterial, 32);
							}
							else if (strlen(tmpbuf) == 64)	// 64 Hex
							{							 
								AtoH(tmpbuf, keyMaterial, 32);
								NdisMoveMemory(pAd->StaCfg.PMK, keyMaterial, 32);
							}
							else	// error
							{
								err = 1;
								DBGPRINT(RT_DEBUG_ERROR, ("WPAPSK key-string required 8 ~ 64 characters!\n"));
							}
							// Initialize WPA state machine
							if (err == 0)
							{  
								if ((pAd->CommonCfg.AuthMode == Ndis802_11AuthModeWPAPSK) ||\
									(pAd->CommonCfg.AuthMode == Ndis802_11AuthModeWPA2PSK))
								{
									// Use RaConfig as PSK agent.
									// Start STA supplicant state machine
									pAd->StaCfg.WpaState = SS_START;
								}
							
								DBGPRINT(RT_DEBUG_TRACE, ("WPAPSK=%s\n", tmpbuf));
							}
						}
    				}
    				
					//DefaultKeyID, KeyType, KeyStr
					rtmp_read_key_parms_from_file(pAd, tmpbuf, buffer);

					//11n High Throughput settings						 
					HTParametersHook(pAd, tmpbuf, buffer);

					DBGPRINT(RT_DEBUG_TRACE, ("<==== Read from the profile\n"));

				}
			}
			else
			{
				DBGPRINT(RT_DEBUG_TRACE, ("--> %s does not have a write method\n", src));
			}
			
			retval=filp_close(srcf,NULL);
			
			if (retval)
			{
				DBGPRINT(RT_DEBUG_TRACE, ("--> Error %d closing %s\n", -retval, src));
			}
		}
	}

	set_fs(orgfs);
	current->fsuid = orgfsuid;
	current->fsgid = orgfsgid;

	os_free_mem(NULL, (PUCHAR)buffer);
	os_free_mem(NULL, (PUCHAR)tmpbuf);

	return (NDIS_STATUS_SUCCESS);	
}

static void	HTParametersHook(
	IN	PRTMP_ADAPTER pAd, 
	IN	CHAR		  *pValueStr,
	IN	CHAR		  *pInput)
{

	INT Value;

    if (RTMPGetKeyParameter("HT_BADecline", pValueStr, 25, pInput))
    {
        Value = simple_strtol(pValueStr, 0, 10);
        if (Value == 0)
        {
			pAd->CommonCfg.BACapability.field.Policy = IMMED_BA;
        }
        else
        {
			pAd->CommonCfg.BACapability.field.Policy = BA_NOTUSE;
        }
        DBGPRINT(RT_DEBUG_TRACE, ("HT: BA Decline  = %s\n", (Value==0) ? "Disable" : "Enable"));
    }

    if (RTMPGetKeyParameter("HT_AutoBA", pValueStr, 25, pInput))
    {
        Value = simple_strtol(pValueStr, 0, 10);
        if (Value == 0)
        {
            pAd->CommonCfg.BACapability.field.AutoBA = FALSE;
        }
        else
        {
            pAd->CommonCfg.BACapability.field.AutoBA = TRUE;
        }
        DBGPRINT(RT_DEBUG_TRACE, ("HT: Auto BA  = %s\n", (Value==0) ? "Disable" : "Enable"));
    }
    
	// Enable HT Link Adaptation Control. (Ralink proprietary feature)
	if (RTMPGetKeyParameter("HT_LinkAdapt", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);
		if (Value == 0)
		{
			pAd->bLinkAdapt = FALSE;
		}
		else
		{
			//pAd->HTCEnable = TRUE;
			pAd->bLinkAdapt = TRUE;
		}
		DBGPRINT(RT_DEBUG_TRACE, ("HT: Link Adaptation Control = %s\n", (Value==0) ? "Disable" : "Enable(+HTC)"));
	}

	// Reverse Direction Mechanism. (Ralink proprietary feature)
    if (RTMPGetKeyParameter("HT_RDG", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);
		if (Value == 0)
		{			
			pAd->CommonCfg.bRdg = FALSE;
		}
		else
		{
			//pAd->HTCEnable = TRUE;
            pAd->CommonCfg.bRdg = TRUE;
		}
		DBGPRINT(RT_DEBUG_TRACE, ("HT: RDG = %s\n", (Value==0) ? "Disable" : "Auto(+HTC)"));
	}

	// Tx A-MSUD
    if (RTMPGetKeyParameter("HT_AMSDU", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);
		if (Value == 0)
		{
			pAd->CommonCfg.BACapability.field.AmsduEnable = FALSE;
		}
		else
		{
            pAd->CommonCfg.BACapability.field.AmsduEnable = TRUE;
		}
		DBGPRINT(RT_DEBUG_TRACE, ("HT: Tx A-MSDU = %s\n", (Value==0) ? "Disable" : "Enable"));
	}

	// MPDU Density
    if (RTMPGetKeyParameter("HT_MpduDensity", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);
		if (Value <=7 && Value >= 0)
		{		
			pAd->CommonCfg.BACapability.field.MpduDensity = Value;
			DBGPRINT(RT_DEBUG_TRACE, ("HT: MPDU Density = %d\n", Value));
		}
		else
		{
			pAd->CommonCfg.BACapability.field.MpduDensity = 0;
			DBGPRINT(RT_DEBUG_TRACE, ("HT: MPDU Density = %d (Default)\n", 0));
		}
	}

	// Guard Interval
	if (RTMPGetKeyParameter("HT_GI", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);

		if (Value == GI_400)
		{
			pAd->CommonCfg.RegTransmitSetting.field.ShortGI = GI_400;
		}
		else
		{
			pAd->CommonCfg.RegTransmitSetting.field.ShortGI = GI_800;
		}
		
		DBGPRINT(RT_DEBUG_TRACE, ("HT: Guard Interval = %s\n", (Value==GI_400) ? "400" : "800" ));
	}

	// HT Operation Mode : Mixed Mode , Green Field
	if (RTMPGetKeyParameter("HT_OpMode", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);

		if (Value == HTMODE_GF)
		{

			pAd->CommonCfg.RegTransmitSetting.field.HTMODE  = HTMODE_GF;
		}
		else
		{
			pAd->CommonCfg.RegTransmitSetting.field.HTMODE  = HTMODE_MM;
		}		

		DBGPRINT(RT_DEBUG_TRACE, ("HT: Operate Mode = %s\n", (Value==HTMODE_GF) ? "Green Field" : "Mixed Mode" ));
	}

	// Channel Width
	if (RTMPGetKeyParameter("HT_BW", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);

		if (Value == BW_40)
		{
			
			pAd->CommonCfg.RegTransmitSetting.field.BW  = BW_40;
		}
		else
		{
            pAd->CommonCfg.RegTransmitSetting.field.BW  = BW_20;
		}		

		DBGPRINT(RT_DEBUG_TRACE, ("HT: Channel Width = %s\n", (Value==BW_40) ? "40 MHz" : "20 MHz" ));
	}

	// MSC
	if (RTMPGetKeyParameter("HT_MCS", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);

		if (Value >= 0 && Value < MCS_AUTO)
		{
			OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_TX_RATE_SWITCH_ENABLED); 
			pAd->CommonCfg.RegTransmitSetting.field.MCS  = Value;
		}
		else
		{
			OPSTATUS_SET_FLAG(pAd, fOP_STATUS_TX_RATE_SWITCH_ENABLED); 
			pAd->CommonCfg.RegTransmitSetting.field.MCS  = MCS_AUTO;
		}		

		if (pAd->CommonCfg.RegTransmitSetting.field.MCS  == MCS_AUTO)
		{
			DBGPRINT(RT_DEBUG_TRACE, ("HT: MCS = Auto\n"));
		}
		else
		{
            DBGPRINT(RT_DEBUG_TRACE, ("HT: MCS = %d\n", pAd->CommonCfg.RegTransmitSetting.field.MCS));
		}
	}

	// STBC 
    if (RTMPGetKeyParameter("HT_STBC", pValueStr, 25, pInput))
	{
		Value = simple_strtol(pValueStr, 0, 10);
		if (Value == STBC_USE)
		{		
			pAd->CommonCfg.RegTransmitSetting.field.STBC = STBC_USE;
		}
		else
		{
			pAd->CommonCfg.RegTransmitSetting.field.STBC = STBC_NONE;
		}
		DBGPRINT(RT_DEBUG_TRACE, ("HT: STBC = %s\n", (Value==0) ? "Disable" : "Auto"));
	}
	
	// HT_MIMOPS
	if (RTMPGetKeyParameter("HT_MIMOPS", pValueStr, 25, pInput))
    {
        Value = simple_strtol(pValueStr, 0, 10);
        
        if (Value == 0)
            pAd->CommonCfg.BACapability.field.MMPSmode = MMPS_STATIC;
        else if (Value == 1)
            pAd->CommonCfg.BACapability.field.MMPSmode = MMPS_DYNAMIC;
		else
			pAd->CommonCfg.BACapability.field.MMPSmode = MMPS_ENABLE;
			
        DBGPRINT(RT_DEBUG_TRACE, ("HT: MIMOPS  = %d\n", pAd->CommonCfg.BACapability.field.MMPSmode));
    }
}


