/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology	5th	Rd.
 * Science-based Industrial	Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002-2006, Ralink Technology, Inc.
 *
 * All rights reserved.	Ralink's source	code is	an unpublished work	and	the
 * use of a	copyright notice does not imply	otherwise. This	source code
 * contains	confidential trade secret material of Ralink Tech. Any attemp
 * or participation	in deciphering,	decoding, reverse engineering or in	any
 * way altering	the	source code	is stricitly prohibited, unless	the	prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************
 
 	Module Name:
	rtusb_io.c

	Abstract:

	Revision History:
	Who			When	    What
	--------	----------  ----------------------------------------------
	Name		Date	    Modification logs
	Paul Lin    06-25-2004  created
*/

#include	"rt_config.h"


/*
	========================================================================
	
	Routine Description: NIC initialization complete

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/

NTSTATUS	RTUSBFirmwareRun(
	IN	PRTMP_ADAPTER	pAd)
{
	NTSTATUS	Status;

	Status = RTUSB_VendorRequest(
		pAd,
		USBD_TRANSFER_DIRECTION_OUT,
		DEVICE_VENDOR_REQUEST_OUT,
		0x01,
		0x8,
		0,
		NULL,
		0);
	
	return Status;
}

/*
	========================================================================
	
	Routine Description: Get current firmware operation mode (Return Value)

	Arguments:

	Return Value: 
		0 or 1 = Downloaded by host driver
		others = Driver doesn't download firmware

	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBFirmwareOpmode(
	IN	PRTMP_ADAPTER	pAd,
	OUT	PUINT32			pValue)
{
	NTSTATUS	Status;

	Status = RTUSB_VendorRequest(
		pAd,
		(USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK),
		DEVICE_VENDOR_REQUEST_IN,
		0x1,
		0x11,
		0,
		pValue,
		4);
	return Status;
}
NTSTATUS	RTUSBVenderReset(
	IN	PRTMP_ADAPTER	pAd)
{
	NTSTATUS	Status;
	DBGPRINT_RAW(RT_DEBUG_ERROR, ("-->RTUSBVenderReset\n"));
	Status = RTUSB_VendorRequest(
		pAd,
		USBD_TRANSFER_DIRECTION_OUT,
		DEVICE_VENDOR_REQUEST_OUT,
		0x01,
		0x1,
		0,
		NULL,
		0);

	DBGPRINT_RAW(RT_DEBUG_ERROR, ("<--RTUSBVenderReset\n"));
	return Status;
}
/*
	========================================================================
	
	Routine Description: Read various length data from RT2573

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBMultiRead(
	IN	PRTMP_ADAPTER	pAd,
	IN	USHORT			Offset,
	OUT	PUCHAR			pData,
	IN	USHORT			length)
{
	NTSTATUS	Status;

	Status = RTUSB_VendorRequest(
		pAd,
		(USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK),
		DEVICE_VENDOR_REQUEST_IN,
		0x7,
		0,
		Offset,
		pData,
		length);
	
	return Status;
}

/*
	========================================================================
	
	Routine Description: Write various length data to RT2573

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBMultiWrite(
	IN	PRTMP_ADAPTER	pAd,
	IN	USHORT			Offset,
	IN	PUCHAR			pData,
	IN	USHORT			length)
{
	NTSTATUS	Status;


	Status = RTUSB_VendorRequest(
		pAd,
		USBD_TRANSFER_DIRECTION_OUT,
		DEVICE_VENDOR_REQUEST_OUT,
		0x6,
		0,
		Offset,
		pData,
		length);
	
	return Status;
}

/*
	========================================================================
	
	Routine Description: Read 32-bit MAC register

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBReadMACRegister(
	IN	PRTMP_ADAPTER	pAd,
	IN	USHORT			Offset,
	OUT	PUINT32			pValue)
{
	NTSTATUS	Status;
#ifdef BIG_ENDIAN		
	UINT32		Value_Big;
#endif

#ifdef BIG_ENDIAN
	Status = RTUSB_VendorRequest(
		pAd,
		(USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK),
		DEVICE_VENDOR_REQUEST_IN,
		0x7,
		0,
		Offset,
		&Value_Big,
		4);
	
	*pValue = SWAP32(Value_Big);
#else
	Status = RTUSB_VendorRequest(
		pAd,
		(USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK),
		DEVICE_VENDOR_REQUEST_IN,
		0x7,
		0,
		Offset,
		pValue,
		4);
#endif
	return Status;
}

/*
	========================================================================
	
	Routine Description: Write 32-bit MAC register

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBWriteMACRegister(
	IN	PRTMP_ADAPTER	pAd,
	IN	USHORT			Offset,
	IN	UINT32			Value)
{
	NTSTATUS	Status;
#ifdef BIG_ENDIAN		
	UINT32		Value_SWAP32;
#endif

#ifdef BIG_ENDIAN
	Value_SWAP32 = SWAP32(Value);

	if(Offset == 0x3040)
	{
		DBGPRINT(RT_DEBUG_TRACE, ("Value_SWAP32 = 0x%08x  Value = 0x%08x \n", Value_SWAP32, Value));	
	}

	Status = RTUSB_VendorRequest(
		pAd,
		USBD_TRANSFER_DIRECTION_OUT,
		DEVICE_VENDOR_REQUEST_OUT,
		0x6,
		0x00,
		Offset,
		&Value_SWAP32,
		4);	
#else
	Status = RTUSB_VendorRequest(
		pAd,
		USBD_TRANSFER_DIRECTION_OUT,
		DEVICE_VENDOR_REQUEST_OUT,
		0x6,
		0x00,
		Offset,
		&Value,
		4);
#endif	
	return Status;
}
#if 0
/*
	========================================================================
	
	Routine Description: Read 8-bit BBP register

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBReadBBPRegister(
	IN	PRTMP_ADAPTER	pAd,
	IN	UCHAR			Id,
	IN	PUCHAR			pValue)
{
	BBP_CSR_CFG_STRUC	BbpCsr;
//	UINT32			temp;
	UINT			i = 0;

	// Verify the busy condition
	do
	{
		RTUSBReadMACRegister(pAd, BBP_CSR_CFG, &BbpCsr.word);
		if (!(BbpCsr.field.Busy == BUSY))
			break;
		i++;
	}
	while ((i < RETRY_LIMIT) && (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)));
	
	if ((i == RETRY_LIMIT) || (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)))
	{
		//
		// Read failed then Return Default value.
		//
		*pValue = pAd->BbpWriteLatch[Id];
	
		DBGPRINT_RAW(RT_DEBUG_ERROR, ("Retry count exhausted or device removed!!!\n"));
		return STATUS_UNSUCCESSFUL;
	}

	// Prepare for write material
	BbpCsr.word 				= 0;
	BbpCsr.field.fRead			= 1;
	BbpCsr.field.Busy			= 1;
	BbpCsr.field.RegNum 		= Id;
	RTUSBWriteMACRegister(pAd, BBP_CSR_CFG, BbpCsr.word);

	i = 0;	
	// Verify the busy condition
	do
	{
		RTUSBReadMACRegister(pAd, BBP_CSR_CFG, &BbpCsr.word);
		if (!(BbpCsr.field.Busy == BUSY))
		{
			*pValue = (UCHAR)BbpCsr.field.Value;
			break;
		}
		i++;
	}
	while ((i < RETRY_LIMIT) && (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)));
	
	if ((i == RETRY_LIMIT) || (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)))
	{
		//
		// Read failed then Return Default value.
		//
		*pValue = pAd->BbpWriteLatch[Id];

		DBGPRINT_RAW(RT_DEBUG_ERROR, ("Retry count exhausted or device removed!!!\n"));
		return STATUS_UNSUCCESSFUL;
	}
	
	return STATUS_SUCCESS;
}
#else
/*
	========================================================================
	
	Routine Description: Read 8-bit BBP register via firmware

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBReadBBPRegister(
	IN	PRTMP_ADAPTER	pAd,
	IN	UCHAR			Id,
	IN	PUCHAR			pValue)
{																		
	BBP_CSR_CFG_STRUC	BbpCsr;											
	int					i, k;											
	for (i=0; i<MAX_BUSY_COUNT; i++)									
	{																	
		RTUSBReadMACRegister(pAd, H2M_BBP_AGENT, &BbpCsr.word);				
		if (BbpCsr.field.Busy == BUSY)									
		{																
			continue;													
		}																
		BbpCsr.word = 0;												
		BbpCsr.field.fRead = 1;											
		BbpCsr.field.BBP_RW_MODE = 1;									
		BbpCsr.field.Busy = 1;											
		BbpCsr.field.RegNum = Id;										
		RTUSBWriteMACRegister(pAd, H2M_BBP_AGENT, BbpCsr.word);				
		AsicSendCommandToMcu(pAd, 0x80, 0xff, 0x0, 0x0);					
		for (k=0; k<MAX_BUSY_COUNT; k++)								
		{																
			RTUSBReadMACRegister(pAd, H2M_BBP_AGENT, &BbpCsr.word);			
			if (BbpCsr.field.Busy == IDLE)								
				break;													
		}																
		if ((BbpCsr.field.Busy == IDLE) &&								
			(BbpCsr.field.RegNum == Id))								
		{																
			*pValue = (UCHAR)BbpCsr.field.Value;							
			break;														
		}																
	}
	if (BbpCsr.field.Busy == BUSY)										
	{																	
		DBGPRINT_ERR(("BBP read R%d=0x%x fail\n", Id, BbpCsr.word));	
		*pValue = pAd->BbpWriteLatch[Id];								
		return STATUS_UNSUCCESSFUL;
	}
	return STATUS_SUCCESS;
}
#endif

#if 0
/*
	========================================================================
	
	Routine Description: Write 8-bit BBP register

	Arguments:

	Return Value:
	
	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBWriteBBPRegister(
	IN	PRTMP_ADAPTER	pAd,
	IN	UCHAR			Id,
	IN	UCHAR			Value)
{
	BBP_CSR_CFG_STRUC	BbpCsr;
//	USHORT			temp;
	UINT			i = 0;

	// Verify the busy condition
	do
	{
		RTUSBReadMACRegister(pAd, BBP_CSR_CFG, &BbpCsr.word);
		if (!(BbpCsr.field.Busy == BUSY))
			break;
		i++;
	}
	while ((i < RETRY_LIMIT) && (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)));
	
	if ((i == RETRY_LIMIT) || (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)))
	{
		DBGPRINT_RAW(RT_DEBUG_ERROR, ("Retry count exhausted or device removed!!!\n"));
		return STATUS_UNSUCCESSFUL;
	}

	// Prepare for write material
	BbpCsr.word 				= 0;
	BbpCsr.field.fRead			= 0;
	BbpCsr.field.Value			= Value;
	BbpCsr.field.Busy			= 1;
	BbpCsr.field.RegNum 		= Id;
	RTUSBWriteMACRegister(pAd, BBP_CSR_CFG, BbpCsr.word);
	
	pAd->BbpWriteLatch[Id] = Value;

	return STATUS_SUCCESS;
}
#else
/*
	========================================================================
	
	Routine Description: Write 8-bit BBP register via firmware

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/

NTSTATUS	RTUSBWriteBBPRegister(
	IN	PRTMP_ADAPTER	pAd,
	IN	UCHAR			Id,
	IN	UCHAR			Value)

{																		
	BBP_CSR_CFG_STRUC	BbpCsr;											
	int					BusyCnt;										
	for (BusyCnt=0; BusyCnt<MAX_BUSY_COUNT; BusyCnt++)				
	{																
		RTMP_IO_READ32(pAd, H2M_BBP_AGENT, &BbpCsr.word);				
		if (BbpCsr.field.Busy == BUSY)									
			continue;												
		BbpCsr.word = 0;												
		BbpCsr.field.fRead = 0;											
		BbpCsr.field.BBP_RW_MODE = 1;									
		BbpCsr.field.Busy = 1;											
		BbpCsr.field.Value = Value;								
		BbpCsr.field.RegNum = Id;										
		RTMP_IO_WRITE32(pAd, H2M_BBP_AGENT, BbpCsr.word);				
		AsicSendCommandToMcu(pAd, 0x80, 0xff, 0x0, 0x0);				
		pAd->BbpWriteLatch[Id] = Value;									
		break;														
	}																
	if (BusyCnt == MAX_BUSY_COUNT)									
	{																
		DBGPRINT_ERR(("BBP write R%d=0x%x fail\n", Id, BbpCsr.word));	
		return STATUS_UNSUCCESSFUL;
	}									
	return STATUS_SUCCESS;
}
#endif
/*
	========================================================================
	
	Routine Description: Write RF register through MAC

	Arguments:

	Return Value:
	
	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBWriteRFRegister(
	IN	PRTMP_ADAPTER	pAd,
	IN	UINT32			Value)
{
	PHY_CSR4_STRUC	PhyCsr4;
	UINT			i = 0;

	do
	{
		RTUSBReadMACRegister(pAd, RF_CSR_CFG0, &PhyCsr4.word);
		if (!(PhyCsr4.field.Busy))
			break;
		i++;
	}
	while ((i < RETRY_LIMIT) && (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)));

	if ((i == RETRY_LIMIT) || (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)))
	{
		DBGPRINT_RAW(RT_DEBUG_ERROR, ("Retry count exhausted or device removed!!!\n"));
		return STATUS_UNSUCCESSFUL;
	}

	RTUSBWriteMACRegister(pAd, RF_CSR_CFG0, Value);
	
	return STATUS_SUCCESS;
}

/*
	========================================================================
	
	Routine Description:

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBReadEEPROM(
	IN	PRTMP_ADAPTER	pAd,
	IN	USHORT			Offset,
	OUT	PUCHAR			pData,
	IN	USHORT			length)
{
	NTSTATUS	Status = STATUS_SUCCESS;

	Status = RTUSB_VendorRequest(
		pAd,
		(USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK),
		DEVICE_VENDOR_REQUEST_IN,
		0x9,
		0,
		Offset,
		pData,
		length);
	
	return Status;
}

/*
	========================================================================
	
	Routine Description:

	Arguments:

	Return Value:
	
	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS	RTUSBWriteEEPROM(
	IN	PRTMP_ADAPTER	pAd,
	IN	USHORT			Offset,
	IN	PUCHAR			pData,
	IN	USHORT			length)
{
	NTSTATUS	Status = STATUS_SUCCESS;

	Status = RTUSB_VendorRequest(
		pAd,
		USBD_TRANSFER_DIRECTION_OUT,
		DEVICE_VENDOR_REQUEST_OUT,
		0x8,
		0,
		Offset,
		pData,
		length);
	
	return Status;
}

/*
	========================================================================
	
	Routine Description:

	Arguments:

	Return Value:
	
	IRQL = 
	
	Note:
	
	========================================================================
*/
VOID RTUSBPutToSleep(
	IN	PRTMP_ADAPTER	pAd)
{
	UINT32		value;
	
	// Timeout 0x40 x 50us
	value = (SLEEPCID<<16)+(OWNERMCU<<24)+ (0x40<<8)+1;
	RTUSBWriteMACRegister(pAd, 0x7010, value);
	RTUSBWriteMACRegister(pAd, 0x404, 0x30);
	//RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS);			
	DBGPRINT_RAW(RT_DEBUG_ERROR, ("Sleep Mailbox testvalue %x\n", value));
	
}

/*
	========================================================================
	
	Routine Description:

	Arguments:

	Return Value:
	
	IRQL = 
	
	Note:
	
	========================================================================
*/
NTSTATUS RTUSBWakeUp(
	IN	PRTMP_ADAPTER	pAd)
{
	NTSTATUS	Status;
	
	Status = RTUSB_VendorRequest(
		pAd,
		USBD_TRANSFER_DIRECTION_OUT,
		DEVICE_VENDOR_REQUEST_OUT,
		0x01,
		0x09,
		0,
		NULL,
		0);
	
	return Status;
}

/*
	========================================================================
	
	Routine Description:

	Arguments:

	Return Value:
	
	IRQL = 
	
	Note:
	
	========================================================================
*/
VOID	RTUSBInitializeCmdQ(
	IN	PCmdQ	cmdq)
{
	cmdq->head = NULL;
	cmdq->tail = NULL;
	cmdq->size = 0;
}

/*
	========================================================================
	
	Routine Description:

	Arguments:

	Return Value:
	
	IRQL = 
	
	Note:
	
	========================================================================
*/
NDIS_STATUS	RTUSBEnqueueCmdFromNdis(
	IN	PRTMP_ADAPTER	pAd,
	IN	NDIS_OID		Oid,
	IN	BOOLEAN			SetInformation,
	IN	PVOID			pInformationBuffer,
	IN	UINT32			InformationBufferLength)
{
	NDIS_STATUS	status;
	PCmdQElmt	cmdqelmt = NULL;
	unsigned long   IrqFlags;
	POS_COOKIE pObj = (POS_COOKIE) pAd->OS_Cookie;

	
	if (pObj->RTUSBCmdThr_pid < 0) 
		return (NDIS_STATUS_RESOURCES);
        
	status = RTMPAllocateMemory((PVOID *)&cmdqelmt, sizeof(CmdQElmt));
	if ((status != NDIS_STATUS_SUCCESS) || (cmdqelmt == NULL))
		return (NDIS_STATUS_RESOURCES);

	if ((Oid == RT_OID_MULTI_READ_MAC) ||
		(Oid == RT_OID_VENDOR_READ_BBP) ||
#ifdef DBG		
		(Oid == RT_OID_802_11_QUERY_HARDWARE_REGISTER) ||
#endif		
		(Oid == RT_OID_USB_VENDOR_EEPROM_READ) ||
		(Oid == OID_802_11_STATISTICS)	||
		(Oid == RT_OID_802_11_RSSI_1) ||
		(Oid == RT_OID_802_11_RSSI_2) )
	{
		cmdqelmt->buffer = pInformationBuffer;
	}
	else
	{
		cmdqelmt->buffer = NULL;
		if (pInformationBuffer != NULL)
		{
			status = RTMPAllocateMemory((PVOID *)&cmdqelmt->buffer, InformationBufferLength);
			if ((status != NDIS_STATUS_SUCCESS) || (cmdqelmt->buffer == NULL))
			{       
				return (NDIS_STATUS_RESOURCES);
			}
			else
			{
				NdisMoveMemory(cmdqelmt->buffer, pInformationBuffer, InformationBufferLength);
				cmdqelmt->bufferlength = InformationBufferLength;
			}
		}
		else
			cmdqelmt->bufferlength = 0;
	}
	
	cmdqelmt->command = Oid;
	cmdqelmt->CmdFromNdis = TRUE;
	if (SetInformation == TRUE)
		cmdqelmt->SetOperation = TRUE;
	else
		cmdqelmt->SetOperation = FALSE;

	NdisAcquireSpinLock(&pAd->CmdQLock, IrqFlags);
	EnqueueCmd((&pAd->CmdQ), cmdqelmt);
	NdisReleaseSpinLock(&pAd->CmdQLock, IrqFlags);
	
	RTUSBCMDUp(pAd);
	
	
    return(NDIS_STATUS_SUCCESS);
}

/*
	========================================================================
	
	Routine Description:

	Arguments:

	Return Value:

	IRQL = 
	
	Note:
	
	========================================================================
*/
NDIS_STATUS RTUSBEnqueueInternalCmd(
	IN PRTMP_ADAPTER	pAd,
	IN NDIS_OID			Oid,
	IN PVOID			pInformationBuffer,
	IN UINT32			InformationBufferLength)	
{
	NDIS_STATUS	status;
	PCmdQElmt	cmdqelmt = NULL;
	unsigned long	IrqFlags;
	
	DBGPRINT(RT_DEBUG_INFO, ("--->RTUSBEnqueueInternalCmd\n"));

	status = RTMPAllocateMemory((PVOID *)&cmdqelmt, sizeof(CmdQElmt));
	if ((status != NDIS_STATUS_SUCCESS) || (cmdqelmt == NULL))
		return (NDIS_STATUS_RESOURCES);

	NdisZeroMemory(cmdqelmt, sizeof(CmdQElmt));

	if(InformationBufferLength > 0)
	{
		status = RTMPAllocateMemory((PVOID *)&cmdqelmt->buffer, InformationBufferLength);
        
		if ((status != NDIS_STATUS_SUCCESS) || (cmdqelmt->buffer == NULL))
		{
			NdisFreeMemory(cmdqelmt, sizeof(CmdQElmt), 0);
			return (NDIS_STATUS_RESOURCES);
		}
		else
		{
			NdisMoveMemory(cmdqelmt->buffer, pInformationBuffer, InformationBufferLength);
			cmdqelmt->bufferlength = InformationBufferLength;
		}
	}
	else
	{
		cmdqelmt->buffer = NULL;
		cmdqelmt->bufferlength = 0;
	}

	cmdqelmt->command = Oid;
	cmdqelmt->CmdFromNdis = FALSE;

	if (cmdqelmt != NULL)
	{
		NdisAcquireSpinLock(&pAd->CmdQLock, IrqFlags);
		EnqueueCmd((&pAd->CmdQ), cmdqelmt);
		NdisReleaseSpinLock(&pAd->CmdQLock, IrqFlags);

		RTUSBCMDUp(pAd);
	}

	DBGPRINT(RT_DEBUG_INFO, ("<---RTUSBEnqueueInternalCmd\n"));
	return(NDIS_STATUS_SUCCESS);
}

/*
	========================================================================
	
	Routine Description:

	Arguments:

	Return Value:
	
	IRQL = 
	
	Note:
	
	========================================================================
*/
VOID	RTUSBDequeueCmd(
	IN	PCmdQ		cmdq,
	OUT	PCmdQElmt	*pcmdqelmt)
{
	*pcmdqelmt = cmdq->head;
	
	if (*pcmdqelmt != NULL)
	{
		cmdq->head = cmdq->head->next;
		cmdq->size--;
		if (cmdq->size == 0)
			cmdq->tail = NULL;
	}
}

/*
    ========================================================================
	  usb_control_msg - Builds a control urb, sends it off and waits for completion
	  @dev: pointer to the usb device to send the message to
	  @pipe: endpoint "pipe" to send the message to
	  @request: USB message request value
	  @requesttype: USB message request type value
	  @value: USB message value
	  @index: USB message index value
	  @data: pointer to the data to send
	  @size: length in bytes of the data to send
	  @timeout: time in jiffies to wait for the message to complete before
			  timing out (if 0 the wait is forever)
	  Context: !in_interrupt ()

	  This function sends a simple control message to a specified endpoint
	  and waits for the message to complete, or timeout.
	  If successful, it returns the number of bytes transferred, otherwise a negative error number.

	 Don't use this function from within an interrupt context, like a
	  bottom half handler.	If you need an asynchronous message, or need to send
	  a message from within interrupt context, use usb_submit_urb()
	  If a thread in your driver uses this call, make sure your disconnect()
	  method can wait for it to complete.  Since you don't have a handle on
	  the URB used, you can't cancel the request.
  
	
	Routine Description:

	Arguments:

	Return Value:
	
	Note:
	
	========================================================================
*/
NTSTATUS    RTUSB_VendorRequest(
	IN	PRTMP_ADAPTER	pAd,
	IN	UINT32			TransferFlags,
	IN	UCHAR			RequestType,
	IN	UCHAR			Request,
	IN	USHORT			Value,
	IN	USHORT			Index,
	IN	PVOID			TransferBuffer,
	IN	UINT32			TransferBufferLength)
{
	int				ret;
	POS_COOKIE		pObj = (POS_COOKIE) pAd->OS_Cookie;

	if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST))
	{
		DBGPRINT(RT_DEBUG_ERROR, ("device disconnected\n"));
		return -1;
	}
	else if (in_interrupt())	
	{
		DBGPRINT(RT_DEBUG_ERROR, ("in_interrupt, return RTUSB_VendorRequest\n"));

		return -1;
	}
	else
	{
	
		if( RequestType == DEVICE_VENDOR_REQUEST_OUT)
			ret=usb_control_msg(pObj->pUsb_Dev, usb_sndctrlpipe( pObj->pUsb_Dev, 0 ), Request, RequestType, Value,Index, TransferBuffer, TransferBufferLength, CONTROL_TIMEOUT_JIFFIES);
		else if(RequestType == DEVICE_VENDOR_REQUEST_IN)
			ret=usb_control_msg(pObj->pUsb_Dev, usb_rcvctrlpipe( pObj->pUsb_Dev, 0 ), Request, RequestType, Value,Index, TransferBuffer, TransferBufferLength, CONTROL_TIMEOUT_JIFFIES);
		else
		{
			DBGPRINT(RT_DEBUG_ERROR, ("vendor request direction is failed\n"));
			ret = -1;
		}

        if (ret < 0)
			DBGPRINT(RT_DEBUG_ERROR, ("USBVendorRequest failed ret=%d \n",ret));
			
#if 0	    
        // retry
		if (ret < 0) {
			int temp_i=0;
			DBGPRINT(RT_DEBUG_ERROR, ("USBVendorRequest failed ret=%d, \n",ret));
			ret = 0;
			do
			{
				if( RequestType == DEVICE_VENDOR_REQUEST_OUT)
					ret=usb_control_msg(pObj->pUsb_Dev, usb_sndctrlpipe( pObj->pUsb_Dev, 0 ), Request, RequestType, Value,Index, TransferBuffer, TransferBufferLength, CONTROL_TIMEOUT_JIFFIES);
				else if(RequestType == DEVICE_VENDOR_REQUEST_IN)
					ret=usb_control_msg(pObj->pUsb_Dev, usb_rcvctrlpipe( pObj->pUsb_Dev, 0 ), Request, RequestType, Value,Index, TransferBuffer, TransferBufferLength, CONTROL_TIMEOUT_JIFFIES);
				temp_i++;
			} while( (ret < 0) && (temp_i <= 1) );

			if( ret >= 0)
				return ret;
			
		}
#endif

	}
	return ret;
}

/*
	========================================================================
	
	Routine Description:
	  Creates an IRP to submite an IOCTL_INTERNAL_USB_RESET_PORT
	  synchronously. Callers of this function must be running at
	  PASSIVE LEVEL.

	Arguments:

	Return Value:

	Note:
	
	========================================================================
*/
NTSTATUS	RTUSB_ResetDevice(
	IN	PRTMP_ADAPTER	pAd)
{
	NTSTATUS		Status = TRUE;

	DBGPRINT_RAW(RT_DEBUG_TRACE, ("--->USB_ResetDevice\n"));
	//RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS);
	return Status;
}
                                                                                                                                                               
VOID CMDHandler(                                                                                                                                                
    IN PRTMP_ADAPTER pAd)                                                                                                                                       
{                                                                                                                                                               
	PCmdQElmt		cmdqelmt;                                                                                                                                       
	PUCHAR			pData;                                                                                                                                          
	NDIS_STATUS		NdisStatus = NDIS_STATUS_SUCCESS;                                                                                                               
	unsigned long	IrqFlags;                                                                                                                                   
	ULONG			Now;                                                                                                                                            
	NTSTATUS		ntStatus;
	
	while (pAd->CmdQ.size > 0)                                                                                                                                  
	{                                                                                                                                                           
		NdisStatus = NDIS_STATUS_SUCCESS;
		                                                                                                                      
		NdisAcquireSpinLock(&pAd->CmdQLock, IrqFlags);                                                                                                          
		RTUSBDequeueCmd(&pAd->CmdQ, &cmdqelmt);                                                                                                                 
		NdisReleaseSpinLock(&pAd->CmdQLock, IrqFlags);
		                                                                                                        
		if (cmdqelmt == NULL)                                                                                                                                   
			break; 
			                                                                                                                                             
		pData = cmdqelmt->buffer;                                                                                                                                                              
		DBGPRINT_RAW(RT_DEBUG_INFO, ("Cmd = %x\n", cmdqelmt->command));                                                                                         
		if((!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)) &&                                                                                               
			(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS)))                                                                                             
		{                                                                                                                                                       
			switch (cmdqelmt->command)                                                                                                                              
			{                                                                                                                                                       
				case RT_OID_CHECK_GPIO:
					{
						UINT32 data;                                                                                                                                     
						DBGPRINT_RAW(RT_DEBUG_ERROR, ("!!! RT_OID_CHECK_GPIO !!!\n"));                                                                              
        	                                                                                                                                                        
						// Read GPIO pin2 as Hardware controlled radio state

						RTUSBReadMACRegister( pAd, GPIO_CTRL_CFG, &data);

						if (data & 0x04)
						{                                                                                                                                               
							pAd->StaCfg.bHwRadio = TRUE;                                                                                                            
						} 
						else                                                                                                                                      
						{                                                                                                                                               
							pAd->StaCfg.bHwRadio = FALSE;                                                                                                           
						}                                                                                                                                               
        	                                                                                                                                                        
						if(pAd->StaCfg.bRadio != (pAd->StaCfg.bHwRadio && pAd->StaCfg.bSwRadio))                                                                    
						{                                                                                                                                               
							pAd->StaCfg.bRadio = (pAd->StaCfg.bHwRadio && pAd->StaCfg.bSwRadio);                                                                    
							if(pAd->StaCfg.bRadio == TRUE)                                                                                                          
							{                                                                                                                                           
								DBGPRINT_RAW(RT_DEBUG_ERROR, ("!!! Radio On !!!\n"));  
                                                                                                                            
								MlmeRadioOn(pAd);                                                                                                                       
								// Update extra information                                                                                                             
								pAd->ExtraInfo = EXTRA_INFO_CLEAR;                                                                                                      
							}                                                                                                                                           
							else                                                                                                                                        
							{                                                                                                                                           
								DBGPRINT_RAW(RT_DEBUG_ERROR, ("!!! Radio Off !!!\n"));
                                                                                                                                    
								MlmeRadioOff(pAd);                                                                                                                      
								// Update extra information                                                                                                             
								pAd->ExtraInfo = HW_RADIO_OFF;                                                                                                          
							}                                                                                                                                           
						}
					}
					break;                                                                                                                                              
        	                                                                                                                                                        
				case RT_OID_PERIODIC_EXECUT:                                                                                                                        
					{
						STAMlmePeriodicExec(pAd);
					}
					break;                                                                                                                                         
        	                                                                                                                                                        
				case OID_802_11_BSSID_LIST_SCAN:                                                                                                                    
					{                                                                                                                                                   
						if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)                                                                                               
						{                                                                                                                                               
							MlmeEnqueue(pAd,                                                                                                                            
						            	MLME_CNTL_STATE_MACHINE,                                                                                                        
						            	RT_CMD_RESET_MLME,                                                                                                              
						            	0,                                                                                                                              
						            	NULL);					                                                                                                        
						}                                                                                                                                               
        	                                                                                                                                                        
						NdisGetSystemUpTime(&Now);                                                                                                                                  
						// Reset Missed scan number                                                                                                                     
						pAd->StaCfg.ScanCnt = 0;                                                                                                                       
						pAd->StaCfg.LastScanTime = Now;                                                                                                                
						MlmeEnqueue(pAd,                                                                                                                                
									MLME_CNTL_STATE_MACHINE,                                                                                                            
									OID_802_11_BSSID_LIST_SCAN,                                                                                                         
									0,                                                                                                                                  
									NULL);                                                                                                                              
						RTUSBMlmeUp(pAd);                                                                                                                               
					}                                                                                                                                                   
					break;                                                                                                                                              
				                                                                                                                                                    
				case RT_OID_802_11_BSSID:               
					{                                                                                                                                                                                                                                                                                                
						if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)                                                                                               
						{                                                                                                                                               
							MlmeEnqueue(pAd,                                                                                                                            
						            	MLME_CNTL_STATE_MACHINE,                                                                                                        
						      			RT_CMD_RESET_MLME,                                                                                                              
						            	0,                                                                                                                              
						            	NULL);					                                                                                                        
						}                                                                                                                                               
        	                                                                                                                                                                	                                                                                                                                                        
						// Reset allowed scan retries                                                                                                                   
						pAd->StaCfg.ScanCnt = 0;                                                                                                                       
        	        	                                                                                                                                                
						MlmeEnqueue(pAd,                                                                                                                                
									MLME_CNTL_STATE_MACHINE,                                                                                                            
									RT_OID_802_11_BSSID,                                                                                                                   
									cmdqelmt->bufferlength,                                                                                                             
									cmdqelmt->buffer);                                                                                                                  
						RTUSBMlmeUp(pAd);                                                                                                                               
					}                                                                                                                                                   
					break;                                                                                                                                              
				                                                                                                                                                    
				case OID_802_11_SSID:                                                                                                                               
					{                                                                                                                                                   
						if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)                                                                                               
						{					                                                                                                                            
							MlmeEnqueue(pAd,                                                                                                                            
						            	MLME_CNTL_STATE_MACHINE,                                                                                                        
						            	RT_CMD_RESET_MLME,                                                                                                              
						            	0,                                                                                                                              
						            	NULL);					                                                                                                        
        	                                                                                                                                                        
						}                                                                                                                                               
        	                                                                                                                                                                	                                                                                                                                                        
						// Reset allowed scan retries                                                                                                                   
						pAd->StaCfg.ScanCnt = 0;                                                                                                                       
						pAd->bConfigChanged = TRUE;                                                                                                                     
        	                                                                                                                                                        
						MlmeEnqueue(pAd,                                                                                                                                
									MLME_CNTL_STATE_MACHINE,                                                                                                            
									OID_802_11_SSID,                                                                                                                    
									cmdqelmt->bufferlength,                                                                                                             
									pData);                                                                                                                             
						RTUSBMlmeUp(pAd);                                                                                                                               
					}                                                                                                                                                   
					break;                                                                                                                                              
        	                                                                                                                                                        
				case OID_802_11_DISASSOCIATE:                                                                                                                       
					{                                                                                                                                                   
						if (pAd->Mlme.CntlMachine.CurrState != CNTL_IDLE)                                                                                               
						{                                                                                                                                               
							MlmeEnqueue(pAd,                                                                                                                            
						            	MLME_CNTL_STATE_MACHINE,                                                                                                        
						            	RT_CMD_RESET_MLME,                                                                                                              
						            	0,                                                                                                                              
						            	NULL);					                                                                                                        
						}                                                                                                                                               
        	                                                                                                                                                        
						// Set to immediately send the media disconnect event                                                                                           
						pAd->MlmeAux.CurrReqIsFromNdis = TRUE;                                                                                                          
        	                                                                                                                                                        
						MlmeEnqueue(pAd,                                                                                                                                
									MLME_CNTL_STATE_MACHINE,                                                                                                            
									OID_802_11_DISASSOCIATE,                                                                                                            
									0,                                                                                                                                  
									NULL);                                                                                                                              
						RTUSBMlmeUp(pAd);                                                                                                                               
					}                                                                                                                                                   
					break;                                                                                                                                              
        	                                                                                                                                                        
				case RT_OID_QKERIODIC_EXECUT:
					{
						StaQuickResponeForRateUpExec(NULL, pAd, NULL, NULL);
					}
					break;
				case OID_802_11_RX_ANTENNA_SELECTED:
					break;

				case OID_802_11_TX_ANTENNA_SELECTED:
					break;
#if 0   	                                                                                                                                                        
				case RT_OID_802_11_QUERY_HARDWARE_REGISTER:
					NdisStatus = RTUSBQueryHardWareRegister(pAd, pData);
					break;

				case RT_OID_802_11_SET_HARDWARE_REGISTER:
					NdisStatus = RTUSBSetHardWareRegister(pAd, pData);
					break;
#endif  	                                                                                                                                                        
				case RT_OID_MULTI_READ_MAC:
					{
						PCMDHandler_TLV	p = (PCMDHandler_TLV)cmdqelmt->buffer;
						RTUSBMultiRead(pAd, p->Offset, &p->DataFirst, p->Length);
					}
					break;
        	                                                                                                                                                        
				case RT_OID_MULTI_WRITE_MAC:
					{
						PCMDHandler_TLV	p = (PCMDHandler_TLV)cmdqelmt->buffer;
						DBGPRINT_RAW(RT_DEBUG_TRACE, ("RT_OID_MULTI_WRITE_MAC Offset%04x Length%04x\n",p->Offset,p->Length));
						RTUSBMultiWrite(pAd, p->Offset, &p->DataFirst, p->Length);
					}
					break;
        	                                                                                                                                                        
				case RT_OID_USB_VENDOR_EEPROM_READ:
					{
						PCMDHandler_TLV	p = (PCMDHandler_TLV)cmdqelmt->buffer;
						RTUSBReadEEPROM(pAd, p->Offset, &p->DataFirst, p->Length);
					}
					break;
				                                                                                                                                                
				case RT_OID_USB_VENDOR_EEPROM_WRITE:
					{
						PCMDHandler_TLV	p = (PCMDHandler_TLV)cmdqelmt->buffer;
						pData = &p->DataFirst;
						while (p->Length > 62)
						{
							RTUSBWriteEEPROM(pAd, p->Offset, pData, 62);
							p->Offset += 62;
							p->Length -= 62;
							pData += 62;
						}
						RTUSBWriteEEPROM(pAd, p->Offset, pData, p->Length);
					}
					break;
                                                                                                                                                                
				case RT_OID_USB_VENDOR_ENTER_TESTMODE:
						RTUSB_VendorRequest(pAd,
											0,
											DEVICE_VENDOR_REQUEST_OUT,
											0x1,
											0x4,
											0x1,
											NULL,
											0);
						break;                                                                                                                                   
            	                                                                                                                                                    
					case RT_OID_USB_VENDOR_EXIT_TESTMODE:
						RTUSB_VendorRequest(pAd,
											0,
											DEVICE_VENDOR_REQUEST_OUT,
											0x1,
											0x4,
											0x0,
											NULL,
											0);
						break;

				case RT_OID_USB_RESET_BULK_OUT:                                                                                                                     
					{
						UINT32		MACValue;
						UCHAR		Index;
						int			ret=0;
						PHT_TX_CONTEXT	pTxContext;
						unsigned long IrqFlag;

						DBGPRINT_RAW(RT_DEBUG_ERROR, ("CmdThread : RT_OID_USB_RESET_BULK_OUT===>\n"));
						// All transfers must be aborted or cancelled before attempting to reset the pipe.						
						//RTUSBCancelPendingBulkOutIRP(pAd);
						// Wait 10ms to let previous packet that are already in HW FIFO to clear. by MAXLEE 12-25-2007
						Index = 0;
						do 
						{
							RTUSBReadMACRegister(pAd, TXRXQ_PCNT, &MACValue);
							if ((MACValue & 0x800000) == 0)
								break;
							Index++;
							NdisMSleep(5000);
						}while(Index < 10);
						MACValue = 0;
						RTUSBReadMACRegister(pAd, USB_DMA_CFG, &MACValue);
						// To prevent Read Register error, we 2nd check the validity.
						if ((MACValue & 0xc00000) == 0)
							RTUSBReadMACRegister(pAd, USB_DMA_CFG, &MACValue);
						// To prevent Read Register error, we 3rd check the validity.
						if ((MACValue & 0xc00000) == 0)
							RTUSBReadMACRegister(pAd, USB_DMA_CFG, &MACValue);
						MACValue |= 0x80000;
						RTUSBWriteMACRegister(pAd, USB_DMA_CFG, MACValue);
						MACValue &= (~0x80000);
						RTUSBWriteMACRegister(pAd, USB_DMA_CFG, MACValue);
						DBGPRINT_RAW(RT_DEBUG_ERROR, ("Set 0x2a0 bit19. Clear USB DMA TX path\n"));
						
						// Wait 10ms to prevent next URB to bulkout before HW reset. by MAXLEE 12-25-2007
						NdisMSleep(5000);

						NdisAcquireSpinLock(&pAd->BulkOutLock[pAd->bulkResetPipeid], IrqFlag);

						if ( pAd->BulkOutPending[pAd->bulkResetPipeid] == FALSE)
						{
							pAd->BulkOutPending[pAd->bulkResetPipeid] = TRUE; 	
							NdisReleaseSpinLock(&pAd->BulkOutLock[pAd->bulkResetPipeid], IrqFlag);
							pTxContext = &(pAd->TxContext[pAd->bulkResetPipeid]);	
							pTxContext->IRPPending = TRUE;
							RTUSBInitHTTxDesc(pAd, pTxContext, pAd->bulkResetPipeid, pTxContext->BulkOutSize, RTUSBBulkOutDataPacketComplete);							

							if((ret = RTUSB_SUBMIT_URB(pTxContext->pUrb))!=0)
							{
								DBGPRINT(RT_DEBUG_ERROR, ("RTUSBBulkOutDataPacket: Submit Tx URB failed %d\n", ret));
							}								
						}
						else
						{
							DBGPRINT_RAW(RT_DEBUG_ERROR, ("TX RECOVER FAIL\n"));
							NdisReleaseSpinLock(&pAd->BulkOutLock[pAd->bulkResetPipeid], IrqFlag);
						} 
						RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);

						/*for (Index = 0; Index < 4; Index++)
						{
							if(pAd->SendTxWaitQueue[Index].Number > 0)
							{
								RTMPDeQueuePacket(pAd, Index);
							}
						}
						RTUSBKickBulkOut(pAd);*/
					}
					/*
						// Don't cancel BULKIN.	
						while ((atomic_read(&pAd->PendingRx) > 0) && 
								(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST))) 
						{
							if (atomic_read(&pAd->PendingRx) > 0)
							{
								DBGPRINT_RAW(RT_DEBUG_ERROR, ("BulkIn IRP Pending!!cancel it!\n"));
								RTUSBCancelPendingBulkInIRP(pAd);
							}
							NdisMSleep(100000);
						}
						
						if ((atomic_read(&pAd->PendingRx) == 0) && (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS)))
						{
							UCHAR	i;
							RTUSBRxPacket(pAd);
							pAd->NextRxBulkInReadIndex = 0;	// Next Rx Read index
							pAd->NextRxBulkInIndex		= 0;	// Rx Bulk pointer
							for (i = 0; i < (RX_RING_SIZE); i++)
							{
								PRX_CONTEXT  pRxContext = &(pAd->RxContext[i]);

								pRxContext->pAd	= pAd;
								pRxContext->InUse		= FALSE;
								pRxContext->IRPPending	= FALSE;
								pRxContext->Readable	= FALSE;
								pRxContext->ReorderInUse = FALSE;
							
							}
							RTUSBBulkReceive(pAd);
							DBGPRINT_RAW(RT_DEBUG_ERROR, ("RTUSBBulkReceive\n"));
						}*/
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("CmdThread : RT_OID_USB_RESET_BULK_OUT<===\n"));
    	   			break;                                                                                                                                              
            	                                                                                                                                                    
				case RT_OID_USB_RESET_BULK_IN:                                                                                                                      
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("CmdThread : RT_OID_USB_RESET_BULK_IN === >\n"));

					// All transfers must be aborted or cancelled before attempting to reset the pipe.
					{
						UINT32		MACValue;
						while ((atomic_read(&pAd->PendingRx) > 0) && (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST))) 
						{
							if (atomic_read(&pAd->PendingRx) > 0)
							{
								DBGPRINT_RAW(RT_DEBUG_ERROR, ("BulkIn IRP Pending!!!\n"));
								RTUSBCancelPendingBulkInIRP(pAd);
							}
							NdisMSleep(100000);
						}
						// Wait 10ms before reading register.
						NdisMSleep(10000);
						ntStatus = RTUSBReadMACRegister(pAd, MAC_CSR0, &MACValue);

						if ((NT_SUCCESS(ntStatus) == TRUE)&& (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS)) &&
							(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF)) &&
							(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS)) &&
							(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)))
						{		
							UCHAR	i;
							pAd->NextRxBulkInPosition = pAd->RxContext[pAd->NextRxBulkInReadIndex].pUrb->actual_length;
							pAd->RxContext[pAd->NextRxBulkInReadIndex].ReadPosOffset = pAd->NextRxBulkInPosition;
							if (pAd->NextRxBulkInPosition == 0)
								pAd->NextRxBulkInPosition = MAX_RXBULK_SIZE;
							DBGPRINT_RAW(RT_DEBUG_ERROR, ("0x%x/0x%x/ 0x%lx. ::x%lx/0x%lx/ 0x%lx.n", pAd->NextRxBulkInIndex,  pAd->NextRxBulkInReadIndex, pAd->NextRxBulkInPosition, pAd->BulkInReq, pAd->BulkInComplete, pAd->BulkInCompleteFail));
 							DBGPRINT_RAW(RT_DEBUG_ERROR, ("[0] IRPPending=%d:  InUse=%d: Readable=%d: [1] IRPPending=%d:  InUse=%d: Readable=%d: [2 IRPPending=%d:  InUse=%d: Readable=%d: [3] IRPPending=%d:  InUse=%d: Readable=%d: \n"
								, pAd->RxContext[0].IRPPending, pAd->RxContext[0].InUse, pAd->RxContext[0].Readable
								, pAd->RxContext[1].IRPPending, pAd->RxContext[1].InUse, pAd->RxContext[1].Readable
								, pAd->RxContext[2].IRPPending, pAd->RxContext[2].InUse, pAd->RxContext[2].Readable
								, pAd->RxContext[3].IRPPending, pAd->RxContext[3].InUse, pAd->RxContext[3].Readable));
							DBGPRINT_RAW(RT_DEBUG_ERROR, ("[4] IRPPending=%d:  InUse=%d: Readable=%d: [5] IRPPending=%d:  InUse=%d: Readable=%d:[6] IRPPending=%d:  InUse=%d: Readable=%d: [7] IRPPending=%d:  InUse=%d: Readable=%d: \n"
								, pAd->RxContext[4].IRPPending, pAd->RxContext[4].InUse, pAd->RxContext[4].Readable
								, pAd->RxContext[5].IRPPending, pAd->RxContext[5].InUse, pAd->RxContext[5].Readable
								, pAd->RxContext[6].IRPPending, pAd->RxContext[6].InUse, pAd->RxContext[6].Readable
								, pAd->RxContext[7].IRPPending, pAd->RxContext[7].InUse, pAd->RxContext[7].Readable));
 							/*

							DBGPRINT_RAW(RT_DEBUG_ERROR, ("==========================================\n"));

							pAd->NextRxBulkInReadIndex = 0;	// Next Rx Read index
							pAd->NextRxBulkInIndex		= 0;	// Rx Bulk pointer
							for (i = 0; i < (RX_RING_SIZE); i++)
							{
								PRX_CONTEXT  pRxContext = &(pAd->RxContext[i]);

								pRxContext->pAd	= pAd;
								pRxContext->InUse		= FALSE;
								pRxContext->IRPPending	= FALSE;
								pRxContext->Readable	= FALSE;
								pRxContext->ReorderInUse = FALSE;
									
							}*/
							RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET);
							for (i = 0; i < pAd->CommonCfg.NumOfBulkInIRP; i ++)
							{
								RTUSBBulkReceive(pAd);
								DBGPRINT_RAW(RT_DEBUG_ERROR, ("RTUSBBulkReceive\n"));
							}					
						}
						else
						{
							// Card must be removed
							RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST);
							DBGPRINT_RAW(RT_DEBUG_ERROR, ("Read Register Failed!Card must be removed,\n"));
							DBGPRINT_RAW(RT_DEBUG_ERROR, ("\n"));
						}
					}                                                                                                                                                   
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("CmdThread : RT_OID_USB_RESET_BULK_IN <===\n"));
					break;                                                                                                                                              
            	                                                                                                                                                    
				case RT_OID_802_11_SET_STA_CONFIG:
					{
						RT_802_11_STA_CONFIG *pStaConfig = (RT_802_11_STA_CONFIG *)pData;
						if (pStaConfig->EnableTxBurst != pAd->CommonCfg.bEnableTxBurst)
						{
							pAd->CommonCfg.bEnableTxBurst = (pStaConfig->EnableTxBurst == 1);
							//Currently Tx burst mode is only implemented in infrastructure mode.
							if (INFRA_ON(pAd))
							{
								if (pAd->CommonCfg.bEnableTxBurst)
								{
									//Extend slot time if any encryption method is used to give ASIC more time to do encryption/decryption during Tx burst mode.
									if (pAd->CommonCfg.WepStatus != Ndis802_11EncryptionDisabled)
									{
									}
									//Set CWmin/CWmax to 0.
								}
								else
								{
									if (pAd->CommonCfg.WepStatus != Ndis802_11EncryptionDisabled)
										AsicSetSlotTime(pAd, (BOOLEAN)pAd->CommonCfg.bUseShortSlotTime);
								}
							}
						}
						//pAd->CommonCfg.EnableTurboRate = pStaConfig->EnableTurboRate;
						pAd->CommonCfg.UseBGProtection = pStaConfig->UseBGProtection;
						//pAd->CommonCfg.bUseShortSlotTime = pStaConfig->UseShortSlotTime;
						pAd->CommonCfg.bUseShortSlotTime = 1; // 2003-10-30 always SHORT SLOT capable
						if (pAd->StaCfg.AdhocMode != pStaConfig->AdhocMode)
						{
							// allow dynamic change of "USE OFDM rate or not" in ADHOC mode
							// if setting changed, need to reset current TX rate as well as BEACON frame format
							pAd->StaCfg.AdhocMode = pStaConfig->AdhocMode;
							if (pAd->StaCfg.BssType == BSS_ADHOC)
							{
								MlmeUpdateTxRates(pAd, FALSE);
								MakeIbssBeacon(pAd);
							}
						}	
						// Add for Monitor mode support
						pAd->bPromiscuous = ((pStaConfig->Rsv1&0x1) == 1) ? TRUE : FALSE;
						if (pAd->bPromiscuous)
						{
							UINT32	Value;
							UCHAR	bbpValue;
							BCN_TIME_CFG_STRUC csr;					
							OPSTATUS_SET_FLAG(pAd, fOP_STATUS_INFRA_ON);
							OPSTATUS_SET_FLAG(pAd, fOP_STATUS_MEDIA_STATE_CONNECTED);
#ifdef WIN_NDIS
							pAd->IndicateMediaState = NdisMediaStateConnected;
							NdisMIndicateStatus(pAd->AdapterHandle, NDIS_STATUS_MEDIA_CONNECT, (PVOID)NULL, 0);
							NdisMIndicateStatusComplete(pAd->AdapterHandle);
#endif							
							// disable all periodic state machine
							pAd->StaCfg.bAutoReconnect = FALSE;
							// reset all mlme state machine
							MlmeRestartStateMachine(pAd);
							// Enable Rx with promiscuous reception
							RTUSBWriteMACRegister(pAd, RX_FILTR_CFG, 0x0);
							// ASIC supporsts sniffer function with replacing RSSI with timestamp.
							RTUSBReadMACRegister(pAd, MAC_SYS_CTRL, &Value);
							Value |= (0x80);
							RTUSBWriteMACRegister(pAd, MAC_SYS_CTRL, Value);
							// disable sync
							RTUSBReadMACRegister(pAd, BCN_TIME_CFG, &csr.word);
							csr.field.bBeaconGen = 0;
							csr.field.bTBTTEnable = 0;
							csr.field.TsfSyncMode = 0;
							RTUSBWriteMACRegister(pAd, BCN_TIME_CFG, csr.word);
							if ((pStaConfig->Rsv1&0x2) == 2)
							{
								// 40MHz
								RTUSBReadBBPRegister(pAd, BBP_R4, &bbpValue);
								bbpValue &= (~0x18);
								bbpValue |= 0x10;
								RTUSBWriteBBPRegister(pAd, BBP_R4, bbpValue);
								pAd->CommonCfg.BBPCurrentBW = BW_40;
								DBGPRINT(RT_DEBUG_TRACE, ("BW_40 \n"));
							}
							else
							{
								// 40MHz
								RTUSBReadBBPRegister(pAd, 4, &bbpValue);
								bbpValue &= (~0x18);
								RTUSBWriteBBPRegister(pAd, 4, bbpValue);
								pAd->CommonCfg.BBPCurrentBW = BW_20;
								DBGPRINT(RT_DEBUG_TRACE, ("BW_20 \n"));
							}
						
							DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::RT_OID_802_11_SET_STA_CONFIG (0x1004 = 0x%x\n",Value));
					
						}
						else
						{
							UINT32	Value;
							UCHAR	bbpValue;
							RTUSBWriteMACRegister(pAd, RX_FILTR_CFG, STANORMAL);
							RTUSBReadMACRegister(pAd, MAC_SYS_CTRL, &Value);
							Value &= (~0x80);
							RTUSBWriteMACRegister(pAd, MAC_SYS_CTRL, Value);
							RTUSBReadBBPRegister(pAd, 4, &bbpValue);
							bbpValue &= (~0x18);
							RTUSBWriteBBPRegister(pAd, 4, bbpValue);
							pAd->CommonCfg.BBPCurrentBW = BW_20;
							DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::RT_OID_802_11_SET_STA_CONFIG (0x1004 = 0x%x\n",Value));
						}
						DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::RT_OID_802_11_SET_STA_CONFIG (bPromiscuous=%d,Burst=%ld,BGprot=%ld,ShortSlot=%ld,Adhoc=%ld,Protection=%ld\n",
							pAd->bPromiscuous,
							pStaConfig->EnableTxBurst,
							pStaConfig->UseBGProtection,
							pStaConfig->UseShortSlotTime,
							pStaConfig->AdhocMode,
							pAd->CommonCfg.UseBGProtection));
					}
					break;
					
				case OID_802_11_CONFIGURATION:
					if (pAd->bPromiscuous == TRUE)
					{
						AsicSwitchChannel(pAd, pAd->CommonCfg.Channel,FALSE);
						AsicLockChannel(pAd, pAd->CommonCfg.Channel);
						DBGPRINT_RAW(RT_DEBUG_ERROR, ("bPromiscuous( switch channel = %d)\n", pAd->CommonCfg.Channel));
					}
					break;
					
				case RT_OID_SET_PSM_BIT_SAVE:
					MlmeSetPsm(pAd, PWR_SAVE);
					RTMPSendNullFrame(pAd, pAd->CommonCfg.TxRate);
					break;

				case RT_OID_SET_RADIO:
					if (pAd->StaCfg.bRadio == TRUE)
					{
						MlmeRadioOn(pAd);
						// Update extra information
						pAd->ExtraInfo = EXTRA_INFO_CLEAR;
					}
					else
					{
						MlmeRadioOff(pAd);
						// Update extra information
						pAd->ExtraInfo = SW_RADIO_OFF;
					}
					break;
                                                                                                                                          
				case RT_OID_RESET_FROM_ERROR:                                                                                                                       
				case RT_OID_RESET_FROM_NDIS:                                                                                                                        
					/*{
						UINT	i = 0;                                                                                                                                  
            	                                                                                                                                                    
						RTUSBRejectPendingPackets(pAd);//reject all NDIS packets waiting in TX queue                                                                    
						RTUSBCleanUpDataBulkOutQueue(pAd);                                                                                                              
						MlmeSuspend(pAd, FALSE);                                                                                                                        
            	                                                                                                                                                    
						//Add code to access necessary registers here.                                                                                                  
						//disable Rx                                                                                                                                    
						RTUSBWriteMACRegister(pAd, TXRX_CSR2, 1);                                                                                                       
						//Ask our device to complete any pending bulk in IRP.                                                                                           
						while ((atomic_read(&pAd->PendingRx) > 0) ||                                                                                                    
								(pAd->BulkOutPending[0] == TRUE) ||                                                                                                      
						   		(pAd->BulkOutPending[1] == TRUE) ||                                                                                                      
						   		(pAd->BulkOutPending[2] == TRUE) ||                                                                                                      
						   		(pAd->BulkOutPending[3] == TRUE))                                                                                                        
            	                                                                                                                                                    
						{                                                                                                                                               
							if (atomic_read(&pAd->PendingRx) > 0)                                                                                                       
							{                                                                                                                                           
								DBGPRINT_RAW(RT_DEBUG_TRACE, "BulkIn IRP Pending!!!\n");                                                                                
								RTUSB_VendorRequest(pAd,                                                                                                                
													0,                                                                                                                  
													DEVICE_VENDOR_REQUEST_OUT,                                                                                          
													0x0C,                                                                                                               
													0x0,                                                                                                                
													0x0,                                                                                                                
													NULL,                                                                                                               
													0);                                                                                                                 
							}                                                                                                                                           
            	        	                                                                                                                                            
							if ((pAd->BulkOutPending[0] == TRUE) ||                                                                                                     
							(pAd->BulkOutPending[1] == TRUE) ||                                                                                                     
							(pAd->BulkOutPending[2] == TRUE) ||                                                                                                     
							(pAd->BulkOutPending[3] == TRUE))                                                                                                       
							{                                                                                                                                           
								DBGPRINT_RAW(RT_DEBUG_TRACE, ("BulkOut IRP Pending!!!\n"));
								if (i == 0)                                                                                                                             
								{                                                                                                                                       
									RTUSBCancelPendingBulkOutIRP(pAd);                                                                                                  
									i++;                                                                                                                                
								}                                                                                                                                       
							}                                                                                                                                           
							NdisMSleep(500000);
						}                                                                                                                                               
            	                                                                                                                                                    
						NICResetFromError(pAd);                                                                                                                         
						if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HARDWARE_ERROR))                                                                                          
						{                                                                                                                                               
							RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_HARDWARE_ERROR);                                                                                         
						}                                                                                                                                               
						if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET))                                                                                            
						{                                                                                                                                               
							RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET);                                                                                           
						}                                                                                                                                               
						if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET))                                                                                           
						{                                                                                                                                               
							RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);                                                                                          
						}                                                                                                                                               
            	                                                                                                                                                    
						RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS);                                                                                          
            	                                                                                                                                                    
						if ((!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF)) &&                                                                                          
							(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS)) &&                                                                                   
							(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)))                                                                                        
						{                                                                                                                                               
							MlmeResume(pAd);                                                                                                                            
							//
							// Support multiple BulkIn IRP, the value on pAd->CommonCfg.NumOfBulkInIRP may be large than 1.
							//
							for(i = 0; i < pAd->CommonCfg.NumOfBulkInIRP; i ++)
							{
								RTUSBBulkReceive(pAd);
							}      
							RTUSBWriteMACRegister(pAd, TXRX_CSR2, 0x7e);                                                                                                
						}                                                                                                                                               
					}*/                                                                                                                                                   
					break;                                                                                                                                              
            	                                                                                                                                                    
				case RT_OID_LINK_DOWN:                                                                                                                              
					DBGPRINT_RAW(RT_DEBUG_TRACE, ("LinkDown(RT_OID_LINK_DOWN)\n"));                                                                                   
					LinkDown(pAd, TRUE);                                                                                                                            
					break;                                                                                                                                              
            	                                                                                                                                                                	                                                                                                                                                  
				case RT_OID_VENDOR_WRITE_BBP:
					{
						UCHAR	Offset, Value, bitmask, BBPValue;
						Offset = *((PUCHAR)pData);
						Value = *((PUCHAR)(pData + 1));
						bitmask = *((PUCHAR)(pData + 2));
						DBGPRINT_RAW(RT_DEBUG_INFO, ("offset = 0x%02x	value = 0x%02x\n", Offset, Value));
						if (bitmask != 0xff)
						{
							RTUSBReadBBPRegister(pAd, Offset, &BBPValue);
							BBPValue&= (~bitmask);
							BBPValue |= Value;
							RTUSBWriteBBPRegister(pAd, Offset, BBPValue);
						}
						else
						{
							RTUSBWriteBBPRegister(pAd, Offset, Value);
						}
					}
					break;
					
				case RT_CMD_SET_ASIC_WCID:
					{
						RT_SET_ASIC_WCID	SetAsicWcid;
						USHORT		offset;
						UINT32		MACValue, MACRValue = 0;
						SetAsicWcid = *((PRT_SET_ASIC_WCID)(pData));
						
						if (SetAsicWcid.WCID >= MAX_LEN_OF_MAC_TABLE)
							return;
						
						offset = MAC_WCID_BASE + ((UCHAR)SetAsicWcid.WCID)*HW_WCID_ENTRY_SIZE;

						DBGPRINT_RAW(RT_DEBUG_TRACE, ("CmdThread : RT_CMD_SET_ASIC_WCID : WCID = %ld, SetTid  = %lx, DeleteTid = %lx.\n", SetAsicWcid.WCID, SetAsicWcid.SetTid, SetAsicWcid.DeleteTid));
						MACValue = (pAd->MacTab.Content[SetAsicWcid.WCID].Addr[3]<<24)+(pAd->MacTab.Content[SetAsicWcid.WCID].Addr[2]<<16)+(pAd->MacTab.Content[SetAsicWcid.WCID].Addr[1]<<8)+(pAd->MacTab.Content[SetAsicWcid.WCID].Addr[0]);
						DBGPRINT_RAW(RT_DEBUG_TRACE, ("1-MACValue= %x,\n", MACValue));
						RTUSBWriteMACRegister(pAd, offset, MACValue);
						// Read bitmask
						RTUSBReadMACRegister(pAd, offset+4, &MACRValue);
						if ( SetAsicWcid.DeleteTid != 0xffffffff)
							MACRValue &= (~SetAsicWcid.DeleteTid);
						if (SetAsicWcid.SetTid != 0xffffffff)
							MACRValue |= (SetAsicWcid.SetTid);
						MACRValue &= 0xffff0000;
					
						MACValue = (pAd->MacTab.Content[SetAsicWcid.WCID].Addr[5]<<8)+pAd->MacTab.Content[SetAsicWcid.WCID].Addr[4];
						MACValue |= MACRValue;
						RTUSBWriteMACRegister(pAd, offset+4, MACValue);
						
						DBGPRINT_RAW(RT_DEBUG_TRACE, ("2-MACValue= %x,\n", MACValue));
					}
					break;
					
				case RT_CMD_SET_ASIC_WCID_CIPHER:
					{
						RT_SET_ASIC_WCID_ATTRI	SetAsicWcidAttri;
						USHORT		offset;
						UINT32		MACRValue = 0;
						SHAREDKEY_MODE_STRUC csr1;
						SetAsicWcidAttri = *((PRT_SET_ASIC_WCID_ATTRI)(pData));
						
						if (SetAsicWcidAttri.WCID >= MAX_LEN_OF_MAC_TABLE)
							return;
						
						offset = MAC_WCID_ATTRIBUTE_BASE + ((UCHAR)SetAsicWcidAttri.WCID)*HW_WCID_ATTRI_SIZE;

						DBGPRINT_RAW(RT_DEBUG_TRACE, ("Cmd : RT_CMD_SET_ASIC_WCID_CIPHER : WCID = %ld, Cipher = %lx.\n", SetAsicWcidAttri.WCID, SetAsicWcidAttri.Cipher));
						// Read bitmask
						RTUSBReadMACRegister(pAd, offset, &MACRValue);
						MACRValue = 0;
						MACRValue |= (((UCHAR)SetAsicWcidAttri.Cipher) << 1);

						RTUSBWriteMACRegister(pAd, offset, MACRValue);
						DBGPRINT_RAW(RT_DEBUG_TRACE, ("2-offset = %x , MACValue= %x,\n", offset, MACRValue));
						
						offset = PAIRWISE_IVEIV_TABLE_BASE + ((UCHAR)SetAsicWcidAttri.WCID)*HW_IVEIV_ENTRY_SIZE;
						MACRValue = 0;							
						if ( (SetAsicWcidAttri.Cipher <= CIPHER_WEP128))
							MACRValue |= ( pAd->CommonCfg.DefaultKeyId << 30);
						else
							MACRValue |= (0x20000000);
						RTUSBWriteMACRegister(pAd, offset, MACRValue);
						DBGPRINT_RAW(RT_DEBUG_TRACE, ("2-offset = %x , MACValue= %x,\n", offset, MACRValue));

						//
						// Update cipher algorithm. WSTA always use BSS0							
						//
						// for adhoc mode only ,because wep status slow than add key, when use zero config
						if (pAd->StaCfg.BssType == BSS_ADHOC )
						{
							offset = MAC_WCID_ATTRIBUTE_BASE;

							RTUSBReadMACRegister(pAd, offset, &MACRValue);
							MACRValue &= (~0xe);
							MACRValue |= (((UCHAR)SetAsicWcidAttri.Cipher) << 1);

							RTUSBWriteMACRegister(pAd, offset, MACRValue);

							//Update group key cipher,,because wep status slow than add key, when use zero config
							RTUSBReadMACRegister(pAd, SHARED_KEY_MODE_BASE+4*(0/2), &csr1.word);
								
							csr1.field.Bss0Key0CipherAlg = SetAsicWcidAttri.Cipher;
							csr1.field.Bss0Key1CipherAlg = SetAsicWcidAttri.Cipher;
																
							RTUSBWriteMACRegister(pAd, SHARED_KEY_MODE_BASE+4*(0/2), csr1.word);
						}
					}
					break;
				case RT_OID_SET_CLIENT_MAC_ENTRY:
					{
						MAC_TABLE_ENTRY *pEntry;
						pEntry = (MAC_TABLE_ENTRY *)pData;
						if ((pEntry->AuthMode <= Ndis802_11AuthModeAutoSwitch) && (pEntry->WepStatus == Ndis802_11Encryption1Enabled))
						{
							UINT32 uIV = 0;
							PUCHAR  ptr;

							ptr = (PUCHAR) &uIV;
							*(ptr + 3) = (pAd->CommonCfg.DefaultKeyId << 6);
							AsicUpdateWCIDIVEIV(pAd, pEntry->Aid, uIV, 0);
							AsicUpdateWCIDAttribute(pAd, pEntry->Aid, BSS0, pAd->SharedKey[BSS0][pAd->CommonCfg.DefaultKeyId].CipherAlg, FALSE);
						}
						else if (pEntry->AuthMode == Ndis802_11AuthModeWPANone)
						{
							UINT32 uIV = 0;
							PUCHAR  ptr;

							ptr = (PUCHAR) &uIV;
							*(ptr + 3) = (pAd->CommonCfg.DefaultKeyId << 6);
							AsicUpdateWCIDIVEIV(pAd, pEntry->Aid, uIV, 0);
							AsicUpdateWCIDAttribute(pAd, pEntry->Aid, BSS0, pAd->SharedKey[BSS0][pAd->CommonCfg.DefaultKeyId].CipherAlg, FALSE);
						}
						else
						{
							//
							// Other case, disable engine.
							// Don't worry WPA key, we will add WPA Key after 4-Way handshaking.
							//
							USHORT   offset;
							offset = MAC_WCID_ATTRIBUTE_BASE + (pEntry->Aid * HW_WCID_ATTRI_SIZE);
							// RX_PKEY_MODE:0 for no security; RX_KEY_TAB:0 for shared key table; BSS_IDX:0
							RTUSBWriteMACRegister(pAd, offset, 0);


						}

#if 0
						// Add this pAddr into ASIC WCID Table. No LEVEL issue, so write register here.
						offset = MAC_WCID_BASE + (i * HW_WCID_ENTRY_SIZE);	
						addr = (pEntry->Addr[0]) + (pEntry->Addr[1]<<8) +(pEntry->Addr[2]<<16) +(pEntry->Addr[3]<<24);
						RTMP_IO_WRITE32(pAd, offset, addr);
						addr = (pEntry->Addr[4]) + (pEntry->Addr[5]<<8);
						RTMP_IO_WRITE32(pAd, offset+4, addr);
#endif
						AsicUpdateRxWCIDTable(pAd, pEntry->Aid, pEntry->Addr);
					}
					break;
					
				case RT_CMD_SET_BW:
					{
						UCHAR	Value, extoffset, centralchannel;
						UCHAR	BBPValue;
						UINT32	MACValue;
						Value = *((PUCHAR)(pData));
						extoffset = *((PUCHAR)(pData+1));
						centralchannel = *((PUCHAR)(pData+2));
						RTUSBReadMACRegister(pAd, TX_BAND_CFG, &MACValue);
						RTUSBReadBBPRegister(pAd, BBP_R3, &BBPValue);
						if ((extoffset == EXTCHA_BELOW))
						{
							MACValue |= 0x1;
							BBPValue |= 0x20;
							RTUSBWriteMACRegister(pAd, TX_BAND_CFG, MACValue);
							RTUSBWriteBBPRegister(pAd, BBP_R3, BBPValue);
							DBGPRINT_RAW(RT_DEBUG_TRACE, (" extoffset == EXTCHA_BELOW\n"));
						}
						else
						{
							MACValue &= 0xfe;
							BBPValue &= (~0x20);
							RTUSBWriteMACRegister(pAd, TX_BAND_CFG, MACValue);
							RTUSBWriteBBPRegister(pAd, BBP_R3, BBPValue);
							DBGPRINT_RAW(RT_DEBUG_TRACE, (" extoffset == EXTCHA_ABOVE\n"));
						} 
						if ((Value%0xf) == BW_40)
						{
							RTUSBReadBBPRegister(pAd, BBP_R4, &BBPValue);
							BBPValue&= (~0x18);
							BBPValue|= (0x10);
							RTUSBWriteBBPRegister(pAd, BBP_R4, BBPValue);
							pAd->CommonCfg.BBPCurrentBW = BW_40;
							AsicSwitchChannel(pAd, centralchannel,FALSE);
							AsicLockChannel(pAd, centralchannel);
							DBGPRINT_RAW(RT_DEBUG_TRACE, (" BBPCurrentBW = BW_40\n"));
						}
						else
						{
							RTUSBReadBBPRegister(pAd, BBP_R4, &BBPValue);
							BBPValue&= (~0x18);
							RTUSBWriteBBPRegister(pAd, BBP_R4, BBPValue);
							pAd->CommonCfg.BBPCurrentBW = BW_20;
							DBGPRINT_RAW(RT_DEBUG_TRACE, (" BBPCurrentBW = BW_20\n"));
						}
					}
					break;

				case RT_OID_VENDOR_READ_BBP:
					{
						UCHAR	Offset = *((PUCHAR)pData);
						PUCHAR	pValue = (PUCHAR)(pData + 1);

						DBGPRINT_RAW(RT_DEBUG_INFO, ("offset = 0x%02x\n", Offset));
						RTUSBReadBBPRegister(pAd, Offset, pValue);
						DBGPRINT_RAW(RT_DEBUG_INFO, ("value = 0x%02x\n", *pValue));
					}
					break;

				case RT_OID_VENDOR_WRITE_RF:
					{
						UINT32	Value = *((PUINT32)pData);

						DBGPRINT_RAW(RT_DEBUG_INFO, ("value = 0x%08x\n", Value));
						RTUSBWriteRFRegister(pAd, Value);
					}
					break;
				    
				case RT_OID_802_11_RESET_COUNTERS:
					{
//						UCHAR	Value[22];

						//RTUSBMultiRead(pAd, STA_CSR0, Value, 24);
					}
					break;

				case RT_OID_USB_VENDOR_RESET:
					RTUSB_VendorRequest(pAd,
											0,
											DEVICE_VENDOR_REQUEST_OUT,
											1,
											1,
											0,
											NULL,
											0);
					break;

				case RT_OID_USB_VENDOR_UNPLUG:
					RTUSB_VendorRequest(pAd,
										0,
										DEVICE_VENDOR_REQUEST_OUT,
										1,
										2,
										0,
										NULL,
										0);
					break;
      	                                                                                                                                                    
				case RT_OID_USB_VENDOR_SWITCH_FUNCTION:                                                                                                             
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("RT_OID_VENDOR_SWITCH_FUNCTION -- NOT SUPPORT !!\n"));
					break;                                                                                                                                              
			                                                                                                                                                
				case RT_OID_VENDOR_FLIP_IQ:
					{
						UINT32	Value1, Value2;
						//RTUSBReadMACRegister(pAd, PHY_CSR5, &Value1);
						//RTUSBReadMACRegister(pAd, PHY_CSR6, &Value2);
						if (*pData == 1)
						{
							DBGPRINT_RAW(RT_DEBUG_ERROR, ("I/Q Flip\n"));
							Value1 = Value1 | 0x0004;
							Value2 = Value2 | 0x0004;
						}
						else
						{
							DBGPRINT_RAW(RT_DEBUG_ERROR, ("I/Q Not Flip\n"));
							Value1 = Value1 & 0xFFFB;
							Value2 = Value2 & 0xFFFB;
						}
						///RTUSBWriteMACRegister(pAd, PHY_CSR5, Value1);
						//RTUSBWriteMACRegister(pAd, PHY_CSR6, Value2);
					}
					break;

				case RT_OID_UPDATE_TX_RATE:					
					NdisZeroMemory(pAd->CommonCfg.DesireRate, MAX_LEN_OF_SUPPORTED_RATES);
					NdisMoveMemory(pAd->CommonCfg.DesireRate, (PUCHAR)pData, sizeof(NDIS_802_11_RATES));
					MlmeUpdateTxRates(pAd, FALSE);
					MlmeUpdateHtTxRates(pAd, 0xff);
					if (ADHOC_ON(pAd))
						MakeIbssBeacon(pAd);
					break;

				case RT_OID_802_11_SET_PREAMBLE:
					{
						USHORT	Preamble = *((PUSHORT)(cmdqelmt->buffer));
						if (Preamble == Rt802_11PreambleShort)
						{
							pAd->CommonCfg.TxPreamble = Preamble;
							MlmeSetTxPreamble(pAd, Rt802_11PreambleShort);
						}
						else if ((Preamble == Rt802_11PreambleLong) || (Preamble == Rt802_11PreambleAuto))
						{
							// if user wants AUTO, initialize to INT32 here, then change according to AP's
							// capability upon association.
							pAd->CommonCfg.TxPreamble = Preamble;
							MlmeSetTxPreamble(pAd, Rt802_11PreambleLong);
						}
						else
							NdisStatus = NDIS_STATUS_INVALID_DATA;
						DBGPRINT(RT_DEBUG_ERROR, ("Set::RT_OID_802_11_SET_PREAMBLE (=%d)\n", Preamble));
					}
					break;

				case OID_802_11_NETWORK_TYPE_IN_USE:
					{
						NDIS_802_11_NETWORK_TYPE	NetType = *(PNDIS_802_11_NETWORK_TYPE)(cmdqelmt->buffer);
						if (NetType == Ndis802_11DS)
							RTMPSetPhyMode(pAd, PHY_11B);
						else if (NetType == Ndis802_11OFDM24)
							RTMPSetPhyMode(pAd, PHY_11BG_MIXED);
						else if (NetType == Ndis802_11OFDM5)
							RTMPSetPhyMode(pAd, PHY_11A);
						else
							NdisStatus = NDIS_STATUS_INVALID_DATA;
						DBGPRINT(RT_DEBUG_ERROR, ("CmdThread::OID_802_11_NETWORK_TYPE_IN_USE (=%d)\n",NetType));
					}
					break;

				case RT_OID_802_11_SET_PHY_MODE:
					{
						UINT32	phymode = *(UINT32 *)(cmdqelmt->buffer);
						pAd->CommonCfg.PhyMode = 0xff;
						RTMPSetPhyMode(pAd, phymode);
						DBGPRINT(RT_DEBUG_ERROR, ("CmdThread::RT_OID_802_11_SET_PHY_MODE (=%x)\n", phymode));
					}
					break;            	                                                                                                                                                    

				case OID_802_11_ADD_WEP:
					{
						UINT	i;
						UINT32	KeyIdx;
						PNDIS_802_11_WEP	pWepKey;

						DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::OID_802_11_ADD_WEP  \n"));
						
						pWepKey = (PNDIS_802_11_WEP)pData;
						KeyIdx = pWepKey->KeyIndex & 0x0fffffff;

						// it is a shared key
						if ((KeyIdx >= 4) || ((pWepKey->KeyLength != 5) && (pWepKey->KeyLength != 13)))
						{
							NdisStatus = NDIS_STATUS_INVALID_DATA;
							DBGPRINT(RT_DEBUG_ERROR, ("CmdThread::OID_802_11_ADD_WEP, INVALID_DATA!!\n"));
						}
						else 
						{
							UCHAR CipherAlg;
							pAd->SharedKey[BSS0][KeyIdx].KeyLen = (UCHAR) pWepKey->KeyLength;
							NdisMoveMemory(pAd->SharedKey[BSS0][KeyIdx].Key, &pWepKey->KeyMaterial, pWepKey->KeyLength);
							CipherAlg = (pAd->SharedKey[BSS0][KeyIdx].KeyLen == 5)? CIPHER_WEP64 : CIPHER_WEP128;

							//
							// Change the WEP cipher to CKIP cipher if CKIP KP on.
							// Funk UI or Meetinghouse UI will add ckip key from this path.
							//	

							if (CKIP_KP_ON(pAd))
							{
								if (CipherAlg == CIPHER_WEP64)  
									CipherAlg = CIPHER_CKIP64;
								else                            
								 	CipherAlg = CIPHER_CKIP128;
							}
							
							if (pAd->OpMode == OPMODE_STA)
						 	{
								pAd->MacTab.Content[BSSID_WCID].PairwiseKey.CipherAlg = pAd->SharedKey[BSS0][KeyIdx].CipherAlg;
								pAd->MacTab.Content[BSSID_WCID].PairwiseKey.KeyLen = pAd->SharedKey[BSS0][KeyIdx].KeyLen;
						 	}
							pAd->SharedKey[BSS0][KeyIdx].CipherAlg = CipherAlg;
							if (pWepKey->KeyIndex & 0x80000000)
							{
								// Default key for tx (shared key)
								UCHAR	IVEIV[8];
								UINT32	WCIDAttri, Value;
								USHORT	offset, offset2;
								NdisZeroMemory(IVEIV, 8);
								pAd->CommonCfg.DefaultKeyId = (UCHAR) KeyIdx;
								// Add BSSID to WCTable. because this is Tx wep key.
								// WCID Attribute UDF:3, BSSIdx:3, Alg:3, Keytable:1=PAIRWISE KEY, BSSIdx is 0
								WCIDAttri = (CipherAlg<<1)|SHAREDKEYTABLE;
								
								offset = MAC_WCID_ATTRIBUTE_BASE + (BSSID_WCID* HW_WCID_ATTRI_SIZE);
								RTUSBWriteMACRegister(pAd, offset, WCIDAttri);
								// 1. IV/EIV
								// Specify key index to find shared key.
								IVEIV[3] = (UCHAR)(KeyIdx<< 6);	//WEP Eiv bit off. groupkey index is not 0
								offset = PAIRWISE_IVEIV_TABLE_BASE + (BSS0Mcast_WCID * HW_IVEIV_ENTRY_SIZE);
								offset2 = PAIRWISE_IVEIV_TABLE_BASE + (BSSID_WCID* HW_IVEIV_ENTRY_SIZE);
								for (i=0; i<8;)
								{
									Value = IVEIV[i];
									Value += (IVEIV[i+1]<<8);
									Value += (IVEIV[i+2]<<16);
									Value += (IVEIV[i+3]<<24);
									RTUSBWriteMACRegister(pAd, offset+i, Value);
									RTUSBWriteMACRegister(pAd, offset2+i, Value);
									i+=4;
								}

								// 2. WCID Attribute UDF:3, BSSIdx:3, Alg:3, Keytable:use share key, BSSIdx is 0
								WCIDAttri = (pAd->SharedKey[BSS0][KeyIdx].CipherAlg<<1)|SHAREDKEYTABLE;
								offset = MAC_WCID_ATTRIBUTE_BASE + (BSS0Mcast_WCID* HW_WCID_ATTRI_SIZE);
							        DBGPRINT(RT_DEBUG_TRACE, ("BSS0Mcast_WCID : offset = %x, WCIDAttri = %x\n", offset, WCIDAttri));
								RTUSBWriteMACRegister(pAd, offset, WCIDAttri);
								
							}
							AsicAddSharedKeyEntry(pAd, BSS0, (UCHAR)KeyIdx, CipherAlg, pWepKey->KeyMaterial, NULL, NULL);
							DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::OID_802_11_ADD_WEP (KeyIdx=%d, Len=%d-byte)\n", KeyIdx, pWepKey->KeyLength));
						}
					}
					break;
				    
				case OID_802_11_REMOVE_WEP:
					{
						UINT32		KeyIdx;

						
						KeyIdx = *(NDIS_802_11_KEY_INDEX *) pData;
						if (KeyIdx & 0x80000000)
						{
							NdisStatus = NDIS_STATUS_INVALID_DATA;
							DBGPRINT(RT_DEBUG_ERROR, ("CmdThread::OID_802_11_REMOVE_WEP, INVALID_DATA!!\n"));
						}
						else
						{
							KeyIdx = KeyIdx & 0x0fffffff;
							if (KeyIdx >= 4)
							{
								NdisStatus = NDIS_STATUS_INVALID_DATA;
								DBGPRINT(RT_DEBUG_ERROR, ("CmdThread::OID_802_11_REMOVE_WEP, Invalid KeyIdx[=%d]!!\n", KeyIdx));
							}
							else
							{
								pAd->SharedKey[BSS0][KeyIdx].KeyLen = 0;
								pAd->SharedKey[BSS0][KeyIdx].CipherAlg = CIPHER_NONE;
								AsicRemoveSharedKeyEntry(pAd, BSS0, (UCHAR)KeyIdx);
								DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::OID_802_11_REMOVE_WEP (KeyIdx=%d)\n", KeyIdx));
							}
						}	
					}
					break;

				case OID_802_11_ADD_KEY_WEP:
					{
						PNDIS_802_11_KEY		pKey;
						UINT32					KeyIdx;						
						USHORT					i;						

						pKey = (PNDIS_802_11_KEY) pData;
						KeyIdx = pKey->KeyIndex & 0x0fffffff;
							 
						 // it is a shared key
						 if (KeyIdx >= 4)
						 {
							 NdisStatus = NDIS_STATUS_INVALID_DATA;
							 DBGPRINT(RT_DEBUG_ERROR, ("CmdThread::OID_802_11_ADD_KEY_WEP, Invalid KeyIdx[=%d]!!\n", KeyIdx));
						 }
						 else 
						 {
							 UCHAR CipherAlg;
							 
							 pAd->SharedKey[BSS0][KeyIdx].KeyLen = (UCHAR) pKey->KeyLength;
							 NdisMoveMemory(pAd->SharedKey[BSS0][KeyIdx].Key, &pKey->KeyMaterial, pKey->KeyLength);

							 if (pKey->KeyLength == 5)
								 CipherAlg = CIPHER_WEP64;
							 else
								 CipherAlg = CIPHER_WEP128;
#ifdef CCX_SUPPORT
							 if (CKIP_KP_ON(pAd))
							 {
								 if (CipherAlg == CIPHER_WEP64)  
								 	CipherAlg = CIPHER_CKIP64;
								 else                            
								 	CipherAlg = CIPHER_CKIP128;
							 }

							 // always expand the KEY to 16-byte here for efficiency sake. so that in case CKIP is used
							 // sometime later we don't have to do key expansion for each TX in RTUSBHardTransmit().
							 // However, we shouldn't change pAd->SharedKey[BSS0][KeyIdx].KeyLen
							 if (pKey->KeyLength < 16)
							 {
								 for(i = 1; i < (16 / pKey->KeyLength); i++)
								 {
									 NdisMoveMemory(&pAd->SharedKey[BSS0][KeyIdx].Key[i * pKey->KeyLength], 
													&pKey->KeyMaterial[0], 
													pKey->KeyLength);
								 }
								 NdisMoveMemory(&pAd->SharedKey[BSS0][KeyIdx].Key[i * pKey->KeyLength], 
												&pKey->KeyMaterial[0], 
												16 - (i * pKey->KeyLength));
							 }
#endif
							pAd->SharedKey[BSS0][KeyIdx].CipherAlg = CipherAlg;
							if (pAd->OpMode == OPMODE_STA)
							{
								pAd->MacTab.Content[BSSID_WCID].PairwiseKey.CipherAlg = pAd->SharedKey[BSS0][KeyIdx].CipherAlg;
								pAd->MacTab.Content[BSSID_WCID].PairwiseKey.KeyLen = pAd->SharedKey[BSS0][KeyIdx].KeyLen;
						 	}
							if (pKey->KeyIndex & 0x80000000)
							{
								// Default key for tx (shared key)
								pAd->CommonCfg.DefaultKeyId = (UCHAR) KeyIdx;
								// Add BSSID to WCTable. because this is Tx wep key.
								RTMPAddBSSIDCipher(pAd, BSSID_WCID, pKey);
							}

							AsicAddSharedKeyEntry(pAd, BSS0, (UCHAR)KeyIdx, CipherAlg, pAd->SharedKey[BSS0][KeyIdx].Key, NULL, NULL);
							DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::OID_802_11_ADD_KEY_WEP (KeyIdx=%d, KeyLen=%d, CipherAlg=%d)\n", pAd->CommonCfg.DefaultKeyId, pAd->SharedKey[BSS0][KeyIdx].KeyLen, pAd->SharedKey[BSS0][KeyIdx].CipherAlg));
							// ad-hoc mode need to specify group key with WCID index=BSS0Mcast_WCID. Let's always set this key here.
							if (pKey->KeyIndex & 0x80000000)
							{
								UCHAR	IVEIV[8];
								UINT32	WCIDAttri, Value;
								USHORT	offset;
								NdisZeroMemory(IVEIV, 8);
								// 1. IV/EIV
								// Specify key index to find shared key.
								IVEIV[3] = (UCHAR)(KeyIdx<< 6);	//WEP Eiv bit off. groupkey index is not 0
								offset = PAIRWISE_IVEIV_TABLE_BASE + (BSS0Mcast_WCID * HW_IVEIV_ENTRY_SIZE);
								for (i=0; i<8;)
								{
									Value = IVEIV[i];
									Value += (IVEIV[i+1]<<8);
									Value += (IVEIV[i+2]<<16);
									Value += (IVEIV[i+3]<<24);
									RTUSBWriteMACRegister(pAd, offset+i, Value);
									i+=4;
								}

								// 2. WCID Attribute UDF:3, BSSIdx:3, Alg:3, Keytable:use share key, BSSIdx is 0
								WCIDAttri = (pAd->SharedKey[BSS0][KeyIdx].CipherAlg<<1)|SHAREDKEYTABLE;
								offset = MAC_WCID_ATTRIBUTE_BASE + (BSS0Mcast_WCID* HW_WCID_ATTRI_SIZE);
							 	DBGPRINT(RT_DEBUG_TRACE, ("BSS0Mcast_WCID : offset = %x, WCIDAttri = %x\n", offset, WCIDAttri));
								RTUSBWriteMACRegister(pAd, offset, WCIDAttri);
								
							}

						}
					}
					break;

				case OID_802_11_ADD_KEY:
					{  
						NdisStatus = RTMPWPAAddKeyProc(pAd, pData);
						DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::OID_802_11_ADD_KEY\n"));
					}
					break;
                                                                                                                                                                                                                                                                              
				case OID_802_11_REMOVE_KEY:
					{
						PNDIS_802_11_REMOVE_KEY  pRemoveKey;
						UINT32  KeyIdx;
						
						pRemoveKey = (PNDIS_802_11_REMOVE_KEY) pData;
						if (pAd->CommonCfg.AuthMode >= Ndis802_11AuthModeWPA)
						{
							NdisStatus = RTMPWPARemoveKeyProc(pAd, pData);
							DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::RTMPWPARemoveKeyProc\n"));
						}
						else 
						{
							KeyIdx = pRemoveKey->KeyIndex;
								
							if (KeyIdx & 0x80000000)
							{
								// Should never set default bit when remove key
								NdisStatus = NDIS_STATUS_INVALID_DATA;
								DBGPRINT(RT_DEBUG_ERROR, ("CmdThread::OID_802_11_REMOVE_KEY, Invalid KeyIdx[=%d]!!\n", KeyIdx));
							}
							else
							{
								KeyIdx = KeyIdx & 0x0fffffff;
								if (KeyIdx >= 4)
								{
									NdisStatus = NDIS_STATUS_INVALID_DATA;
									DBGPRINT(RT_DEBUG_ERROR, ("CmdThread::OID_802_11_REMOVE_KEY, Invalid KeyIdx[=%d]!!\n", KeyIdx));										
								}
								else
								{
									pAd->SharedKey[BSS0][KeyIdx].KeyLen = 0;
									pAd->SharedKey[BSS0][KeyIdx].CipherAlg = CIPHER_NONE;
									AsicRemoveSharedKeyEntry(pAd, BSS0, (UCHAR)KeyIdx);
									DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::AsicRemoveSharedKeyEntry(KeyIdx=%d)\n", KeyIdx));
								}
							}
						}

					}
					break;
				case RT_OID_802_11_RSSI_1:
					{
						UCHAR	BBPValue;
						UINT32	ulInfo;

						RTUSBReadBBPRegister(pAd, BBP_R3, &BBPValue);
						if (((BBPValue & 0x18) >> 3) >= 1)
						{
							ulInfo = pAd->StaCfg.LastRssi2;
						}
						else
						{
							ulInfo = -200;
						}
						NdisMoveMemory(pData, &ulInfo, sizeof(UINT32));
						DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::RT_OID_802_11_RSSI_1(=%d)\n", ulInfo));
					}
						
					break;
				case RT_OID_802_11_RSSI_2:
					{
						UCHAR	BBPValue;
						UINT32	ulInfo;
						RTUSBReadBBPRegister(pAd, BBP_R3, &BBPValue);
						if (((BBPValue & 0x18) >> 3) == 2)
						{
							ulInfo = pAd->StaCfg.LastRssi3;
						}
						else
						{
							ulInfo = -200;
						}
						NdisMoveMemory(pData, &ulInfo, sizeof(UINT32));
						DBGPRINT(RT_DEBUG_INFO, ("CmdThread::RT_OID_802_11_RSSI_2(=%d)\n", ulInfo));

					}
					break;
				case OID_802_11_STATISTICS:
					{
						NDIS_802_11_STATISTICS      Statistics;

						DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::OID_802_11_STATISTICS pData= %lx  \n", (ULONG)pData));

						NICUpdateRawCounters(pAd);
						// Sanity check for calculation of sucessful count
						if (pAd->WlanCounters.TransmittedFragmentCount.QuadPart < pAd->WlanCounters.RetryCount.QuadPart)
							pAd->WlanCounters.TransmittedFragmentCount = pAd->WlanCounters.RetryCount;
						Statistics.TransmittedFragmentCount = pAd->WlanCounters.TransmittedFragmentCount;
						Statistics.MulticastTransmittedFrameCount = pAd->WlanCounters.MulticastTransmittedFrameCount;
						Statistics.FailedCount = pAd->WlanCounters.FailedCount;
						Statistics.RetryCount = pAd->WlanCounters.RetryCount;
						Statistics.MultipleRetryCount = pAd->WlanCounters.MultipleRetryCount;
						Statistics.RTSSuccessCount = pAd->WlanCounters.RTSSuccessCount;
						Statistics.RTSFailureCount = pAd->WlanCounters.RTSFailureCount;
						Statistics.ACKFailureCount = pAd->WlanCounters.ACKFailureCount;
						Statistics.FrameDuplicateCount = pAd->WlanCounters.FrameDuplicateCount;
						Statistics.ReceivedFragmentCount = pAd->WlanCounters.ReceivedFragmentCount;
						Statistics.MulticastReceivedFrameCount = pAd->WlanCounters.MulticastReceivedFrameCount;

#ifdef DBG
						Statistics.FCSErrorCount = pAd->RalinkCounters.RealFcsErrCount;
#else
						Statistics.FCSErrorCount = pAd->WlanCounters.FCSErrorCount;
						Statistics.FrameDuplicateCount.LowPart = pAd->WlanCounters.FrameDuplicateCount.LowPart / 100;
#endif							
						NdisMoveMemory(pData, &Statistics, sizeof(NDIS_802_11_STATISTICS));
		
		
					}
					break;
					
				case OID_802_11_POWER_MODE:
					{
						NDIS_802_11_POWER_MODE PowerMode = *(PNDIS_802_11_POWER_MODE) pData;
						DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::OID_802_11_POWER_MODE (=%d)\n",PowerMode));
						
						// save user's policy here, but not change StaCfg.Psm immediately
						if (PowerMode == Ndis802_11PowerModeCAM) 
						{
							// clear PSM bit immediately
							MlmeSetPsm(pAd, PWR_ACTIVE);
			            
							OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM); 
							if (pAd->StaCfg.bWindowsACCAMEnable == FALSE)
								pAd->StaCfg.WindowsPowerMode = PowerMode;
							pAd->StaCfg.WindowsBatteryPowerMode = PowerMode;
						} 
						else if (PowerMode == Ndis802_11PowerModeMAX_PSP) 
						{
							// do NOT turn on PSM bit here, wait until MlmeCheckPsmChange()
							// to exclude certain situations.
							//     MlmeSetPsm(pAd, PWR_SAVE);
							if (pAd->StaCfg.bWindowsACCAMEnable == FALSE)
								pAd->StaCfg.WindowsPowerMode = PowerMode;
							pAd->StaCfg.WindowsBatteryPowerMode = PowerMode;
							OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM); 
							pAd->StaCfg.DefaultListenCount = 5;
						} 
						else if (PowerMode == Ndis802_11PowerModeFast_PSP) 
						{
							// do NOT turn on PSM bit here, wait until MlmeCheckPsmChange()
							// to exclude certain situations.
							//     MlmeSetPsm(pAd, PWR_SAVE);
							OPSTATUS_SET_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM);
							if (pAd->StaCfg.bWindowsACCAMEnable == FALSE)
								pAd->StaCfg.WindowsPowerMode = PowerMode;
							pAd->StaCfg.WindowsBatteryPowerMode = PowerMode;
							pAd->StaCfg.DefaultListenCount = 3;
						} 
					}					
					break;

				case RT_PERFORM_SOFT_DIVERSITY:
						AsicRxAntEvalAction(pAd);
					break;
					
				case RT_OID_FORCE_WAKE_UP:
					AsicForceWakeup(pAd);
					break;

				case RT_OID_SET_PSM_BIT_ACTIVE:
					MlmeSetPsm(pAd, PWR_ACTIVE);
					break;
#if 0	// not ready !					
				case RT_OID_802_11_ADD_IMME_BA:
					{
						MAC_TABLE_ENTRY *pEntry;
						OID_ADD_BA_ENTRY BA = *(OID_ADD_BA_ENTRY *)pData;
						pEntry = MacTableLookup(pAd, BA.MACAddr);
						if (pEntry)
						{
							BATableInsertEntry(pAd, pEntry->Aid, 0, 0xffff, BA.TID, BA.nMSDU, Originator_WaitRes, BA.IsRecipient);
							DBGPRINT(RT_DEBUG_ERROR, ("--> RT_OID_802_11_ADD_IMME_BA \n"));    
						}
					}
					break;
				case RT_OID_802_11_TEAR_IMME_BA:
					{
						MAC_TABLE_ENTRY *pEntry;
						OID_ADD_BA_ENTRY BA = *(OID_ADD_BA_ENTRY *)pData;
						pEntry = MacTableLookup(pAd, BA.MACAddr);
						if (pEntry)
						{
#if 0						
							BATableTearRECEntry( pAd, BA.TID, (UCHAR)pEntry->Aid, FALSE);
#else
							BARecSessionTearDown(pAd, (UCHAR)pEntry->Aid, BA.TID, FALSE);
#endif
							DBGPRINT(RT_DEBUG_ERROR, ("--> RT_OID_802_11_TEAR_IMME_BA \n"));    
						}
					}
					break;
#endif					
				default:
					DBGPRINT(RT_DEBUG_ERROR, ("--> Control Thread !! ERROR !! Unknown !! \n"));
					break;                                                                                                                                              
			}                                                                                                                                                       
		}                                                                                                                                                          
                                                                                                                                                                
		if (cmdqelmt->CmdFromNdis == TRUE)                                                                                                                      
		{                                                                                                                                                       
			if ((cmdqelmt->command != OID_802_11_BSSID_LIST_SCAN) &&                                                                                            
				(cmdqelmt->command != RT_OID_802_11_BSSID) &&                                                                                                   
				(cmdqelmt->command != OID_802_11_SSID) &&                                                                                                       
				(cmdqelmt->command != OID_802_11_DISASSOCIATE))                                                                                                 
			{ // do nothing!                                                                                                                                               
			}                                                                                                                                                   
                                                                                                                                                                
			if ((cmdqelmt->command != OID_802_11_STATISTICS) &&
				(cmdqelmt->command != RT_OID_802_11_RSSI_1) &&
				(cmdqelmt->command != RT_OID_802_11_RSSI_2) &&
				(cmdqelmt->command != RT_OID_MULTI_READ_MAC) &&
				(cmdqelmt->command != RT_OID_VENDOR_READ_BBP) &&                                                                                                
#ifdef DBG					                                                                                                                                    
				(cmdqelmt->command != RT_OID_802_11_QUERY_HARDWARE_REGISTER) &&                                                                                 
#endif					                                                                                                                                        
				(cmdqelmt->command != RT_OID_USB_VENDOR_EEPROM_READ))                                                                                           
			{                                                                                                                                                   
				if (cmdqelmt->buffer != NULL)                                                                                                                   
					NdisFreeMemory(cmdqelmt->buffer, cmdqelmt->bufferlength, 0);
			}                                                                                                                                                   
			                                                                                                                                                    
			NdisFreeMemory(cmdqelmt, sizeof(CmdQElmt), 0);
		}                                                                                                                                                       
		else                                                                                                                                                    
		{
			if ((cmdqelmt->buffer != NULL) && (cmdqelmt->bufferlength != 0))
				NdisFreeMemory(cmdqelmt->buffer, cmdqelmt->bufferlength, 0);
            {                                                                                                                                                    
				NdisFreeMemory(cmdqelmt, sizeof(CmdQElmt), 0);
			}
		}
	}	/* end of while */	                                                                                                                                                           
}
#if 0
#ifdef DBG
#define HARDWARE_MAC	0
#define HARDWARE_BBP	1
#define HARDWARE_RF		2
#define HARDWARE_EEP	3
NDIS_STATUS RTUSBQueryHardWareRegister(
	IN	PRTMP_ADAPTER	pAdapter,
	IN	PVOID			pBuf)
{
	PRT_802_11_HARDWARE_REGISTER	pHardwareRegister;
	UINT32							Value;
	USHORT							Offset;
	UCHAR							bbpValue;
	UCHAR							bbpID;
	USHORT							eepID;
	USHORT							eepValue;
	NDIS_STATUS						Status = NDIS_STATUS_SUCCESS;	

	pHardwareRegister = (PRT_802_11_HARDWARE_REGISTER) pBuf;

	if (pHardwareRegister->HardwareType == HARDWARE_MAC)
	{
		//Check Offset is valid?
		if (pHardwareRegister->Offset > 0xF4)
			Status = NDIS_STATUS_INVALID_DATA;
		
		Offset = (USHORT) pHardwareRegister->Offset;
		RTUSBReadMACRegister(pAdapter, Offset, &Value);
		pHardwareRegister->Data = Value;
		DBGPRINT(RT_DEBUG_TRACE, ("MAC:Offset[0x%04x]=[0x%04x]\n", Offset, Value));
	}
	else if (pHardwareRegister->HardwareType == HARDWARE_BBP)
	{
		bbpID = (UCHAR) pHardwareRegister->Offset;
		
		RTUSBReadBBPRegister(pAdapter, bbpID, &bbpValue);		
		pHardwareRegister->Data = bbpValue;
		DBGPRINT(RT_DEBUG_TRACE, ("BBP:ID[0x%02x]=[0x%02x]\n", bbpID, bbpValue));		
	}
	else if (pHardwareRegister->HardwareType == HARDWARE_EEP)
	{
		eepID = (USHORT) pHardwareRegister->Offset;
		eepValue = (USHORT) pHardwareRegister->Data;
		RTUSBReadEEPROM(pAdapter, eepID, (PUCHAR)&eepValue,2);
		pHardwareRegister->Data = eepValue;
		DBGPRINT(RT_DEBUG_TRACE, ("EEP:ID[0x%04x]=[0x%04x]\n", eepID, eepValue));		
	}
	else
		Status = NDIS_STATUS_INVALID_DATA;
	
	return Status;
}

NDIS_STATUS RTUSBSetHardWareRegister(
	IN	PRTMP_ADAPTER	pAd,
	IN	PVOID			pBuf)
{
	PRT_802_11_HARDWARE_REGISTER	pHardwareRegister;
	UINT32							Value;
	USHORT							Offset;
	UCHAR							bbpValue;
	UCHAR							bbpID;
	USHORT							eepID;
	USHORT							eepValue;
	NDIS_STATUS						Status = NDIS_STATUS_SUCCESS;	
	NTSTATUS		ntStatus;

	pHardwareRegister = (PRT_802_11_HARDWARE_REGISTER) pBuf;

	if (((RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)) ||
		(RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS)) ))
			 return NDIS_STATUS_INVALID_DATA;

	
	if (pHardwareRegister->HardwareType == HARDWARE_MAC)
	{
		//Check Offset is valid?
		if (pHardwareRegister->Offset > 0xF4)
			Status = NDIS_STATUS_INVALID_DATA;

		Offset = (USHORT) pHardwareRegister->Offset;
		Value = (UINT32) pHardwareRegister->Data;
		if ((!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)) && 
			(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS)))
				RTUSBWriteMACRegister(pAd, Offset, Value);		
		DBGPRINT(RT_DEBUG_TRACE, ("CmdThread::RT_OID_802_11_SET_HARDWARE_REGISTER (MAC offset=0x%08x, data=0x%08x)\n", pHardwareRegister->Offset, pHardwareRegister->Data));
		// 2004-11-08 a special 16-byte on-chip memory is used for RaConfig to pass debugging parameters to driver
		// for debug-tuning only
		if ((pHardwareRegister->Offset >= HW_DEBUG_SETTING_BASE))
		{
			// 0x2bf0: test power-saving feature
			if (pHardwareRegister->Offset == HW_DEBUG_SETTING_BASE)
			{
#if 0			
				UINT32 isr, imr, gimr;
				USHORT tbtt = 3;

				RTMP_IO_READ32(pAd, MCU_INT_SOURCE_CSR, &isr);
				RTMP_IO_READ32(pAd, MCU_INT_MASK_CSR, &imr);
				RTMP_IO_READ32(pAd, INT_MASK_CSR, &gimr);
				DBGPRINT(RT_DEBUG_TRACE, ("Sleep %d TBTT, 8051 IMR=%08x, ISR=%08x, MAC IMR=%08x\n", tbtt, imr, isr, gimr));
				AsicSleepThenAutoWakeup(pAd, tbtt);	
#endif				
				if (pHardwareRegister->Data == 0x8)
				{
					{
						pAd->bForcePrintTX = TRUE;
						DBGPRINT(RT_DEBUG_ERROR, ("bForcePrintTX = %d \n", pAd->bForcePrintTX));
					}
				}
				else if (pHardwareRegister->Data == 0x9)
				{
					{
						pAd->bForcePrintTX = FALSE;
						DBGPRINT(RT_DEBUG_ERROR, ("bForcePrintTX = %d \n", pAd->bForcePrintTX));
					}
				}
				else if (pHardwareRegister->Data == 0xe)
				{
					{
						pAd->bForcePrintRX = TRUE;
						DBGPRINT(RT_DEBUG_ERROR, ("bForcePrintRX = %d \n", pAd->bForcePrintRX));
					}
				}
				else if (pHardwareRegister->Data == 0xf)
				{
					{
						pAd->bForcePrintRX = FALSE;
						DBGPRINT(RT_DEBUG_ERROR, ("bForcePrintRX = %d \n", pAd->bForcePrintRX));
					}
				}
				else if (pHardwareRegister->Data == 0x31)
				{
					UCHAR	i;
					PTX_CONTEXT		pTxContext;
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("==========================================\n"));
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("pAd->Flags=0x%x, pAd->BulkFlags=0x%x\n", pAd->Flags, pAd->BulkFlags));
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("pAd->PrioRingFirstIndex = %d, PopMgmtIndex = %d, PushMgmtIndex = %d\n", pAd->PrioRingFirstIndex,  pAd->PopMgmtIndex, pAd->PushMgmtIndex));
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("=============MLME==================\n"));
					for (i=0;i<TX_RING_SIZE;i++)
					{
						pTxContext= (PTX_CONTEXT)&pAd->MLMEContext[i];
						DBGPRINT_RAW(RT_DEBUG_ERROR, ("[%d] IRPPending=%d:  InUse=%d: bWaitingBulkOut=%d: BulkOutSize=%05d: bFullForBulkOut=%d \n " , i, pTxContext->IRPPending, pTxContext->InUse, pTxContext->bWaitingBulkOut, pTxContext->BulkOutSize, pTxContext->bFullForBulkOut));
					}
				}
				else if (pHardwareRegister->Data == 0x32)
				{
					UCHAR	i;
					PTX_CONTEXT		pTxContext;
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("==========================================\n"));
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("pAd->Flags=0x%x, pAd->BulkFlags=0x%x\n", pAd->Flags, pAd->BulkFlags));
					DBGPRINT_RAW(RT_DEBUG_ERROR, (">>BulkOutReq=0x%x, BulkOutComplete=0x%x\n", pAd->BulkOutReq, pAd->BulkOutComplete));
					DBGPRINT_RAW(RT_DEBUG_ERROR, (">>BulkOutCompleteCancel=0x%x, BulkOutCompleteOther=0x%x\n", pAd->BulkOutCompleteCancel, pAd->BulkOutCompleteOther));
					DBGPRINT(RT_DEBUG_TRACE,(" --> CurWritePosition = %d.  NextBulkOutPosition = %d. \n", pAd->TxContext[0].CurWritePosition, pAd->TxContext[0].NextBulkOutPosition));
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x33)
				{
					UCHAR	i;
					PRX_CONTEXT		pRxContext;
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("==========================================\n"));
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("NextRxBulkInIndex = 0x%x, NextRxBulkInReadIndex =0x%x, pAd->NextRxBulkInPosition = 0x%x. \n", pAd->NextRxBulkInIndex,  pAd->NextRxBulkInReadIndex, pAd->NextRxBulkInPosition));
					DBGPRINT_RAW(RT_DEBUG_ERROR, (">>BulkInReq=0x%x, BulkInComplete=0x%x, BulkInCompleteFail=%x\n", pAd->BulkInReq, pAd->BulkInComplete, pAd->BulkInCompleteFail));
					for (i=0;i<RX_RING_SIZE;i++)
					{
						pRxContext= (PRX_CONTEXT)&pAd->RxContext[i];
						DBGPRINT_RAW(RT_DEBUG_ERROR, ("[%d] IRPPending=%d:  InUse=%d: Readable=%d: \n "
							, i, pRxContext->IRPPending, pRxContext->InUse, pRxContext->Readable));
					}
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x34)
				{
					UINT32		value;
					value = (SLEEPCID<<16)+(OWNERMCU<<24)+ (0x40<<8)+1;
					RTUSBWriteMACRegister(pAd, 0x7010, value);
					RTUSBWriteMACRegister(pAd, 0x404, 0x30);
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("Sleep Mailbox testvalue %x\n", value));
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x35)
				{
					BOOLEAN	 	Cancelled;
					UINT32		value;
					value = (WAKECID<<16)+(OWNERMCU<<24);
					RTUSBWriteMACRegister(pAd, 0x7010, value);
					RTUSBWriteMACRegister(pAd, 0x404, 0x31);
			
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("Wakeup Mailbox testvalue%x\n", value));
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x36)
				{
					BOOLEAN	 	Cancelled;
					UINT32		value, ledcs;
					ledcs = (pHardwareRegister->Data&0xffff00)>>8;
					value = (OWNERMCU<<24)+(LEDMODECID<<16)+ledcs;
					RTUSBWriteMACRegister(pAd, 0x7010, value);
					RTUSBWriteMACRegister(pAd, 0x404, 0x50);
			
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("LED Mailbox test value %x\n", value));
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x37)
				{
					BOOLEAN	 	Cancelled;
					UINT32		value, stren;
					stren = (pHardwareRegister->Data&0xffff00)>>8;
					value = (OWNERMCU<<24)+(SIGSTRENCID<<16)+stren;
					RTUSBWriteMACRegister(pAd, 0x7010, value);
					RTUSBWriteMACRegister(pAd, 0x404, 0x51);
					//RTMPCancelTimer(&pAd->MailboxTimer,		&Cancelled);
					//RTMPSetTimer(pAd, &pAd->MailboxTimer, 100);
			
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("SignalStr Mailbox test value %x\n", value));
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x38)
				{
					UINT32		commandid, returncode;
					RTUSBReadMACRegister( pAd, 0x7014, &commandid);
					RTUSBReadMACRegister( pAd, 0x701c, &returncode);
					DBGPRINT(RT_DEBUG_ERROR, ("MailBoxTOExec read 0x7014 command id : 0x%x\n", commandid));
					DBGPRINT(RT_DEBUG_ERROR, ("MailBoxTOExec read 0x701c code : 0x%x\n",returncode));
					if (commandid != 0xffffffff)
						RTUSBWriteMACRegister( pAd, 0x7014, 0xffffffff);
			
					DBGPRINT_RAW(RT_DEBUG_ERROR, (" Clear Command\n"));
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x39)
				{
					UINT32		value;
					UINT32		commandid, returncode;
					value = (QUERYPOWERCID<<16)+(OWNERMCU<<24)+ (0x40<<8)+1;
					{

						RTUSBReadMACRegister( pAd, 0x7014, &commandid);
						RTUSBReadMACRegister( pAd, 0x701c, &returncode);
						DBGPRINT(RT_DEBUG_ERROR, ("MailBoxTOExec read 0x7014 command id : 0x%x\n", commandid));
						DBGPRINT(RT_DEBUG_ERROR, ("MailBoxTOExec read 0x701c code : 0x%x\n",returncode));
					}
					RTUSBWriteMACRegister( pAd, 0x7014, 0xffffffff);
					RTUSBWriteMACRegister(pAd, 0x7010, value);
					RTUSBWriteMACRegister(pAd, 0x404, 0x32);
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("QUERY POWER Mailbox testvalue %x\n", value));
					/*RTUSBReadMACRegister( pAd, 0x7014, &commandid);
					RTUSBReadMACRegister( pAd, 0x701c, &returncode);
					DBGPRINT(RT_DEBUG_ERROR, ("MailBoxTOExec read 0x7014 command id : 0x%x\n", commandid));
					DBGPRINT(RT_DEBUG_ERROR, ("MailBoxTOExec read 0x701c code : 0x%x\n",returncode));
					RTUSBWriteMACRegister( pAd, 0x7014, 0xffffffff);
					Status = RTUSB_VendorRequest(
						pAd,
						(USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK),
						DEVICE_VENDOR_REQUEST_IN,
						0x1,
						0x12,
						0,
						&returncode,
						4);*/
			
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x40)
				{
					UINT32		TbttNumToNextWakeUp;
					AUTO_WAKEUP_STRUC	WakeupCfg;
					UINT32		value;
					TbttNumToNextWakeUp = (pHardwareRegister->Data&0xff00) >> 8;

					value = (SLEEPCID<<16)+(OWNERMCU<<24)+ (0x40<<8)+1;
					RTUSBWriteMACRegister(pAd, 0x7010, value);
					RTUSBWriteMACRegister(pAd, 0x404, 0x30);
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("Sleep Mailbox testvalue %x\n", value));
					if (TbttNumToNextWakeUp > 0)
					{
						WakeupCfg.field.EnableAutoWakeup = 1;
						WakeupCfg.field.NumofSleepingTbtt = TbttNumToNextWakeUp-1;
						// lead time should set to minimum HW wakeuptime. use 7 first. The correct time needs measurement.
						WakeupCfg.field.AutoLeadTime = 0xa;
						RTUSBWriteMACRegister(pAd, AUTO_WAKEUP_CFG, WakeupCfg.word);
					}
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("MlmeSet TbttNumToNextWakeUp=%d. Data = %x\n", TbttNumToNextWakeUp, pHardwareRegister->Data));
			
					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x41)
				{
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("pAd->MacTab.Content[1].HTPhyMode = %x. \n", pAd->MacTab.Content[1].HTPhyMode.word));					
				}
				else if ((pHardwareRegister->Data&0xff) == 0x42)
				{
					pAd->bForceDisableautowake = TRUE;
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("Force Disable autowakeup =%d. \n", pAd->bForceDisableautowake));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x43)
				{
					pAd->bForceDisableautowake = FALSE;
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("Force Disable autowakeup =%d. \n", pAd->bForceDisableautowake));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x44)
				{
					RTUSB_SET_BULK_FLAG(pAd, fRTUSB_BULK_OUT_PSPOLL);
					RTUSBKickBulkOut(pAd);
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("Send a Ps-Poll frame. \n"));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x45)
				{
					RTUSB_SET_BULK_FLAG(pAd, fRTUSB_BULK_OUT_DATA_NULL);
					RTUSBKickBulkOut(pAd);
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("Send a NULL frame. \n"));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x46)
				{
				}
				else if ((pHardwareRegister->Data&0xff) == 0x47)
				{
					pAd->bDisablescanning= FALSE;
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("Normal SCANNING. \n"));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x48)
				{
					RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST);
					RTUSBBulkReceive(pAd);
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("RTUSBBulkReceive \n"));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x4a)
				{
					pAd->QuickTimerP= 500;
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("pAd->QuickTimerP = %d\n",pAd->QuickTimerP));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x49)
				{
					pAd->QuickTimerP= 100;
					DBGPRINT_RAW(RT_DEBUG_ERROR, ("pAd->QuickTimerP = %d\n",pAd->QuickTimerP));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x4c)
				{
					DBGPRINT(RT_DEBUG_ERROR, (" MacTab.Content[1].word = 0x%x \n",pAd->MacTab.Content[1].HTPhyMode.word));
					DBGPRINT(RT_DEBUG_ERROR, (" MacTab.Content[0].word = 0x%x \n",pAd->MacTab.Content[0].HTPhyMode.word));
					DBGPRINT(RT_DEBUG_ERROR, (" mlme.word = 0x%x \n",pAd->CommonCfg.MlmeTransmit.word));
				}
				else if ((pHardwareRegister->Data&0xf000) != 0)
				{
								{
								UCHAR	MCS;
								
								MCS = (UCHAR)(pHardwareRegister->Data)&0xf;
								pAd->CommonCfg.HTPhyMode.field.MCS = MCS;
								pAd->MacTab.Content[1].HTPhyMode.field.MCS = MCS;

								if (((UCHAR)(pHardwareRegister->Data) & 0xff) == MCS_AUTO)
								{
									PRTMP_TX_RATE_SWITCH	pCurrTxRate;

									pAd->CommonCfg.TxRateIndex = RateSwitchTable[1];
									pCurrTxRate = (PRTMP_TX_RATE_SWITCH) &RateSwitchTable[(pAd->CommonCfg.TxRateIndex+1)*5];
									pAd->CommonCfg.HTPhyMode.field.MCS = pCurrTxRate->CurrMCS;
									pAd->MacTab.Content[1].HTPhyMode.field.MCS = pCurrTxRate->CurrMCS;
										
									pAd->CommonCfg.bAutoTxRateSwitch = TRUE;
								}
								else
								{
									// bit10 for BW control
									if ((pHardwareRegister->Data & 0x0400) == 0x400)
									{
										UCHAR	BBPValue;
										UINT32	MACValue;
										pAd->CommonCfg.HTPhyMode.field.BW = BW_40;
										if (ADHOC_ON(pAd) && (pAd->CommonCfg.Channel <3))
										{
			RTUSBReadMACRegister(pAd, TX_BAND_CFG, &MACValue);
			MACValue &= 0xfe;
			MACValue |= 0x1;
			RTUSBWriteMACRegister(pAd, TX_BAND_CFG, MACValue);
			RTUSBReadBBPRegister(pAd, BBP_R4, &BBPValue);
			BBPValue&= (~0x18);
			BBPValue|= (0x10);
			RTUSBWriteBBPRegister(pAd, BBP_R4, BBPValue);
			RTUSBReadBBPRegister(pAd, BBP_R3, &BBPValue);
			BBPValue&= (~0x20);
			RTUSBWriteBBPRegister(pAd, BBP_R3, BBPValue);
			pAd->CommonCfg.BBPCurrentBW = BW_40;
			pAd->CommonCfg.CentralChannel = pAd->CommonCfg.Channel +2;
			AsicSwitchChannel(pAd, pAd->CommonCfg.CentralChannel,FALSE);
			AsicLockChannel(pAd, pAd->CommonCfg.CentralChannel);
										}
										if (ADHOC_ON(pAd) && (pAd->CommonCfg.Channel >=3))
										{
			RTUSBReadMACRegister(pAd, TX_BAND_CFG, &MACValue);
			MACValue &= 0xfe;
			RTUSBWriteMACRegister(pAd, TX_BAND_CFG, MACValue);
			RTUSBReadBBPRegister(pAd, BBP_R4, &BBPValue);
			BBPValue&= (~0x18);
			BBPValue|= (0x10);
			RTUSBWriteBBPRegister(pAd, BBP_R4, BBPValue);
			RTUSBReadBBPRegister(pAd, BBP_R3, &BBPValue);
			BBPValue&= (~0x20);
			RTUSBWriteBBPRegister(pAd, BBP_R3, BBPValue);
			pAd->CommonCfg.BBPCurrentBW = BW_40;
			pAd->CommonCfg.CentralChannel = pAd->CommonCfg.Channel -2;
			AsicSwitchChannel(pAd, pAd->CommonCfg.CentralChannel,FALSE);
			AsicLockChannel(pAd, pAd->CommonCfg.CentralChannel);
										}
									}
									else
									{
										UCHAR	BBPValue;
										if (ADHOC_ON(pAd))
										{
											RTUSBReadBBPRegister(pAd, BBP_R4, &BBPValue);
											BBPValue &= (~0x08);
											RTUSBWriteBBPRegister(pAd, BBP_R4, BBPValue);
										}
										pAd->CommonCfg.BBPCurrentBW = BW_20;
										pAd->CommonCfg.HTPhyMode.field.BW = BW_20;
									}
									
									// bit8, 9 for MODE control
									pAd->CommonCfg.HTPhyMode.field.MODE = (UCHAR)((pHardwareRegister->Data & 0x0300) >> 8);
									
									// bit7 for STBC control
									if ((pHardwareRegister->Data & 0x0080) == 0x0080)
									{
										pAd->CommonCfg.HTPhyMode.field.STBC = STBC_USE;
									}
									else
									{
										pAd->CommonCfg.HTPhyMode.field.STBC = STBC_NONE;
									}

									// bit6 for ShortGI control
									if ((pHardwareRegister->Data & 0x0040) == 0x0040)
									{
										pAd->CommonCfg.HTPhyMode.field.ShortGI = GI_400;
									}
									else
									{
										pAd->CommonCfg.HTPhyMode.field.ShortGI = GI_800;
									}

									// bit0~5 for MCS control
									
									pAd->MacTab.Content[1].HTPhyMode.field.STBC = pAd->CommonCfg.HTPhyMode.field.STBC;
									pAd->MacTab.Content[1].HTPhyMode.field.ShortGI = pAd->CommonCfg.HTPhyMode.field.ShortGI;
									pAd->MacTab.Content[1].HTPhyMode.field.BW = pAd->CommonCfg.HTPhyMode.field.BW;
									pAd->MacTab.Content[1].HTPhyMode.field.MODE = pAd->CommonCfg.HTPhyMode.field.MODE;

									pAd->CommonCfg.bAutoTxRateSwitch = FALSE;

									DBGPRINT(RT_DEBUG_TRACE, ("DRS: STBC=%d, ShortGI=%d, MODE=%d, BW=%d \n",
										pAd->CommonCfg.HTPhyMode.field.STBC, pAd->CommonCfg.HTPhyMode.field.ShortGI,
										pAd->CommonCfg.HTPhyMode.field.MODE, pAd->CommonCfg.HTPhyMode.field.BW));
								}
								}
					DBGPRINT(RT_DEBUG_ERROR, (" HTPhyMode.MODE = 0x%d \n",pAd->CommonCfg.HTPhyMode.field.MODE));
					DBGPRINT(RT_DEBUG_ERROR, (" HTPhyMode.MCS = 0x%d \n",pAd->CommonCfg.HTPhyMode.field.MCS));
					DBGPRINT(RT_DEBUG_ERROR, (" HTPhyMode.BW = 0x%d \n",pAd->CommonCfg.HTPhyMode.field.BW));
					DBGPRINT(RT_DEBUG_ERROR, (" HTPhyMode.ShortGI = 0x%d \n",pAd->CommonCfg.HTPhyMode.field.ShortGI));
					DBGPRINT(RT_DEBUG_ERROR, (" BSSID_WCID.MpduDensity = 0x%d \n",pAd->MacTab.Content[BSSID_WCID].MpduDensity));
				}
				else if ((pHardwareRegister->Data&0xff) == 0x99)
				{

						RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET);
						DBGPRINT_RAW(RT_DEBUG_ERROR, ("USB_RESET_BULK_IN\n"));
						//
						// From DDK,
						// All transfers must be aborted or cancelled before attempting to reset the pipe.
						//	
						while ((atomic_read(&pAd->PendingRx) > 0) && (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST))) 
						{
							if (atomic_read(&pAd->PendingRx) > 0)
							{
								DBGPRINT_RAW(RT_DEBUG_TRACE, ("BulkIn IRP Pending, pAd->PendingRx=%d!!!\n", atomic_read(&pAd->PendingRx)));
								ntStatus = RTUSB_VendorRequest(pAd,
													0,
													DEVICE_VENDOR_REQUEST_OUT,
													0x0C,
													0x0,
													0x0,
													NULL,
													0);
								if (!NT_SUCCESS(ntStatus))
								{
									DBGPRINT_RAW(RT_DEBUG_TRACE, ("Stop RX failed, ntStatus=0x%08x\n", ntStatus));
									break;
								}
							}
						}

						if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST))
						{
						 	UCHAR	i = 0;
							//To distinguish it's really pipe error or fake error which might because temporary miss.
							do
							{
								ntStatus = RTUSB_ResetPipe(pAd, 0);
								i++;
							}
							while ((i<1) && (NT_SUCCESS(ntStatus) != TRUE));
							
							if (NT_SUCCESS(ntStatus) == TRUE)
							{
								USHORT	temp;
								UCHAR	index;
								DBGPRINT_RAW(RT_DEBUG_ERROR, ("Bulk In Reset Successed\n"));

								RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKIN_RESET);
								if ((!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS)) &&
									(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF)) &&
									(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_HALT_IN_PROGRESS)) &&
									(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST)))
								{						
									DBGPRINT_RAW(RT_DEBUG_ERROR, ("NumOfBulkInIRP=%d\n",pAd->CommonCfg.NumOfBulkInIRP));
									for (index = 0; index < pAd->CommonCfg.NumOfBulkInIRP; index ++)
									{
										RTUSBBulkReceive(pAd);
										DBGPRINT_RAW(RT_DEBUG_ERROR, ("RTUSBBulkReceive\n"));
									}
								}
							}
							else
							{
								// Card must be removed
								RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST);
								DBGPRINT_RAW(RT_DEBUG_ERROR, ("Bulk In Reset Failed! Card must be removed, Status=0x%x\n", ntStatus));
								DBGPRINT_RAW(RT_DEBUG_ERROR, ("\n"));
							}
						}

				}

				else if ((pHardwareRegister->Data&0xff) == 0x97)
				{

						RTMP_SET_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
						DBGPRINT_RAW(RT_DEBUG_ERROR, ("USB_RESET_BULK_OUT\n"));
						//
						// From DDK,
						// All transfers must be aborted or cancelled before attempting to reset the pipe.
						//							
						RTUSBRejectPendingPackets(pAd); //reject all NDIS packets waiting in TX queue						
						RTUSBCancelPendingBulkOutIRP(pAd);
						RTUSBCleanUpDataBulkOutQueue(pAd);

						do
						{
							UCHAR	Index;
							

							for (Index=1; Index < pAd->NumberOfPipes; Index++)
							{
								if (!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_NIC_NOT_EXIST))
								{
									ntStatus = RTUSB_ResetPipe(pAd, Index);
									if (NT_SUCCESS(ntStatus) == TRUE)
									{
										DBGPRINT(RT_DEBUG_TRACE, ("Bulk Out Reset on Pipe[=%d] Successed\n", Index));
									}
	 								else
									{
										DBGPRINT(RT_DEBUG_TRACE, ("Bulk Out Reset on Pipe[=%d] Failed, Status=0x%x\n", Index, ntStatus));
									}
								}
							}
								
							if (NT_SUCCESS(ntStatus) == TRUE)
							{
								RTMP_CLEAR_FLAG(pAd, fRTMP_ADAPTER_BULKOUT_RESET);
								for (Index = 0; Index < 4; Index++)
								{
									if(pAd->SendTxWaitQueue[Index].Number > 0)
									{
										RTMPDeQueuePacket(pAd, Index);
									}
								}

								RTUSBKickBulkOut(pAd);
							}

						} while (FALSE);
				}

			}
			// 0x2bf4: test H2M_MAILBOX. byte3: Host command, byte2: token, byte1-0: arguments
			else if (pHardwareRegister->Offset == (HW_DEBUG_SETTING_BASE + 4))
			{
							RTUSBReadMACRegister(pAd, TX_LINK_CFG, &Value);
							Value &= (~0x800);
							RTUSBWriteMACRegister(pAd, TX_LINK_CFG, Value);
							if (pHardwareRegister->Data ==0x11)
							{
								OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED);
								pAd->HTCEnable = TRUE;
								pAd->CommonCfg.bRdg = TRUE;	
								pAd->bLinkAdapt = FALSE;	
								RTUSBReadMACRegister(pAd, TX_LINK_CFG, &Value);
								Value |= 0x800;
								RTUSBWriteMACRegister(pAd, TX_LINK_CFG, Value);
								DBGPRINT(RT_DEBUG_ERROR, ("HTCEnable = %d , RDG = %d\n", pAd->HTCEnable, pAd->CommonCfg.bRdg));
							}
							else if (pHardwareRegister->Data ==0x21)
							{
								OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED);
								pAd->HTCEnable = TRUE;
								pAd->bLinkAdapt = TRUE;	
								pAd->CommonCfg.bRdg = FALSE;	
								RTUSBReadMACRegister(pAd, TX_LINK_CFG, &Value);
								Value |= 0x400;
								RTUSBWriteMACRegister(pAd, TX_LINK_CFG, Value);
							}
							else if (pHardwareRegister->Data ==0x1)
							{
								OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED);
								pAd->HTCEnable = TRUE;
								pAd->CommonCfg.bRdg = FALSE;	
								pAd->bLinkAdapt = FALSE;	
							}
							else
							{
								pAd->HTCEnable = FALSE;
								pAd->CommonCfg.bRdg = FALSE;	
							}
							
							DBGPRINT(RT_DEBUG_ERROR, ("HTCEnable = %d , RDG = %d\n", pAd->HTCEnable, pAd->CommonCfg.bRdg));
			}
			// 0x2bf8: test ACK policy and QOS format in ADHOC mode
			else if (pHardwareRegister->Offset == (HW_DEBUG_SETTING_BASE + 8))
			{
				PUCHAR pAckStr[4] = {"NORMAL", "NO-ACK", "NO-EXPLICIT-ACK", "BLOCK-ACK"};
				EDCA_PARM DefaultEdcaParm;

				// byte0 b1-0 means ACK POLICY - 0: normal ACK, 1: no ACK, 2:no explicit ACK, 3:BA
				pAd->CommonCfg.AckPolicy[0] = ((UCHAR)pHardwareRegister->Data & 0x02) << 5;
				pAd->CommonCfg.AckPolicy[1] = ((UCHAR)pHardwareRegister->Data & 0x02) << 5;
				pAd->CommonCfg.AckPolicy[2] = ((UCHAR)pHardwareRegister->Data & 0x02) << 5;
				pAd->CommonCfg.AckPolicy[3] = ((UCHAR)pHardwareRegister->Data & 0x02) << 5;
				DBGPRINT(RT_DEBUG_EMU, ("ACK policy = %s\n", pAckStr[(UCHAR)pHardwareRegister->Data & 0x02]));

				// any non-ZERO value in byte1 turn on EDCA & QOS format
				if (pHardwareRegister->Data & 0x0000ff00) 
				{
					NdisZeroMemory(&DefaultEdcaParm, sizeof(EDCA_PARM));
					DefaultEdcaParm.bValid = TRUE;
					DefaultEdcaParm.Aifsn[0] = 3;
					DefaultEdcaParm.Aifsn[1] = 7;
					DefaultEdcaParm.Aifsn[2] = 2;
					DefaultEdcaParm.Aifsn[3] = 2;

					DefaultEdcaParm.Cwmin[0] = 4;
					DefaultEdcaParm.Cwmin[1] = 4;
					DefaultEdcaParm.Cwmin[2] = 3;
					DefaultEdcaParm.Cwmin[3] = 2;

					DefaultEdcaParm.Cwmax[0] = 10;
					DefaultEdcaParm.Cwmax[1] = 10;
					DefaultEdcaParm.Cwmax[2] = 4;
					DefaultEdcaParm.Cwmax[3] = 3;

					DefaultEdcaParm.Txop[0]  = 0;
					DefaultEdcaParm.Txop[0]  = (USHORT)(pHardwareRegister->Data&0xff);
					DefaultEdcaParm.Txop[1]  = 0;
					DefaultEdcaParm.Txop[2]  = 96;
					DefaultEdcaParm.Txop[3]  = 48;
					AsicSetEdcaParm(pAd, &DefaultEdcaParm);
				}
				else
					AsicSetEdcaParm(pAd, NULL);
			}
			// 0x2bfc: turn ON/OFF TX aggregation
			else if (pHardwareRegister->Offset == (HW_DEBUG_SETTING_BASE + 12))
			{
				if (pHardwareRegister->Data)
					OPSTATUS_SET_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED);
				else
					OPSTATUS_CLEAR_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED);
				DBGPRINT(RT_DEBUG_TRACE, ("AGGREGATION = %d\n", OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED)));
			}
			else
				Status = NDIS_STATUS_NOT_SUPPORTED;				
		}
	}
	else if (pHardwareRegister->HardwareType == HARDWARE_BBP)
	{
		bbpID = (UCHAR) pHardwareRegister->Offset;
		bbpValue = (UCHAR) pHardwareRegister->Data;
		RTUSBWriteBBPRegister(pAd, bbpID, bbpValue);
		DBGPRINT(RT_DEBUG_TRACE, ("BBP:ID[0x%02x]=[0x%02x]\n", bbpID, bbpValue));		
	}
	else if (pHardwareRegister->HardwareType == HARDWARE_EEP)
	{
		eepID = (USHORT) pHardwareRegister->Offset;
		eepValue= (USHORT) pHardwareRegister->Data;
		RTUSBWriteEEPROM(pAd, eepID, (PUCHAR)&eepValue, 2);
		DBGPRINT(RT_DEBUG_TRACE, ("EEP:ID[0x%04x]=[0x%04x]\n", eepID, eepValue));		
	}
	Status = NDIS_STATUS_SUCCESS;
	return Status;
}
#endif
#endif

